/** @file fifo_model.h
 * This file contains the reference model class declaration.
 *
 * @author Mikhail Chupilko (chupilko@ispras.ru)
 */

#pragma once

#include <hw/model.hpp>

#include <fifo_coverage.h>
#include <fifo_msg.h>

namespace cpptesk {
namespace examples {
namespace fifo {

/**
 * @brief FIFO reference model
 */
MODEL(FIFO) {
public:
    FIFO();
    virtual ~FIFO();

    /// Input interface iface1 to write
    DECLARE_INPUT(iface1);
    /// Input interface iface2 to start reading
    DECLARE_INPUT(iface2);
    /// Output interface iface3 to receive results of the reading
    DECLARE_OUTPUT(iface3);

    /// Check if iface1 is ready. @return iface1.isReady()
    virtual bool isIface1Ready() const;
    /// Check if iface2 is ready. @return iface2.isReady()
    virtual bool isIface2Ready() const;

    /// Stimulus "Write data into FIFO"
    DECLARE_STIMULUS(push_msg);
    /// Stimulus "Start reading data from FIFO"
    DECLARE_STIMULUS(pop_msg);
    /// Reaction "Receive result of reading from FIFO"
    DECLARE_REACTION(get_pop_msg);

    /// Push data into the FIFO reference model
    void push_item(uint8_t data);
    /// Pop and return data from the FIFO reference model
    uint8_t pop_item(void);
    /// Check whether the FIFO reference model is empty
    bool is_empty(void);
    /// Check whether the FIFO reference model is full
    bool is_full(void);
    /// Return fullness of the FIFO reference model. @return size()
    int fifo_fullness(void);

    /// Functional coverage
    FIFOCoverage coverage;
    /// Update functional coverage values (this function is usually called in NOP scenario function)
    void refresh_coverage(void);
protected:
    /// FIFO reference model
    std::vector<uint8_t> fifo;
};

}}}

