/** @file multififo_model.cpp
 * Model of united FIFOs.
 *
 * @author Mikhail Chupilko (chupilko@ispras.ru)
 */

#include <multififo_model.h>

namespace cpptesk {
namespace examples {
namespace fifo {

Multififo::Multififo() : fifo0(), fifo1(), fifo2()
{
    /// Registration of interfaces in the test system.
    CPPTESK_ADD_INPUT(iface1a);
    CPPTESK_ADD_INPUT(iface1b);
    CPPTESK_ADD_INPUT(iface2);
    CPPTESK_ADD_OUTPUT(iface3);
}

Multififo::~Multififo() {}

bool Multififo::isIface1aReady() const
{
    /// return true if iface1a is not busy by any operations.
    return iface1a.isReady();
}

bool Multififo::isIface1bReady() const
{
    /// return true if iface1b is not busy by any operations.
    return iface1b.isReady();
}

bool Multififo::isIface2Ready() const
{
    /// return true if iface2 is not busy by any operations.
    return iface2.isReady();
}

CPPTESK_DEFINE_STIMULUS(Multififo::push_msg_a)
{
    // Copy data to local variable.
    InputData data = CPPTESK_CAST_MESSAGE(InputData);

    // Apply the stimulus to the selected iface in parallel mode.
    CPPTESK_START_STIMULUS(PARALLEL);

    // Refresh model of Multififo.
    fifo0.push_item(data.get_data());
    CPPTESK_CALL_STIMULUS_OF(fifo0, FIFO::pop_msg, fifo0.iface2, data);

    // The process waits for 1 cycle.
    CPPTESK_CYCLE();

    // Finish of stimulus processing.
    CPPTESK_STOP_STIMULUS();
}

CPPTESK_DEFINE_STIMULUS(Multififo::push_msg_b)
{
    // Copy data to local variable.
    InputData data = CPPTESK_CAST_MESSAGE(InputData);

    // Apply the stimulus ot the selected iface in parallel mode.
    CPPTESK_START_STIMULUS(PARALLEL);

    // Refresh model of Multififo.
    fifo1.push_item(data.get_data());
    CPPTESK_CALL_STIMULUS_OF(fifo1, FIFO::pop_msg, fifo1.iface2, data);

    // The process Waits for 1 cycle
    CPPTESK_CYCLE();

    // Finish of stimulus processing.
    CPPTESK_STOP_STIMULUS();
}

CPPTESK_DEFINE_STIMULUS(Multififo::pop_msg)
{
    // Copy data to local variable.
    InputData data = CPPTESK_CAST_MESSAGE(InputData);

    // Start of stimulus processing.
    CPPTESK_START_STIMULUS(PARALLEL);

    // Preparation of model reaction to check data given from DUT.
    OutputData outdata = OutputData(fifo2.pop_item());

    for(int i = 0; i < 3; i++)
        CPPTESK_CYCLE();

    // Start reaction "take output data as a result of reading".
    get_pop_msg(process, iface3, outdata);

    // The process Waits for 1 cycle
    CPPTESK_CYCLE();

    // Finish of stimulus processing.
    CPPTESK_STOP_STIMULUS();
}

CPPTESK_DEFINE_REACTION(Multififo::get_pop_msg)
{
    CPPTESK_START_REACTION(SEQUENTIAL);
    CPPTESK_STOP_REACTION();
}

}}}
