/*
 * Copyright 2010-2012 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Debug printing facilities 
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/errors.hpp>
#include <hw/process/process_stack.hpp>

#include <stdio.h>

#include <iostream>
#include <fstream>
#include <sstream>
#include <string>

///////////////////////////////////////////////////////////////////////////////
// Debug
///////////////////////////////////////////////////////////////////////////////

#define __CPPTESK_DEBUG_BUFFER_SIZE 1024

#define CPPTESK_DEBUG_STYLE(fg_color, bg_color) \
    (::cpptesk::hw::debug_style_t){fg_color, bg_color}

#define CPPTESK_DEBUG_STREAM() std::cout

#define CPPTESK_PRINT_SEPARATOR() \
    Debug::get().print_separator(CPPTESK_DEBUG_STREAM())

#ifdef HW_SHORT_NAMES
    #define DEBUG_STYLE(fg_color, bg_color) CPPTESK_DEBUG_STYLE(fg_color, bg_color)
    #define DEBUG_STREAM()                  CPPTESK_DEBUG_STREAM()
    #define PRINT_SEPARATOR()               CPPTESK_PRINT_SEPARATOR()
#endif // HW_SHORT_NAMES

#define CPPTESK_COLORED_DEBUG_PRINT(debug_level, fg_color, bg_color, debug_message) \
    do { \
        /* Dump process stack if an error or warning occured */ \
        const bool __cpptesk_dump_stack = debug_level == ::cpptesk::hw::DEBUG_ERROR || \
                                          debug_level == ::cpptesk::hw::DEBUG_WARNING; \
        std::stringstream __cpptesk_buffer; \
        __cpptesk_buffer << debug_message; \
        \
        if(__cpptesk_dump_stack && __cpptesk_inside_process && __cpptesk_process_stack) { \
            __cpptesk_buffer << std::endl << *((::cpptesk::hw::process_stack_t*)__cpptesk_process_stack); \
        } \
        ::cpptesk::hw::Debug::get().print(debug_level, __cpptesk_buffer.str(), fg_color, bg_color); \
    } while(false)

#define CPPTESK_DEBUG_PRINT(debug_level, debug_message) \
    CPPTESK_COLORED_DEBUG_PRINT(debug_level, ::cpptesk::hw::DEFAULT, ::cpptesk::hw::DEFAULT, debug_message)

#define CPPTESK_COLORED_DEBUG_PRINTF(debug_level, fg_color, bg_color, debug_message, ...) \
    do { \
        char __cpptesk_buffer[__CPPTESK_DEBUG_BUFFER_SIZE]; \
        if(snprintf(__cpptesk_buffer, __CPPTESK_DEBUG_BUFFER_SIZE, debug_message, __VA_ARGS__) >= __CPPTESK_DEBUG_BUFFER_SIZE) { \
            CPPTESK_ASSERTION(false, "Too long debug message"); \
        } else { \
            ::cpptesk::hw::Debug::get().print(debug_level, __cpptesk_buffer, fg_color, bg_color); \
        } \
    } while(false)

#define CPPTESK_DEBUG_PRINTF(debug_level, debug_message, ...) \
    CPPTESK_COLORED_DEBUG_PRINTF(debug_level, ::cpptesk::hw::DEFAULT, ::cpptesk::hw::DEFAULT, debug_message, __VA_ARGS__)

#define __CPPTESK_COLORED_DEBUG_PRINT_FILE_LINE(debug_level, fg_color, bg_color, debug_message, file, line) \
    CPPTESK_COLORED_DEBUG_PRINT(debug_level, fg_color, bg_color, debug_message << " (" << file << ":" << std::dec << line << std::hex << ")")

#define __CPPTESK_DEBUG_PRINT_FILE_LINE(debug_level, debug_message, file, line) \
    __CPPTESK_COLORED_DEBUG_PRINT_FILE_LINE(debug_level, ::cpptesk::hw::DEFAULT, ::cpptesk::hw::DEFAULT, debug_message, file, line)

#define __CPPTESK_COLORED_DEBUG_PRINTF_FILE_LINE(debug_level, fg_color, bg_color, debug_message, file, line, ...) \
    CPPTESK_COLORED_DEBUG_PRINTF(debug_level, fg_color, bg_color, debug_message " (%s:%d)", __VA_ARGS__, file, line)

#define __CPPTESK_DEBUG_PRINTF_FILE_LINE(debug_level, debug_message, file, line, ...) \
    __CPPTESK_COLORED_DEBUG_PRINTF_FILE_LINE(debug_level, ::cpptesk::hw::DEFAULT, ::cpptesk::hw::DEFAULT, debug_message, file, line, __VA_ARGS__)

#define CPPTESK_COLORED_DEBUG_PRINT_FILE_LINE(debug_level, fg_color, bg_color, debug_message) \
    __CPPTESK_COLORED_DEBUG_PRINT_FILE_LINE(debug_level, fg_color, bg_color, debug_message, __FILE__, __LINE__)

#define CPPTESK_DEBUG_PRINT_FILE_LINE(debug_level, debug_message) \
    CPPTESK_COLORED_DEBUG_PRINT_FILE_LINE(debug_level, ::cpptesk::hw::DEFAULT, ::cpptesk::hw::DEFAULT, debug_message)

#define CPPTESK_COLORED_DEBUG_PRINTF_FILE_LINE(debug_level, fg_color, bg_color, debug_message, ...) \
    __CPPTESK_COLORED_DEBUG_PRINTF_FILE_LINE(debug_level, fg_color, bg_color, debug_message, __FILE__, __LINE__, __VA_ARGS__)

#define CPPTESK_DEBUG_PRINTF_FILE_LINE(debug_level, debug_message, ...) \
    CPPTESK_COLORED_DEBUG_PRINTF_FILE_LINE(debug_level, ::cpptesk::hw::DEFAULT, ::cpptesk::hw::DEFAULT, debug_message, __VA_ARGS__)

#define CPPTESK_SET_DEBUG_LEVEL(debug_level, debug_color) \
    ::cpptesk::hw::debug(debug_level, debug_color)

#define CPPTESK_SET_DEBUG_STYLE(debug_level, fg_level, bg_level, fg_message, bg_message) \
    ::cpptesk::hw::Debug::get().setStyle(debug_level, fg_level, bg_level, fg_message, bg_message)

#define CPPTESK_COLORED_USER_PRINT(fg_color, bg_color, debug_message) \
    CPPTESK_COLORED_DEBUG_PRINT(::cpptesk::hw::DEBUG_USER, fg_color, bg_color, debug_message)

#define CPPTESK_USER_PRINT(debug_message) \
    CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_USER, debug_message)

#define CPPTESK_COLORED_USER_PRINT_FILE_LINE(fg_color, bg_color, debug_message) \
    CPPTESK_COLORED_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_USER, fg_color, bg_color, debug_message)

#define CPPTESK_USER_PRINT_FILE_LINE(debug_message) \
    CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_USER, debug_message)

#define CPPTESK_COLORED_USER_PRINTF(fg_color, bg_color, debug_message, ...) \
    CPPTESK_COLORED_DEBUG_PRINTF(::cpptesk::hw::DEBUG_USER, fg_color, bg_color, debug_message, __VA_ARGS__)

#define CPPTESK_USER_PRINTF(debug_message, ...) \
    CPPTESK_DEBUG_PRINTF(::cpptesk::hw::DEBUG_USER, debug_message, __VA_ARGS__)

#define CPPTESK_COLORED_USER_PRINTF_FILE_LINE(fg_color, bg_color, debug_message, ...) \
    CPPTESK_COLORED_DEBUG_PRINTF_FILE_LINE(::cpptesk::hw::DEBUG_USER, fg_color, bg_color, debug_message, __VA_ARGS__)

#define CPPTESK_USER_PRINTF_FILE_LINE(debug_message, ...) \
    CPPTESK_DEBUG_PRINTF_FILE_LINE(::cpptesk::hw::DEBUG_USER, debug_message, __VA_ARGS__)

////////////////////////////////////////////////////////////////////////////////
// Debug printing indent
////////////////////////////////////////////////////////////////////////////////

#define CPPTESK_SET_DEBUG_INDENT(debug_indent) \
    ::cpptesk::hw::Debug::get().setIndent(debug_indent)

#define CPPTESK_DEBUG_INDENT_BEGIN \
    ::cpptesk::hw::Debug::get().incIndent()

#define CPPTESK_START_DEBUG_INDENT() \
    ::cpptesk::hw::Debug::get().incIndent()

#define CPPTESK_DEBUG_INDENT_END \
    ::cpptesk::hw::Debug::get().decIndent()

#define CPPTESK_STOP_DEBUG_INDENT() \
    ::cpptesk::hw::Debug::get().decIndent()

////////////////////////////////////////////////////////////////////////////////
// Duplicating debug output to a file
////////////////////////////////////////////////////////////////////////////////

#define CPPTESK_SET_DEBUG_LOGFILE(file_name) \
    ::cpptesk::hw::Debug::get().setLogFile(file_name)

////////////////////////////////////////////////////////////////////////////////
// Debug printing short names
////////////////////////////////////////////////////////////////////////////////

#ifdef HW_SHORT_NAMES
    #define COLORED_DEBUG_PRINT(debug_level, fg_color, bg_color, debug_message)                 CPPTESK_COLORED_DEBUG_PRINT(debug_level, fg_color, bg_color, debug_message)
    #define DEBUG_PRINT(debug_level, debug_message)                                             CPPTESK_DEBUG_PRINT(debug_level, debug_message)
    #define COLORED_DEBUG_PRINTF(debug_level, fg_color, bg_color, debug_message, ...)           CPPTESK_COLORED_DEBUG_PRINTF(debug_level, fg_color, bg_color, debug_message, __VA_ARGS__)
    #define DEBUG_PRINTF(debug_level, debug_message, ...)                                       CPPTESK_DEBUG_PRINTF(debug_level, debug_message, __VA_ARGS__)
    #define COLORED_DEBUG_PRINT_FILE_LINE(debug_level,fg_color, bg_color, debug_message)        CPPTESK_COLORED_DEBUG_PRINT_FILE_LINE(debug_level, fg_color, bg_color, debug_message)
    #define DEBUG_PRINT_FILE_LINE(debug_level, debug_message)                                   CPPTESK_DEBUG_PRINT_FILE_LINE(debug_level, debug_message)
    #define COLORED_DEBUG_PRINTF_FILE_LINE(debug_level, fg_color, bg_color, debug_message, ...) CPPTESK_COLORED_DEBUG_PRINTF_FILE_LINE(debug_level, fg_color, bg_color, debug_message, __VA_ARGS__)
    #define DEBUG_PRINTF_FILE_LINE(debug_level, debug_message, ...)                             CPPTESK_DEBUG_PRINTF(debug_level, debug_message, __VA_ARGS__)
    #define COLORED_USER_PRINT(fg_color, bg_color, debug_message)                               CPPTESK_COLORED_USER_PRINT(fg_color, bg_color, debug_message)
    #define USER_PRINT(debug_message)                                                           CPPTESK_USER_PRINT(debug_message)
    #define COLORED_USER_PRINT_FILE_LINE(fg_color, bg_color, debug_message)                     CPPTESK_COLORED_USER_PRINT_FILE_LINE(fg_color, bg_color, debug_message)
    #define USER_PRINT_FILE_LINE(debug_message)                                                 CPPTESK_USER_PRINT_FILE_LINE(debug_message)
    #define COLORED_USER_PRINTF(fg_color, bg_color, debug_message, ...)                         CPPTESK_COLORED_USER_PRINTF(fg_color, bg_color, debug_message, __VA_ARGS__)
    #define USER_PRINTF(debug_message, ...)                                                     CPPTESK_USER_PRINTF(debug_message, __VA_ARGS__)
    #define COLORED_USER_PRINTF_FILE_LINE(fg_color, bg_color, debug_message, ...)               CPPTESK_COLORED_USER_PRINTF_FILE_LINE(fg_color, bg_color, debug_message, __VA_ARGS__)
    #define USER_PRINTF_FILE_LINE(debug_message, ...)                                           CPPTESK_USER_PRINTF_FILE_LINE(debug_message, __VA_ARGS__)
    #define SET_DEBUG_LEVEL(debug_level, debug_color)                                           CPPTESK_SET_DEBUG_LEVEL(debug_level, debug_color)
    #define SET_DEBUG_STYLE(debug_level, fg_level, bg_level, fg_message, bg_message)            CPPTESK_SET_DEBUG_STYLE(debug_level, fg_level, bg_level, fg_message, bg_message)
    #define SET_DEBUG_INDENT(debug_indent)                                                      CPPTESK_SET_DEBUG_INDENT(debug_indent)
    #define DEBUG_INDENT_BEGIN                                                                  CPPTESK_DEBUG_INDENT_BEGIN
    #define DEBUG_INDENT_END                                                                    CPPTESK_DEBUG_INDENT_END
    #define START_DEBUG_INDENT()                                                                CPPTESK_START_DEBUG_INDENT()
    #define STOP_DEBUG_INDENT()                                                                 CPPTESK_STOP_DEBUG_INDENT()
    #define SET_DEBUG_LOGFILE(file_name)                                                        CPPTESK_SET_DEBUG_LOGFILE(file_name)
#endif // HW_SHORT_NAMES

namespace cpptesk {
namespace hw {

enum debug_level_t
{
    /// Print all messages.
    DEBUG_TOOL     =  5,
    DEBUG_INTERNAL =  5,

    ////////////////////////////////////////////////////////////////////////////
    // Internal Debug Levels
    ////////////////////////////////////////////////////////////////////////////
    DEBUG_ALL      =  4,
    ////////////////////////////////////////////////////////////////////////////
    // External Debug Levels
    ////////////////////////////////////////////////////////////////////////////

    /// Print all events including raw-level ones.
    DEBUG_MORE     =  4,
    DEBUG_DETAILED =  4,
    /// Print basic events only.
    DEBUG_INFO     =  3,
    DEBUG_BASIC    =  3,
    /// Print messages, warnings and errors.
    DEBUG_USER     =  2,
    DEBUG_MESSAGE  =  2,
    /// Print warnings and errors.
    DEBUG_WARN     =  1,
    DEBUG_WARNING  =  1,
    /// Print errors.
    DEBUG_FAIL     =  0,
    DEBUG_ERROR    =  0,
    /// Do not print debug messages.
    DEBUG_NONE     = -1,
};

enum debug_color_t
{
    DEFAULT = -2,
    NORMAL  = -1,
    BLACK   =  0,
    RED     =  1,
    GREEN   =  2,
    YELLOW  =  3,
    BLUE    =  4,
    MAGENTA =  5,
    CYAN    =  6,
    WHITE   =  7,
};

struct debug_style_t
{
    debug_color_t fg_color;
    debug_color_t bg_color;
};

std::ostream& operator << (std::ostream& out, const debug_style_t &style);

class Debug
{
public:
    void setLevel(debug_level_t debug_level);

    void enableColoredPrint(bool debug_color);

    debug_style_t getMessageStyle(debug_level_t debug_level);

    void setStyle(debug_level_t debug_level, debug_color_t fg_level, debug_color_t bg_level,
        debug_color_t fg_message, debug_color_t bg_message);

    /// Returns the old indent value.
    unsigned setIndent(unsigned debug_indent);

    void incIndent();
    void decIndent();

    void setLogFile(const char* file_name);

    /// Prints a debug message.
    void print(debug_level_t debug_level, const std::string &debug_message,
        debug_color_t fg_color = DEFAULT, debug_color_t bg_color= DEFAULT);

    /// Prints a separator.
    void print_separator(std::ostream &out)
    {
        out.width(80);
        out.fill('-');
        out << "" << std::endl;
    }

    /// Prints a message number.
    void print_number(std::ostream &out, unsigned number)
    {
        out.width(6);
        out.fill('0');
        out << std::right << std::dec << number << " " << std::hex;
    }

    /// Prints a message tag.
    void print_tag(std::ostream &out, const std::string &tag)
    {
        out << "[" << tag << "]";
    }

    /// Prints a message indent.
    void print_indent(std::ostream &out, unsigned indent)
    {
        out.width(4 * indent);
        out.fill(' ');
        out << " ";
    }

    /// Prints a message.
    void print_message(std::ostream &out, const std::string &message)
    {
        out << message;
    }

    /// Prints a new line.
    void print_newline(std::ostream &out)
    {
        out << std::endl;
    }

private:
    struct debug_config_t
    {
        std::string   tag;
        debug_style_t lvl_style;
        debug_style_t msg_style;
    };

private:
    Debug() : _debug_level(DEBUG_INFO), _debug_color(true), _last_debug_level(DEBUG_NONE),
              _debug_count(0), _debug_indent(0), _debug_logfile(0)
    {
        _debug_config[0].tag = "FAIL"; _debug_config[0].lvl_style.fg_color = BLACK;
                                       _debug_config[0].lvl_style.bg_color = RED;
                                       _debug_config[0].msg_style.fg_color = RED;
                                       _debug_config[0].msg_style.bg_color = NORMAL;

        _debug_config[1].tag = "WARN"; _debug_config[1].lvl_style.fg_color = BLACK;
                                       _debug_config[1].lvl_style.bg_color = YELLOW;
                                       _debug_config[1].msg_style.fg_color = YELLOW;
                                       _debug_config[1].msg_style.bg_color = NORMAL;

        _debug_config[2].tag = "USER"; _debug_config[2].lvl_style.fg_color = NORMAL;
                                       _debug_config[2].lvl_style.bg_color = NORMAL;
                                       _debug_config[2].msg_style.fg_color = NORMAL;
                                       _debug_config[2].msg_style.bg_color = NORMAL;

        _debug_config[3].tag = "INFO"; _debug_config[3].lvl_style.fg_color = BLACK;
                                       _debug_config[3].lvl_style.bg_color = GREEN;
                                       _debug_config[3].msg_style.fg_color = NORMAL;
                                       _debug_config[3].msg_style.bg_color = NORMAL;

        _debug_config[4].tag = "MORE"; _debug_config[4].lvl_style.fg_color = BLACK;
                                       _debug_config[4].lvl_style.bg_color = GREEN;
                                       _debug_config[4].msg_style.fg_color = NORMAL;
                                       _debug_config[4].msg_style.bg_color = NORMAL;

        _debug_config[5].tag = "TOOL"; _debug_config[5].lvl_style.fg_color = BLACK;
                                       _debug_config[5].lvl_style.bg_color = MAGENTA;
                                       _debug_config[5].msg_style.fg_color = NORMAL;
                                       _debug_config[5].msg_style.bg_color = NORMAL;
    }

public:
   ~Debug() {}
   static Debug &get() { if(!_instance) { _instance = new Debug(); } return *_instance; }

protected:
   static Debug *_instance;

private:
    debug_config_t _debug_config[DEBUG_TOOL + 1];
    debug_level_t  _debug_level;
    bool           _debug_color;

    debug_level_t  _last_debug_level;
    unsigned       _debug_count;

    unsigned       _debug_indent;

    /// To duplicate debug output into a file (feature #1900).
    std::ofstream *_debug_logfile;
};

inline void debug(debug_level_t debug_level = DEBUG_ALL, bool debug_color = true)
{
    ::cpptesk::hw::Debug::get().setLevel(debug_level);
    ::cpptesk::hw::Debug::get().enableColoredPrint(debug_color);
}

}} // namespace cpptesk::hw

