/* Copyright 2010-2012 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Protsenko (protsenko@ispras.ru)
// SYNOPSIS: Verdict generation subsystem main file
///////////////////////////////////////////////////////////////////////////////

#include <hw/diagnostics/diagnostics.hpp>

namespace cpptesk {
namespace hw {
namespace diagnostics {

Diagnostics::Diagnostics(const event_list_t &reactions)
{
    // Copy reactions for analysis
    _reactionSet.initialization(reactions);
}

void Diagnostics::run()
{
    // Initialize list of rules and register all of the rules
    _errorModel.initialization();
    
    if(!_reactionSet.isEmpty())
    {
        int n1, n2; // numbers of the first and second reactions

        // Try to apply normal rules for collapsing of reactions
        for(n1 = 0; n1 < _reactionSet.size(); n1++)
        {
            ReactionUnit &r1 = _reactionSet.getReactionUnit(n1);

            if(!r1.isCollapsed())
            {
                for(n2 = 0; n2 < _reactionSet.size(); n2++)
                {
                    ReactionUnit &r2 = _reactionSet.getReactionUnit(n2);

                    if(!r2.isCollapsed())
                    {
                        ReactionSet reactionSet = _errorModel.transform(r2, r1);
                        _reactionSet.addReactionSet(reactionSet, n1, n2);
                        break;
                    }
                }
            }
        }

        FuzzyMatchResult global_result, local_result;

        // Try to apply fuzzy rules for collapsing of reactions
        for(n1 = 0; n1 < _reactionSet.size(); n1++)
        {
            global_result.weight = cpptesk::hw::diagnostics::Rule::FUZZY_MATCH;
            global_result.rule_number = cpptesk::hw::diagnostics::Rule::EMPTY_RULE;

            cpptesk::hw::diagnostics::ReactionUnitPair rupair;

            ReactionUnit &r1 = _reactionSet.getReactionUnit(n1);

            if(!r1.isCollapsed())
            {
                for(n2 = 0; n2 < _reactionSet.size(); n2++)
                {
                    ReactionUnit &r2 = _reactionSet.getReactionUnit(n2);

                    if(!r2.isCollapsed())
                    {
                        local_result = _errorModel.fuzzyMatch(r2.getReaction(), r1.getReaction());

                        if(local_result.weight < global_result.weight)
                        {
                            global_result = local_result;
                            rupair.r1 = &r1;
                            rupair.r2 = &r2;
                            rupair.r1_id = n1;
                            rupair.r2_id = n2;
                        }
                    }
                }

                if(global_result.weight < cpptesk::hw::diagnostics::Rule::FUZZY_MATCH)
                {
                    ReactionSet reactionSet =
                         _errorModel.fuzzyTransform(*(rupair.r1), *(rupair.r2),
                                                    global_result.rule_number);

                    _reactionSet.addReactionSet(reactionSet, rupair.r1_id, rupair.r2_id);

                    _reactionSet.getReactionUnit(rupair.r1_id).setCollapsed();
                    _reactionSet.getReactionUnit(rupair.r1_id).setCollapseRule(
                                                                   global_result.rule_number);
                    _reactionSet.getReactionUnit(rupair.r2_id).setCollapsed();
                    _reactionSet.getReactionUnit(rupair.r2_id).setCollapseRule(
                                                                   global_result.rule_number);
                }
            }
        }
    }
    else
    {
        CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_TOOL, "Reaction set is empty!");
    }

    // printVector(true);
    printDiag();
    
    _reactionSet.clear();
    _errorModel.removeRules();
}

void Diagnostics::printVector(bool debug)
{
    bool header_was_shown = false;

    if(debug)
        CPPTESK_SET_DEBUG_LEVEL(cpptesk::hw::DEBUG_TOOL, true);

    for(int i = 0; i < _reactionSet.size(); i++)
    {
        ReactionUnit &runit = _reactionSet.getReactionUnit(i);
      //  Reaction &reaction = runit.getReaction();
 
    /*    CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_TOOL, ".");
        CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_TOOL, "#" << i << "================================");
        CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_TOOL, "Reaction / Vector Size = " << i <<
                                                     "/" <<_reactionSet.size());
*/
       // if(reaction.isError())
      //  {
            int parent1, parent2;

            runit.getParents(&parent1, &parent2);

            //if((parent1 == -1 || parent2 == -1) && !debug) continue;

            if(!header_was_shown)
            {
                CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_ERROR, "Error diagnostics: ");
                header_was_shown = true;
            }

            CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_TOOL, "Reaction number in the vector: " << i);

            // Out: print short description, impl & spec message
            runit.printDescription(cpptesk::hw::DEBUG_ERROR);

            CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_TOOL,
                "Collapsed: " << (runit.isCollapsed()? "Yes" : "No") << ", " <<
                "Collapse Rule: " << runit.getCollapseRule());

            if(parent1 != -1 && parent2 != -1)
            {
                int parents[2] = {parent1, parent2};

                CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_TOOL,"Reaction Parents:");

                for(unsigned int j = 0; j < 2; j++)
                {
                    CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_TOOL, " " << parents[j]);

                    ReactionUnit &parent = _reactionSet.getReactionUnit(parents[j]);
                    Reaction &r = parent.getReaction();

                    if(j == 0)
                        CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_ERROR,
                             "Parent 1 on interface: " << (r.where())->getName());
                    else
                        CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_ERROR,
                             "Parent 2 on interface: " << (r.where())->getName() << std::endl);

                    parent.printDescription(cpptesk::hw::DEBUG_TOOL);

                    CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_TOOL,
                        "Collapsed: " << (parent.isCollapsed() ? "Yes" : "No") << ", " <<
                        "Collapse Rule: " << parent.getCollapseRule());
                }
            }
       // }
    }
}

void Diagnostics::printDiag()
{
    ReactionGroup reaction_group;
    
    int parent_1, parent_2;
    for(int i = 0; i < _reactionSet.size(); i++)
    {
        ReactionUnit &runit = _reactionSet.getReactionUnit(i);
        runit.getParents(&parent_1, &parent_2);
        Reaction &r = runit.getReaction();

        //CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_INFO, "P1=" << parent_1 << " P2=" << parent_2 << " | ");

        if((parent_1 != -1 || parent_2 != -1) && (!runit.isCollapsed() || (r.getError() == NORMAL_REACTION)))
        {
            //CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_ERROR, "Collapsed| P1=" << parent_1 << " P2=" << parent_2 << " | ");

            ReactionPair reaction_pair;
            
            reaction_pair.addCollapsedReactionUnit(runit);
                        
            // Original specification predecessor
            if(runit.getParentSpec() != -1)
            {
                ReactionUnit parent_spec = _reactionSet.getReactionUnit(runit.getParentSpec());

                while(parent_spec.getParentSpec() != -1)
                {
                    parent_spec = _reactionSet.getReactionUnit(parent_spec.getParentSpec());
                }

                reaction_pair.addSpecReactionUnit(parent_spec);
            }

            // Original implementation predecessor
            if(runit.getParentImpl() != -1)
            {
                ReactionUnit parent_impl = _reactionSet.getReactionUnit(runit.getParentImpl());
                
                while(parent_impl.getParentImpl() != -1)
                {
                    parent_impl = _reactionSet.getReactionUnit(parent_impl.getParentImpl());
                }

                reaction_pair.addImplReactionUnit(parent_impl);
            }

            reaction_pair.setCollapsed();
            reaction_pair.findSimilarity();
            reaction_group.addReactionPair(reaction_pair);
     
        }
        else if ((parent_1 == -1 || parent_2 == -1)&&!runit.isCollapsed())
        {
           // CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_INFO, "printErrorNotCollapsed");

            
            
            
            ReactionPair reaction_pair;
            
            reaction_pair.addCollapsedReactionUnit(runit);
            reaction_pair.setCollapsed(false);
            reaction_pair.findSimilarity();
            reaction_group.addReactionPair(reaction_pair);
        }
    }
    
    reaction_group.print();
    //reaction_group.printXML();
}

}}}
