/*Copyright 2010-2012 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Protsenko (protsenko@ispras.ru)
// SYNOPSIS: 
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <vector>
#include <fstream>
#include <string>
#include <stdio.h>

#include <hw/component.hpp>
#include <hw/message.hpp>
#include <hw/diagnostics/reaction_pair.hpp>

namespace cpptesk {
namespace hw {
namespace diagnostics {

class ReactionGroup
{
public:
    const static int LINE_SIZE          = 80; // Maximum 80 characters in a line
    const static int TITLE_DATA_MESSAGE =  0; // print title
    const static int SPEC_DATA_MESSAGE  =  1; // print spec
    const static int IMPL_DATA_MESSAGE  =  2; // print impl
protected:
    std::vector<ReactionPair> _reaction_group;
    std::vector<std::string> _type_group;
    
public:
    ReactionGroup() {}
    ~ReactionGroup() {}

    bool isEmpty() { return _reaction_group.size() == 0; }
    int size() { return _reaction_group.size(); }
    void clear() { _reaction_group.clear(); }

    bool addNewFieldType(std::string field_type)
    {
        for(int i = 0; i < (int)(_type_group.size()); i++)
        {
            if(_type_group[i] == field_type)
            {
                return false;
            }
        }
        _type_group.push_back(field_type);
        return true;
    }
    
    ReactionPair& getReactionPair(int index)
    {
        return _reaction_group[index];
    }

    void addReactionPair(ReactionPair _reaction_pair)
    {
        _reaction_group.push_back(_reaction_pair);
    }
    
    // Printing N tags in line, before 'tag'
    void printTag(std::fstream &stream, char* tag, int shift)
    {
        for(int i = 0; i < shift; i++)
            stream << "\t";
        stream << tag << std::endl;
    }
  
    // Printing N tag in 'string_stream'
    void printTag(std::stringstream &string_stream, int shift)
    {
        for(int i = 0; i < shift; i++)
            string_stream << "\t";
    }
    
    void addSymbol(std::ostringstream &string_stream, char* symbol, int iterator)
    {
        for(int i = 0; i < iterator; i++)
            string_stream << symbol;
    }

    void addSymbolTop(std::string &str, char* symbol, int iterator)
    {
        for(int i = 0; i < iterator; i++)
            str.insert(0, symbol);
    }
    
    void addSymbol(std::string &str, char* symbol, int iterator)
    {
        for(int i = 0; i < iterator; i++)
            str.append(symbol);
    }

    void printTag(std::fstream &f_stream, int shift)
    {
        for(int i = 0; i < shift; i++)
            f_stream << "\t";
    }
 
    void printTag(std::fstream &stream, int num, int shift)
    {
        std::stringstream string;
        string << num;
        printTag(stream, (char*)(string.str().c_str()), shift);
    }
    
    void print()
    {
        int error_num = 0;
        for(int i = ReactionPair::DATA_TIME_IFACE_ERROR; i >= ReactionPair::NONE_ERROR; i-- )
        {
            _print_error_of_this_types(true, error_num, i, INCORRECT_REACTION, NORMAL_REACTION);
            _print_error_of_this_types(true, error_num, i, MISSING_REACTION);
            _print_error_of_this_types(true, error_num, i, UNEXPECTED_REACTION);

            _print_error_of_this_types(false, error_num, i, INCORRECT_REACTION, NORMAL_REACTION);
            _print_error_of_this_types(false, error_num, i, MISSING_REACTION);
            _print_error_of_this_types(false, error_num, i, UNEXPECTED_REACTION);
        }
    }
    
    void _print_error_group(int &error_num, ReactionPair example_pair, int group)
    {
        error_num ++;
        
        if(!example_pair.isCollapsed())
        {
            //CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_INFO, "isCollapsed");
            _print_error_not_collapsed(group, example_pair, error_num);   
        }
        else 
        {
            Reaction comp_reaction = example_pair.getCollapsedReactionUnit().getReaction();
            
            //CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_INFO, "MISSING_REACTION");
            //CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_INFO, "else");
            
            if(comp_reaction.getError() == MISSING_REACTION)
            {
                //CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_INFO, "MISSING_REACTION");
                print_error_collapsed_missing(group, example_pair, error_num);
            }
            else if(comp_reaction.getError() == UNEXPECTED_REACTION)
            {
                //CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_INFO, "UNEXPECTED_REACTION");
                _print_error_collapsed_unexpected(group, example_pair, error_num);
            }
            else // NORMAL or INCORRECT
            {
                //CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_INFO, "NORMAL or INCORRECT");
                _print_full_collapsed_incorrect(group, example_pair, error_num);
            }
        }
    }

    void _print_error_not_collapsed(int errors_type, ReactionPair example_pair, int error_num)
    {
         _print_separation_line();
        
        Reaction comp_reaction = example_pair.getCollapsedReactionUnit().getReaction();
        
        if(comp_reaction.getError() == MISSING_REACTION)
        {
            // Get size of table title
            int title_width = _get_title_width(error_num, comp_reaction);
             
            // Print the table title
            _print_title_error(title_width, error_num, errors_type, "Unknown", "Unknown", false);
            
            const ::cpptesk::hw::Message *message = comp_reaction.getSpecMessage();

            std::list <int> data_column_width = _get_data_full_fields_width(message);

            _print_line();
            _print_data_error_message(title_width, comp_reaction, data_column_width, TITLE_DATA_MESSAGE);
            _print_data_error_message(title_width, comp_reaction, data_column_width, SPEC_DATA_MESSAGE);

            const ::cpptesk::hw::Interface *iface = comp_reaction.where();
             
            _print_line();
            _print_time_error(title_width, comp_reaction, "Missing");
            _print_line();
            _print_iface_error(title_width, iface);
        }
        else if (comp_reaction.getError() == UNEXPECTED_REACTION)
        {
            // Get size of table title
            int title_width = _get_title_width(error_num, comp_reaction);
             
            const ::cpptesk::hw::Interface *iface = comp_reaction.where();
        
            // Print the table title
            _print_title_error(title_width, error_num, errors_type, iface, comp_reaction, true);
            
            const ::cpptesk::hw::Message *message = comp_reaction.getImplMessage();

            std::list <int> data_column_width = _get_data_full_fields_width(message);
            
            _print_line();
            _print_data_error_message(title_width, comp_reaction, data_column_width, TITLE_DATA_MESSAGE);
            _print_data_error_message(title_width, comp_reaction, data_column_width, IMPL_DATA_MESSAGE);
            
            _print_line();
            _print_time_error(title_width, comp_reaction, IMPL_DATA_MESSAGE);
            _print_line();
            _print_iface_error(title_width, iface, IMPL_DATA_MESSAGE); 
        } 
        else if ((comp_reaction.getError() == INCORRECT_REACTION) || (comp_reaction.getError() == NORMAL_REACTION))
        {
            _print_error_not_collapsed_incorrect(errors_type, example_pair, error_num);
        }

        _print_separation_line();
    }
    
    void _print_error_not_collapsed_incorrect(int errors_type, ReactionPair example_pair, int error_num)
    {
        Reaction comp_reaction = example_pair.getCollapsedReactionUnit().getReaction();
        
        // Get size of table title
        int title_width = _get_title_width(error_num, comp_reaction);

        const ::cpptesk::hw::Interface *iface = comp_reaction.where();
        
        // Print the table title
        _print_title_error(title_width, error_num, errors_type, iface, comp_reaction, true); // get_impl=true, if impl reaction not null
        
        if((ReactionPair::DATA_ERROR & errors_type) != 0)
        {
            _print_line();
            _print_incorrect_data_error( title_width, comp_reaction);
        }

        if((ReactionPair::TIME_ERROR & errors_type) != 0)
        {
            _print_line();
            _print_time_error(title_width, comp_reaction);
        }

        if((ReactionPair::IFACE_ERROR & errors_type) != 0)
        {
            _print_line();
            _print_iface_error(title_width, iface);
        }
    }
    
    void print_error_collapsed_missing(int errors_type, ReactionPair example_pair, int error_num)
    {
        _print_separation_line();
 
        Reaction comp_reaction = example_pair.getCollapsedReactionUnit().getReaction();

        // Get size of table title
        int title_width = _get_title_width(error_num, comp_reaction);
            
        // Print the table title
        _print_title_error(title_width, error_num, errors_type, "Unknown", "Unknown", false);
        
        const ::cpptesk::hw::Message *message = comp_reaction.getSpecMessage();

        std::list <int> data_column_width = _get_data_full_fields_width(message);

        _print_line();
        _print_data_error_message(title_width, comp_reaction, data_column_width, TITLE_DATA_MESSAGE);
        _print_data_error_message(title_width, comp_reaction, data_column_width, SPEC_DATA_MESSAGE);

        const ::cpptesk::hw::Interface *iface = comp_reaction.where();
            
        _print_line();
        _print_time_error(title_width, comp_reaction, "Missing");
        _print_line();
        _print_iface_error(title_width, iface);

        _print_separation_line();
    }

    void _print_error_collapsed_unexpected(int errors_type, ReactionPair example_pair, int error_num)
    {
        _print_separation_line();
 
        Reaction comp_reaction = example_pair.getCollapsedReactionUnit().getReaction();
        
        // Get size of table title
        int title_width = _get_title_width(error_num, comp_reaction);
            
        const ::cpptesk::hw::Interface *iface = comp_reaction.where();
    
        // Print the table title
        _print_title_error(title_width, error_num, errors_type, iface, comp_reaction, true);
        
        const ::cpptesk::hw::Message *message = comp_reaction.getImplMessage();

        std::list <int> data_column_width = _get_data_full_fields_width(message);
        
        _print_line();
        _print_data_error_message(title_width, comp_reaction, data_column_width, TITLE_DATA_MESSAGE);
        _print_data_error_message(title_width, comp_reaction, data_column_width, IMPL_DATA_MESSAGE);
        
        _print_line();
        _print_time_error(title_width, comp_reaction, IMPL_DATA_MESSAGE);
        _print_line();
        _print_iface_error(title_width, iface, IMPL_DATA_MESSAGE); 
        
        _print_separation_line();
    }

    std::list <int> _get_data_full_fields_width(const ::cpptesk::hw::Message *message)
    // Return width of columns for all fields
    // For Missing & Unexpected reactions
    {
         std::list <int> fieid_size;
        
        for(::cpptesk::hw::Message::const_iterator i_message = message->begin();
            i_message != message->end(); 
            i_message++)
        {
            const Field<> &field_message = (*message).**i_message;
 
                std::string str_message;
                std::string str_name;
                    
                std::stringstream buffer;
                buffer << std::hex << field_message.get();
                buffer >> str_message;
                
                str_name = std::string(field_message.getShortName());
            
                int len_message = str_message.length();
                int len_name = str_name.length();
                
                if(len_message > len_name) fieid_size.push_back(len_message);
                else fieid_size.push_back(len_name);
        }
        
        return fieid_size;
    }

    std::list <int> _get_data_error_fields_width(Reaction comp_reaction)
    // Return width of columns for error fields
    // For Normal & Incorrect reactions
    {
        const ::cpptesk::hw::Message *collapsed_impl = comp_reaction.getImplMessage();
        const ::cpptesk::hw::Message *collapsed_spec = comp_reaction.getSpecMessage();

        std::list <int> fieid_size;
        
        for(::cpptesk::hw::Message::const_iterator i_message_spec = collapsed_spec->begin(), 
            j_message_impl = collapsed_impl->begin(); i_message_spec != collapsed_spec->end(); 
            i_message_spec++, j_message_impl++)
        {
            const Field<> &field_spec = (*collapsed_spec).**i_message_spec;
            const Field<> &field_impl = (*collapsed_impl).**j_message_impl;

            if(!field_spec.equals(field_impl))
            {
                std::string str_impl;
                std::string str_spec;
                std::string str_name;
                    
                std::stringstream buffer;
                buffer << std::hex << field_impl.get();
                buffer >> str_impl;
                
                buffer.clear();
                buffer << std::hex << field_spec.get();
                buffer >> str_spec;
                
                str_name = std::string(field_spec.getShortName());
            
                int len_impl = str_impl.length();
                int len_spec = str_spec.length();
                int len_name = str_name.length();
                
                if(len_impl > len_spec)
                {
                    if(len_impl > len_name) fieid_size.push_back(len_impl);
                    else fieid_size.push_back(len_name);
                }
                else
                    if(len_spec > len_name) fieid_size.push_back(len_spec);
                    else fieid_size.push_back(len_name);
            }
        }
        
        return fieid_size;
    }

    void _print_data_error_message(int title_width, Reaction comp_reaction, std::list <int> data_column_width, int data_type)
    // Print columns title for error fields
    {
        const ::cpptesk::hw::Message *message;
        
        if(comp_reaction.getError() == MISSING_REACTION)
        {
            message = comp_reaction.getSpecMessage();

        }else if (comp_reaction.getError() == UNEXPECTED_REACTION)
        {
            message = comp_reaction.getImplMessage();            
        }
        
        std::string msg_title;
        if(data_type == IMPL_DATA_MESSAGE)
        {
            msg_title = "Excepted";
        }
        else if (data_type == SPEC_DATA_MESSAGE)
        {
            msg_title = "Received";
        }
        else 
        {
            msg_title = message->getName();            
        }

        msg_title.insert(0, "| ", 0, 2);
        addSymbol(msg_title, (char*)" ", title_width - msg_title.length());
        msg_title.append(" | ");

        CPPTESK_DEBUG_STREAM() << msg_title;
        
        int line_lenght = 0;
        
        std::list<int>::iterator i_size = data_column_width.begin();
        for(::cpptesk::hw::Message::const_iterator i_message = message->begin();
            i_message != message->end(); 
            i_message++)
        {
            const Field<> &field = (*message).**i_message;

            int msg_size = *i_size;
            
            std::string str_name;
            
            if((data_type == IMPL_DATA_MESSAGE) || (data_type == SPEC_DATA_MESSAGE))
            {
                std::stringstream buffer;
                buffer << std::hex << field.get();
                buffer >> str_name;
            }
            else 
            {
                str_name = std::string(field.getShortName());
            }
            
            addSymbolTop(str_name, (char*)" ", msg_size - str_name.length() + 1); // + " " in top

            line_lenght = line_lenght + str_name.length() + 2; // 2 = " |"

            if((int)(msg_title.length() + line_lenght) > LINE_SIZE) break;

            CPPTESK_DEBUG_STREAM() << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL) << str_name << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL) << " |";

            i_size ++;
        }
        
        std::string str_name;
        str_name.clear();
        addSymbol(str_name, (char*)" ", (LINE_SIZE - msg_title.length() - line_lenght - 2)); // 2 = " |"
        str_name.append(" |");
        
        CPPTESK_DEBUG_STREAM() << str_name << std::endl;
    }
    
    void printXML()
    {
        std::fstream diag_xml;
        diag_xml.open("failures.xml", std::ios::out);
        diag_xml << "<pair_list>" << std::endl; // open main failure tag
        
        std::fstream field_xml;
        field_xml.open("fields.xml", std::ios::out);
        field_xml << "<msg_list>" << std::endl;

        for(int i = ReactionPair::DATA_TIME_IFACE_ERROR; i >= ReactionPair::IFACE_ERROR; i-- )
        {
            diag_xml << "\t" << "<type id=\"" << i <<"\">" << std::endl; // One type of "ReactionPair" errors

            int errors_size = 0;
            for(int j = 0; j < this->size(); j++ )
            {
                if(_reaction_group[j].getGroup() == i)
                {
                    errors_size ++;
                    
                    diag_xml << "\t" << "\t" << "<reaction_pair num=\"" << j <<"\">" << std::endl;

                    // Reactions
                    ReactionPair error_pair = _reaction_group[j];

                    // ***=====DATA=====***
                    Reaction collapsed_reaction = error_pair.getCollapsedReactionUnit().getReaction();
                    printTag(diag_xml, (char*)"<data>", 3);

                    
                    const ::cpptesk::hw::Message *collapsed_spec = collapsed_reaction.getSpecMessage();
                    const ::cpptesk::hw::Message *collapsed_impl = collapsed_reaction.getImplMessage();
    
                    std::stringstream string_field_impl;
                    std::stringstream string_field_spec;
                    
                    bool new_type = false;
                    
                    new_type = addNewFieldType(collapsed_impl->getName());                    
                    
                    if(new_type)                    
                    {
                        field_xml << "\t" << "<msg type=\"" << _type_group.back() << "\">" << std::endl;
                    }
 
                    printTag(string_field_impl, 4);
                    printTag(string_field_spec, 4);
                    string_field_impl << "<impl type=\"" << collapsed_impl->getName() << "\">" << std::endl;
                    string_field_spec << "<spec type=\"" << collapsed_spec->getName() << "\">" << std::endl;
                   
                    for(::cpptesk::hw::Message::const_iterator i_message_spec = collapsed_spec->begin(), 
                        j_message_impl = collapsed_impl->begin(); i_message_spec != collapsed_spec->end(); i_message_spec++, j_message_impl++)
                    {
                        const Field<> &field_spec = (*collapsed_spec).**i_message_spec;
                        const Field<> &field_impl = (*collapsed_impl).**j_message_impl;

                        printTag(string_field_impl, 5);
                        printTag(string_field_spec, 5);
                        
                        string_field_impl << "<field name=\"";
                        string_field_spec << "<field name=\"";

                        string_field_impl << field_spec.getShortName() << "\" value=\"0x" << std::hex << field_spec.get() << "\"";
                        string_field_spec << field_impl.getShortName() << "\" value=\"0x" << std::hex << field_impl.get() << "\"";

                        string_field_impl << "/>" << std::endl;
                        string_field_spec << "/>" << std::endl;

                        if(new_type)                    
                        {
                            field_xml << "\t" << "\t" << "<field name=\"" << field_spec.getShortName() << "\"/>" << std::endl;
                        }
                    }

                    if(new_type)                    
                    {
                        field_xml << "\t" << "</msg>" << std::endl;
                    }
 
                    diag_xml << string_field_impl.str().c_str(); //tab=5
                    printTag(diag_xml, (char*)"</impl>", 4);
                    
                    diag_xml << string_field_spec.str().c_str(); //tab=5
                    printTag(diag_xml, (char*)"</spec>", 4);
                    
                    printTag(diag_xml, (char*)"</data>", 3);
                   
                    // ***=====TIME=====***
                    
                    ReactionUnit specReactionUnit;
                    ReactionUnit implReactionUnit;
                    error_pair.getSpecReactionUnit(specReactionUnit);
                    error_pair.getImplReactionUnit(implReactionUnit);

                
 
                    Reaction impl_reaction = implReactionUnit.getReaction();
                    Reaction spec_reaction = specReactionUnit.getReaction();

                    printTag(diag_xml, (char*)"<time>", 3);

                    // Impl time
                    printTag(diag_xml, 4);
                    diag_xml << "<impl started=\"" << impl_reaction.whenStarted() << "\"" << " stopped=\"" << impl_reaction.whenStopped() << "\"" <<"/>" << std::endl;
                    
                    // Spec time
                    printTag(diag_xml, 4);
                    diag_xml << "<spec started=\"" << spec_reaction.whenStarted() << "\"" << " stopped=\"" << spec_reaction.whenStopped() << "\"" <<"/>" << std::endl;

                    printTag(diag_xml, (char*)"</time>", 3);
                     
                    // ***=====IFACE=====***
                    printTag(diag_xml, (char*)"<iface>", 3);

                    const ::cpptesk::hw::Interface *iface_impl = impl_reaction.where();
                    const ::cpptesk::hw::Interface *iface_spec = spec_reaction.where();
                    printTag(diag_xml, 4);
                    diag_xml << "<impl name=\"" << iface_impl << "\"" <<"/>" << std::endl;

                    printTag(diag_xml, 4);
                    diag_xml << "<spec name=\"" << iface_spec << "\"" <<"/>" << std::endl;

                    printTag(diag_xml, (char*)"</iface>", 3);                    
                    // end <reaction pair>
                    printTag(diag_xml, 2);
                    diag_xml << "</reaction_pair>" << std::endl;                    
                }
            }
            diag_xml << "\t" << "</type>" << std::endl;
        }
        field_xml << "</msg_list>" << std::endl;
        field_xml.close();

        diag_xml << "</pair_list>" << std::endl; // close main failure tag
        diag_xml.close();
    }
    
    protected: 

    void _print_error_of_this_types(bool collapsed, int &error_num, int iterator, int error_type_one)
    // Main
    {
        _print_error_of_this_types(collapsed, error_num, iterator, error_type_one, error_type_one);
    }

    
    void _print_error_of_this_types(bool collapsed, int &error_num, int iterator, int error_type_one, int error_type_two)
    // Main
    {
        ReactionPair example_pair;
        int errors_size = 0;
        bool example = false;

        for(int j = 0; j < this->size(); j++ )
        {
            Reaction reaction = _reaction_group[j].getCollapsedReactionUnit().getReaction();
            
            if((_reaction_group[j].getGroup() == iterator) && (_reaction_group[j].isCollapsed() == collapsed) &&
                ((reaction.getError() == error_type_one) || (reaction.getError() == error_type_two)))
            {
                errors_size ++;
                if(!example)
                {
                    example = true;
                    example_pair = _reaction_group[j];
                }
            }
        }
        
        if(errors_size > 0)
        {
            _print_error_group(error_num, example_pair, iterator);
        }
    }
    
    void _print_full_collapsed_incorrect(int errors_type, ReactionPair example_pair, int error_num)
    // Branch #1
    // Print diagnostic for collapsed NORMAL & INCORRECT reactions
    {
        _print_separation_line();
        
        ReactionUnit specReactionUnit;
        ReactionUnit implReactionUnit;
        
        Reaction comp_reaction = example_pair.getCollapsedReactionUnit().getReaction();
        
        // Get size of table title
        int title_width = _get_title_width(error_num, comp_reaction);

        example_pair.getSpecReactionUnit(specReactionUnit);
        bool get_impl = example_pair.getImplReactionUnit(implReactionUnit);

        Reaction spec_reaction = specReactionUnit.getReaction();
        Reaction impl_reaction = implReactionUnit.getReaction();
        
        const ::cpptesk::hw::Interface *iface_spec = spec_reaction.where();
        const ::cpptesk::hw::Interface *iface_impl = impl_reaction.where();
        
        // Print the table title
        _print_title_error(title_width, error_num, errors_type, iface_impl, impl_reaction, get_impl); // get_impl=true, if impl reaction not null
        
        if((ReactionPair::DATA_ERROR & errors_type) != 0)
        {
            _print_line();
            _print_incorrect_data_error( title_width, comp_reaction);
        }

        if((ReactionPair::TIME_ERROR & errors_type) != 0)
        {
            _print_line();
            Reaction impl_reaction = implReactionUnit.getReaction();

            _print_time_error(title_width, spec_reaction);
        }

        if((ReactionPair::IFACE_ERROR & errors_type) != 0)
        {
            _print_line();
            _print_iface_error(title_width, iface_spec);
        }

        _print_separation_line();       
    }
    
    void _print_incorrect_data_error(int title_width, Reaction comp_reaction)
    // Branch #1
    // Print DATA for collapsed NORMAL & INCORRECT reactions
    {
        std::list <int> data_column_width = _get_data_error_fields_width(comp_reaction);
        
        _print_data_error(title_width, comp_reaction, data_column_width, TITLE_DATA_MESSAGE);
        _print_data_error(title_width, comp_reaction, data_column_width, IMPL_DATA_MESSAGE);
        _print_data_error(title_width, comp_reaction, data_column_width, SPEC_DATA_MESSAGE);
    }

    void _print_data_error(int title_width, Reaction comp_reaction, std::list <int> data_column_width, int data_type)
    // Branch #1
    // Print columns & title for error fields 
    // from  DATA collapsed NORMAL & INCORRECT reactions
    {
        const ::cpptesk::hw::Message *collapsed_impl = comp_reaction.getImplMessage();
        const ::cpptesk::hw::Message *collapsed_spec = comp_reaction.getSpecMessage();
        
        std::string msg_title;
        if(data_type == IMPL_DATA_MESSAGE)
        {
            msg_title = "Excepted";
        }
        else if (data_type == SPEC_DATA_MESSAGE)
        {
            msg_title = "Received";
        }
        else 
        {
            msg_title = collapsed_impl->getName();            
        }

        msg_title.insert(0, "| ", 0, 2);
        addSymbol(msg_title, (char*)" ", title_width - msg_title.length());
        msg_title.append(" | ");

        CPPTESK_DEBUG_STREAM() << msg_title;
        
        int line_lenght = 0;
        
        std::list<int>::iterator i_size = data_column_width.begin();
        for(::cpptesk::hw::Message::const_iterator i_message_spec = collapsed_spec->begin(), 
            j_message_impl = collapsed_impl->begin(); i_message_spec != collapsed_spec->end(); 
            i_message_spec++, j_message_impl++)
        {
            const Field<> &field_spec = (*collapsed_spec).**i_message_spec;
            const Field<> &field_impl = (*collapsed_impl).**j_message_impl;

            if(!field_spec.equals(field_impl))
            {
                int msg_size = *i_size;
                
                std::string str_name;
                
                if(data_type == IMPL_DATA_MESSAGE)
                {
                    std::stringstream buffer;
                    buffer << std::hex << field_impl.get();
                    buffer >> str_name;
                }
                else if (data_type == SPEC_DATA_MESSAGE)
                {
                    std::stringstream buffer;
                    buffer << std::hex << field_spec.get();
                    buffer >> str_name;
                }
                else 
                {
                    str_name = std::string(field_spec.getShortName());
                }
                
                addSymbolTop(str_name, (char*)" ", msg_size - str_name.length() + 1); // + " " in top

                line_lenght = line_lenght + str_name.length() + 2; // 2 = " |"

                if((int)(msg_title.length() + line_lenght) > LINE_SIZE) break;

                if(data_type == IMPL_DATA_MESSAGE)
                    CPPTESK_DEBUG_STREAM() << CPPTESK_DEBUG_STYLE(RED, NORMAL) << str_name << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL) << " |";
                else if(data_type == SPEC_DATA_MESSAGE)
                    CPPTESK_DEBUG_STREAM() << CPPTESK_DEBUG_STYLE(GREEN, NORMAL) << str_name << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL) << " |";
                else
                    CPPTESK_DEBUG_STREAM() << str_name << " |";

                i_size ++;
            }
        }
        
        std::string str_name;
        str_name.clear();
        addSymbol(str_name, (char*)" ", (LINE_SIZE - msg_title.length() - line_lenght - 2)); // 2 = " |"
        str_name.append(" |");
        
        CPPTESK_DEBUG_STREAM() << str_name << std::endl;
    }

    void _print_time_error(int shift_size, Reaction impl_reaction)
    // Branch #1
    {        
        _print_time_error( shift_size, impl_reaction, "Timing");
    }
        
    void _print_time_error(int shift_size, Reaction impl_reaction, std::string time_type)
    // Branch #1
    {        
        std::string str_time;
        str_time.append("| ");
        str_time.append(time_type);
        addSymbol(str_time, (char*)" ", shift_size - str_time.length());
        
        CPPTESK_DEBUG_STREAM() << str_time << " | expected at ["  << CPPTESK_DEBUG_STYLE(GREEN, NORMAL) << std::dec << impl_reaction.whenStarted() << ".." 
        << impl_reaction.whenStopped() << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL) << "]";

        // Get length the string
        str_time.append(" | expected at [..]");
        int lenght = str_time.length();
        str_time.clear();
        std::stringstream o_str;
        o_str << impl_reaction.whenStarted();
        o_str >> str_time;
        lenght = lenght + str_time.length();
        o_str.clear();
        str_time.clear();
        o_str << impl_reaction.whenStopped();
        o_str >> str_time;
        lenght = lenght + str_time.length();

        // Add to the end "|"  
        std::string str_t;
        addSymbol(str_t, (char*)" ", LINE_SIZE - lenght - 1);
        str_t.append("|");
        CPPTESK_DEBUG_STREAM() << str_t << std::endl;
    }
        
    void _print_iface_error(int shift_size, const ::cpptesk::hw::Interface *iface)
    // Branch #1
    {
        std::string str;
        str.append("| ");
        str.append("Interface");
        addSymbol(str, (char*)" ", shift_size - str.length());
        str.append(" | expected on [");
        CPPTESK_DEBUG_STREAM() << str;
        int lenght = str.length();
        str.clear();
        std::stringstream o_str;
        o_str << ((Component*)iface)->getName();
        o_str >> str;
        lenght = lenght + str.length();
        str.clear();
        CPPTESK_DEBUG_STREAM() << CPPTESK_DEBUG_STYLE(GREEN, NORMAL) << ((Component*)iface)->getName() << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL);
        str.append("]");  // lenght - 1      
        addSymbol(str, (char*)" ", LINE_SIZE - lenght - 3); // 1 characters = "]" and 2 characters = " |"
        CPPTESK_DEBUG_STREAM() << str << " |" << std::endl;
    }
    
    void _print_iface_error(int shift_size, const ::cpptesk::hw::Interface *iface, int type)
    // Branch #1
    {
        if(type == IMPL_DATA_MESSAGE)
        {
            std::string str;
            str.append("| ");
            str.append("Interface");
            addSymbol(str, (char*)" ", shift_size - str.length());
            str.append(" | ");
            CPPTESK_DEBUG_STREAM() << str;
            
            str.append("unexpected");
            int lenght = str.length();
            str.clear();

            CPPTESK_DEBUG_STREAM() << CPPTESK_DEBUG_STYLE(RED, NORMAL) << "unexpected" << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL);
            addSymbol(str, (char*)" ", LINE_SIZE - lenght - 2); // 1 characters = "]" and 2 characters = " |"
            CPPTESK_DEBUG_STREAM() << str << " |" << std::endl;
        }
        else
            _print_iface_error( shift_size, iface);
    }

    void _print_time_error(int shift_size, Reaction impl_reaction, int type)
    {
        if(type == IMPL_DATA_MESSAGE)
        {
            std::string str_time;
            str_time.append("| Timing");
            addSymbol(str_time, (char*)" ", shift_size - str_time.length());
            
          /*  CPPTESK_DEBUG_STREAM() << str_time << " | expected at ["  << CPPTESK_DEBUG_STYLE(GREEN, NORMAL) 
            << "Unknown" << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL) << "]";

            // Get length the string
            str_time.append(" | expected at [Unknown]");
            int lenght = str_time.length();
            str_time.clear();*/
          
            CPPTESK_DEBUG_STREAM() << str_time << " | " << CPPTESK_DEBUG_STYLE(RED, NORMAL) << 
            "unexpected"  << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL);

            // Get length the string
            str_time.append(" | unexpected");
            int lenght = str_time.length();
            str_time.clear();

            // Add to the end "|"  
            std::string str_t;
            addSymbol(str_t, (char*)" ", LINE_SIZE - lenght - 1);
            str_t.append("|");
            CPPTESK_DEBUG_STREAM() << str_t << std::endl;
        }
        else
            _print_time_error( shift_size, impl_reaction);
    }

    void _print_title_error(int title_width, int error_num, int errors_type, const ::cpptesk::hw::Interface *iface_impl, Reaction &impl_reaction, bool impl)
    {
        std::stringstream o_str;
        std::string str_iface_name;
        std::string str_cycle;
        
        if(impl)
        {
            o_str << ((Component*)iface_impl)->getName();
            o_str >> str_iface_name;
            o_str.clear();
            o_str << std::dec << impl_reaction.whenStarted();
            o_str >> str_cycle;
            _print_title_error(title_width, error_num, errors_type, str_iface_name, str_cycle, impl); // impl=true, if impl reaction not null
        }
        else
        {
            _print_title_error(title_width, error_num, errors_type, "Unknown", "Unknown", impl);
        }
    }

    void _print_title_error(int title_width, int error_num, int errors_type, std::string str_iface_name, std::string str_cycle, bool impl)
    // Print the table header. 
    // Format: | Failure: #error_num | [iface_impl] at cycle X | 
    {
        CPPTESK_DEBUG_STREAM() << "| " << CPPTESK_DEBUG_STYLE(RED, NORMAL) << "Failure: #" << std::dec << error_num 
        << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL);
        
        std::stringstream o_str;
        o_str << error_num;
        std::string str_title;
        o_str >> str_title;
        str_title.append("| Failure: #");
        std::string str_temp;
        addSymbol(str_temp, (char*)" ", title_width - str_title.length());
        
        CPPTESK_DEBUG_STREAM() << str_temp << " | [";
        
        o_str.clear();
        str_title.clear();
        
        if(impl)
        {
            if((ReactionPair::IFACE_ERROR & errors_type) != 0)
                CPPTESK_DEBUG_STREAM() << CPPTESK_DEBUG_STYLE(RED, NORMAL) << str_iface_name << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL);
            else 
                CPPTESK_DEBUG_STREAM() << str_iface_name;
            
            CPPTESK_DEBUG_STREAM() << "] at cycle "; 
            
            if((ReactionPair::TIME_ERROR & errors_type) != 0)
                CPPTESK_DEBUG_STREAM() << CPPTESK_DEBUG_STYLE(RED, NORMAL) << str_cycle << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL);
            else 
                CPPTESK_DEBUG_STREAM() << str_cycle;
        }
        else
        {
            CPPTESK_DEBUG_STREAM() << CPPTESK_DEBUG_STYLE(RED, NORMAL) << str_iface_name << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL);
            CPPTESK_DEBUG_STREAM() << "] at cycle "; 
            CPPTESK_DEBUG_STREAM() << CPPTESK_DEBUG_STYLE(RED, NORMAL) << str_cycle << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL);
        }
        
        str_title.append(" | [] at cycle ");
        
        int lenght = str_title.length();
        str_title.clear();
        o_str << str_iface_name << str_cycle;
        o_str >> str_title;
        
    //    CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_INFO, std::endl << lenght << "  " << str_title.length() << "  " << title_width << std::endl);
        
        lenght = lenght + str_title.length() + title_width;

    //    CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_INFO, std::endl << (LINE_SIZE - lenght) << std::endl);

        // Add to the end "|"  
        std::string str_t;
        addSymbol(str_t, (char*)" ", LINE_SIZE - lenght - 1);
        str_t.append("|");
        
        CPPTESK_DEBUG_STREAM() << str_t << std::endl;
    }

    int _get_title_width(int error_num, Reaction comp_reaction) 
    // Get the width of the first column
    // to the next "|"
    {
        int lenght_msg = 0;
        std::string str_msg;
        
        if(comp_reaction.getError() != UNEXPECTED_REACTION)
        {
            const ::cpptesk::hw::Message *message = comp_reaction.getSpecMessage();
            str_msg = std::string(message->getName());
        }
        else
        {
            const ::cpptesk::hw::Message *message = comp_reaction.getImplMessage();
            str_msg = std::string(message->getName());
        }

        lenght_msg = str_msg.length() + 2; // 2 = "| "

        std::stringstream failure;
        failure << "| Failure: #" << error_num;
        std::string str_failure;
        str_failure.append(failure.str());

        int title_width = 0;
        
        if((int)str_failure.length() > lenght_msg)
            title_width = str_failure.length();
        else
            title_width = lenght_msg;

        return title_width;
    }
    
    void _print_line(int length)
    {
        std::string str_line;
        addSymbol(str_line, (char*)"-", length);            
        CPPTESK_DEBUG_STREAM() << str_line << std::endl;
    }

    void _print_line()
    {
        _print_line(LINE_SIZE);    
    }

    void _print_separation_line(int length)
    {
        std::string str_big_line;
        addSymbol(str_big_line, (char*)"=", length);            
        CPPTESK_DEBUG_STREAM() << str_big_line << std::endl;
    }

    void _print_separation_line()
    {
         _print_separation_line(LINE_SIZE);            
    }
};

}}}
