/*Copyright 2010-2012 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Protsenko (protsenko@ispras.ru)
// SYNOPSIS: ReactionUnit class keeps Reaction field and its predecessors and 
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/event.hpp>
#include <hw/debug.hpp>

namespace cpptesk {
namespace hw {
namespace diagnostics {

class ReactionUnit
{
public:
    const static int COLLAPSE_TYPE_FUZZY  = 2;
    const static int COLLAPSE_TYPE_NORMAL = 1;
    const static int COLLAPSE_TYPE_NONE   = 0;

    ReactionUnit()
    {
    }
    
    ReactionUnit(const ReactionUnit &reaction_unit): _reaction(reaction_unit._reaction)
    {
        this->_parentSpec      = reaction_unit._parentSpec;
        this->_parentImpl      = reaction_unit._parentImpl;
        this->_parentReaction2 = reaction_unit._parentReaction2;
        this->_parentReaction1 = reaction_unit._parentReaction1;        
        
        this->_collapsed    = reaction_unit._collapsed;
        this->_myIdIs       = reaction_unit._myIdIs;
        this->_collapseType = reaction_unit._collapseType;
        this->_collapseRule = reaction_unit._collapseRule;
    }
    
    ReactionUnit(const ::cpptesk::hw::Reaction &_r) : _reaction(_r)
    {
        _parentReaction1 = -1;
        _parentReaction2 = -1;
        _parentImpl = -1;
        _parentSpec = -1;
        _collapsed = false;
        _collapseRule = -1;
        _myIdIs = -1;
        setCollapseType(COLLAPSE_TYPE_NONE);
    }

    ReactionUnit(::cpptesk::hw::Reaction &_r, int parentSpec, int parentImpl) : _reaction(_r)
    {
        _parentSpec = parentSpec;
        _parentImpl = parentImpl;
        _parentReaction2 = parentSpec;
        _parentReaction1 = parentImpl;        
        _collapsed = false;
        _collapseRule = -1;
        _myIdIs = -1;
        setCollapseType(COLLAPSE_TYPE_NONE);
    }

    virtual ~ReactionUnit() {}

    void setCollapsed() { _collapsed = true; }
    bool isCollapsed() { return _collapsed; }

    void setCollapseRule(int rule) { _collapseRule = rule; }
    int getCollapseRule() { return _collapseRule; }
    
    void setCollapseType(int collapseType)
    {
        _collapseType = collapseType;
    }

    int getCollapseType()
    {
        return _collapseType;
    }

    void setIdentifier(int id)
    {
        _myIdIs = id;
    }
    
    int getIdentifier()
    {
        return _myIdIs;
    }
    void setParentSpec(int parent_id)
    {
        _parentSpec = parent_id;
    }

    int getParentSpec()
    {
        return _parentSpec;
    }

    void setParentImpl(int parent_id)
    {
        _parentImpl = parent_id;
    }

    int getParentImpl()
    {
        return _parentImpl;
    }
 
    void setParents(int parent1_id, int parent2_id)
    {
        _parentReaction1 = parent1_id;
        _parentReaction2 = parent2_id;
    }

    void getParents(int *parent1_id, int *parent2_id)
    {
        *parent1_id = _parentReaction1;
        *parent2_id = _parentReaction2;
    }
    
    void printDescription(cpptesk::hw::debug_level_t debug_level, debug_color_t fg_level, debug_color_t bg_level)
    {
        if(_reaction.getError() == MISSING_REACTION)
        {
            CPPTESK_COLORED_DEBUG_PRINT(debug_level, fg_level, bg_level, 
                _reaction.getShortDescription() << std::endl <<
                " SpecMessage: " << _reaction.getSpecMessage()->forceToString());
        }
        else if (_reaction.getError() == UNEXPECTED_REACTION)
        {
            CPPTESK_COLORED_DEBUG_PRINT(debug_level, fg_level, bg_level, 
                _reaction.getShortDescription() << std::endl <<
                " ImplMessage: " << _reaction.getImplMessage()->forceToString());
        }
        else
        {
            CPPTESK_COLORED_DEBUG_PRINT(debug_level, fg_level, bg_level, 
                _reaction.getShortDescription() << std::endl <<
                " SpecMessage: " << _reaction.getSpecMessage()->forceToString() << std::endl <<
                " ImplMessage: " << _reaction.getImplMessage()->forceToString());
        }
    }
 
    void printDescription(cpptesk::hw::debug_level_t debug_level)
    {
        printDescription(debug_level, ::cpptesk::hw::DEFAULT, ::cpptesk::hw::DEFAULT);
    }

    void printDescription()
    {
        printDescription(cpptesk::hw::DEBUG_BASIC, ::cpptesk::hw::DEFAULT, ::cpptesk::hw::DEFAULT);
    }

    void printShortDescription(cpptesk::hw::debug_level_t debug_level, debug_color_t fg_level, debug_color_t bg_level, bool spec)
    {
        if(spec)
        {
            CPPTESK_COLORED_DEBUG_PRINT(debug_level, fg_level, bg_level, 
                 _reaction.getSpecMessage()->forceToString());
        }
        else if (!spec)
        {
            CPPTESK_COLORED_DEBUG_PRINT(debug_level, fg_level, bg_level, 
                 _reaction.getImplMessage()->forceToString());
        }
    }
 
    void printShortDescription(cpptesk::hw::debug_level_t debug_level)
    {
        printDescription(debug_level, ::cpptesk::hw::DEFAULT, ::cpptesk::hw::DEFAULT);
    }

::cpptesk::hw::Reaction& getReaction() { return _reaction; }

protected:
    bool _collapsed;
    int  _parentSpec, _parentImpl;
    int  _myIdIs;
    int  _parentReaction1, _parentReaction2;
    int  _collapseRule;
    int  _collapseType;
    ::cpptesk::hw::Reaction _reaction;
};

}}}
