/*Copyright 2010-2012 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Protsenko (protsenko@ispras.ru)
// SYNOPSIS: Impl Intersection Elimination 
//           (Rule # 6 in C++TESK.ErrorDiagnostics.ru.doc):
//           {[a:b]+[x:a] => [a:a]+[x:b]}
///////////////////////////////////////////////////////////////////////////////

#include <hw/diagnostics/rule/impl_intersection.hpp>

namespace cpptesk {
namespace hw {
namespace diagnostics {

    ReactionSet ImplIntersectionEliminationRule::transform(::cpptesk::hw::diagnostics::ReactionUnit &reaction_unit_1, 
                                                           ::cpptesk::hw::diagnostics::ReactionUnit &reaction_unit_2)    
    // Collapse the reactions
    {
        ReactionSet transform_out;
        
        //Message from reaction 1
        const ::cpptesk::hw::Message *message_1_spec,
                                     *message_1_impl;
        //Message from reaction 2
        const ::cpptesk::hw::Message *message_2_impl,
                                     *message_2_spec;

        const ::cpptesk::hw::Interface *iface_1;
        const ::cpptesk::hw::Interface *iface_2;
    
        // Get "start" and "end" time of reactions
        cycle_t start_time   =  reaction_unit_1.getReaction().whenStarted();
        cycle_t start_time_2 =  reaction_unit_2.getReaction().whenStarted();
        cycle_t end_time     =  reaction_unit_1.getReaction().whenStopped();
        cycle_t end_time_2   =  reaction_unit_2.getReaction().whenStopped();
        
        if(start_time > start_time_2) start_time = start_time_2; // Min start time
        if(end_time < end_time_2)     end_time = end_time_2;     // Max end time
        
        if((reaction_unit_1.getReaction().getError() == INCORRECT_REACTION)&&(reaction_unit_2.getReaction().getError() == UNEXPECTED_REACTION))
        {
            message_1_spec = reaction_unit_1.getReaction().getSpecMessage();
            message_2_impl = reaction_unit_2.getReaction().getImplMessage();

            //_message2Spec = _reactionUnit_2.getReaction().getSpecMessage();
            message_1_impl = reaction_unit_1.getReaction().getImplMessage();

            iface_1 = reaction_unit_1.getReaction().where();
            iface_2 = reaction_unit_2.getReaction().where();

            ::cpptesk::hw::Reaction reaction_unexpected = Reaction(UNEXPECTED_REACTION, *iface_2, *message_1_impl, 
                                                                  start_time, end_time, "unexpected reaction from rule#6");
            ::cpptesk::hw::Reaction reaction_normal =     Reaction(NORMAL_REACTION,     *iface_1, *message_1_spec, *message_2_impl, 
                                                                  start_time, end_time, "normal reaction from rule#6");
            
            transform_out.addReaction(reaction_unexpected,                              -1, reaction_unit_1.getIdentifier());
            transform_out.addReaction(reaction_normal,     reaction_unit_1.getIdentifier(), reaction_unit_2.getIdentifier());
        }
        else
        {
            //_message1Spec = _reactionUnit_2.getReaction().getSpecMessage();
            message_2_impl = reaction_unit_2.getReaction().getImplMessage();

            message_2_spec = reaction_unit_2.getReaction().getSpecMessage();
            message_1_impl = reaction_unit_1.getReaction().getImplMessage();
            
            iface_1 = reaction_unit_2.getReaction().where();
            iface_2 = reaction_unit_1.getReaction().where();


            ::cpptesk::hw::Reaction reaction_unexpected = Reaction(UNEXPECTED_REACTION, *iface_2, *message_2_impl, 
                                                                  start_time, end_time, "unexpected reaction from rule#6");
            ::cpptesk::hw::Reaction reaction_normal =     Reaction(NORMAL_REACTION,     *iface_1, *message_2_spec, *message_1_impl, 
                                                                  start_time, end_time, "normal reaction from rule#6");
            
            transform_out.addReaction(reaction_unexpected,                              -1, reaction_unit_2.getIdentifier());
            transform_out.addReaction(reaction_normal,     reaction_unit_2.getIdentifier(), reaction_unit_1.getIdentifier());
        }
        
        // Set flag
        reaction_unit_1.setCollapsed();
        reaction_unit_2.setCollapsed();
       
        return transform_out;
    }

    int ImplIntersectionEliminationRule::match(::cpptesk::hw::Reaction reaction_1, 
                                               ::cpptesk::hw::Reaction reaction_2)
    // Compare the reactions
    {
        int match = FALSE_MATCH;

        if((reaction_1.getError() == INCORRECT_REACTION)&&(reaction_2.getError() == UNEXPECTED_REACTION))
        {
            // Message from reaction 1
            const ::cpptesk::hw::Message *message_1_spec = reaction_1.getSpecMessage();
             // Message from reaction 2
            const ::cpptesk::hw::Message *message_2_impl = reaction_2.getImplMessage();
            
            if(message_1_spec->size() == message_2_impl->size())
            {
                match = EXACT_MATCH;

                for(::cpptesk::hw::Message::const_iterator i = message_1_spec->begin(), 
                    j = message_2_impl->begin(); i != message_1_spec->end(); i++, j++)
                {
                    const Field<> &field_spec = (*message_1_spec).**i;
                    const Field<> &field_impl = (*message_2_impl).**j;

                    if(!field_spec.equals(field_impl))
                    {
                        match = FUZZY_MATCH;
                    }
                    
                }
            }
        }

        if((reaction_2.getError() == INCORRECT_REACTION)&&(reaction_1.getError() == UNEXPECTED_REACTION))
        {
            // Message from reaction 1
            const ::cpptesk::hw::Message *message_1_spec = reaction_2.getSpecMessage();
            // Message from reaction 2
            const ::cpptesk::hw::Message *message_2_impl = reaction_1.getImplMessage();
            
            if(message_1_spec->size() == message_2_impl->size())
            {
                match = EXACT_MATCH;

                for(::cpptesk::hw::Message::const_iterator i = message_1_spec->begin(), 
                    j = message_2_impl->begin(); i != message_1_spec->end(); i++, j++)
                {
                    const Field<> &field_spec = (*message_1_spec).**i;
                    const Field<> &field_impl = (*message_2_impl).**j;

                    if(!field_spec.equals(field_impl))
                    {
                        match = FUZZY_MATCH;
                    }
                    
                }
            }
        }
         
        return match;
    }
}}}
