/*Copyright 2010-2012 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Protsenko (protsenko@ispras.ru)
// SYNOPSIS: Union Elimination 
//           (Rule # 4 in C++TESK.ErrorDiagnostics.ru.doc):
//           {[a:x]+[x:a] => [a:a]+[x:x]}
///////////////////////////////////////////////////////////////////////////////

#include <hw/diagnostics/rule/union.hpp>

namespace cpptesk {
namespace hw {
namespace diagnostics {

    ReactionSet UnionEliminationRule::transform(::cpptesk::hw::diagnostics::ReactionUnit &reaction_unit_1, 
                                                ::cpptesk::hw::diagnostics::ReactionUnit &reaction_unit_2)    
    // Collapse the reactions
    {
        ReactionSet transform_out;
        
        const ::cpptesk::hw::Message *message_1_spec,
                                     *message_2_impl;
 
        const ::cpptesk::hw::Interface *iface_1;
                
        if((reaction_unit_1.getReaction().getError() == MISSING_REACTION)&&(reaction_unit_2.getReaction().getError() == UNEXPECTED_REACTION))
        {
            // Message from reaction 1
            message_1_spec = reaction_unit_1.getReaction().getSpecMessage();
            // Message from reaction 2
            message_2_impl = reaction_unit_2.getReaction().getImplMessage();
            
            iface_1 = reaction_unit_1.getReaction().where();
        }
        else
        {
            // Message from reaction 1
            message_1_spec = reaction_unit_2.getReaction().getSpecMessage();
            // Message from reaction 2
            message_2_impl = reaction_unit_1.getReaction().getImplMessage();
            
            iface_1 = reaction_unit_2.getReaction().where();
        }

        // Get "start" and "end" time of reactions
        cycle_t start_time   =  reaction_unit_1.getReaction().whenStarted();
        cycle_t start_time_2 =  reaction_unit_2.getReaction().whenStarted();
        cycle_t end_time     =  reaction_unit_1.getReaction().whenStopped();
        cycle_t end_time_2   =  reaction_unit_2.getReaction().whenStopped();
        
        if(start_time > start_time_2) start_time = start_time_2; // Min start time
        if(end_time < end_time_2)     end_time = end_time_2;     // Max end time
        
        ::cpptesk::hw::Reaction reaction_normal = Reaction(NORMAL_REACTION, *iface_1, *message_1_spec, *message_2_impl, 
                                                          start_time, end_time, "normal reaction from rule#4");
        
        transform_out.addReaction(reaction_normal, reaction_unit_1.getIdentifier(), reaction_unit_1.getIdentifier());
        
        // Set flag
        reaction_unit_1.setCollapsed();
        reaction_unit_2.setCollapsed();
       
        return transform_out;
    }

    int UnionEliminationRule::match(::cpptesk::hw::Reaction reaction_1, 
                                    ::cpptesk::hw::Reaction reaction_2)
    // Compare the reactions
    {
        int match = Rule::FALSE_MATCH;

        if((reaction_1.getError() == MISSING_REACTION)&&(reaction_2.getError() == UNEXPECTED_REACTION))
        {
            // Message from reaction 1
            const ::cpptesk::hw::Message *message_1_spec = reaction_1.getSpecMessage();
            // Message from reaction 2
            const ::cpptesk::hw::Message *message_2_impl = reaction_2.getImplMessage();
    
            if(message_1_spec->size() == message_2_impl->size())
            {
                match = EXACT_MATCH;

                for(::cpptesk::hw::Message::const_iterator i = message_1_spec->begin(), j = message_2_impl->begin();
                    i != message_1_spec->end(); i++, j++)
                {
                    const Field<> &field_spec = (*message_1_spec).**i;
                    const Field<> &field_impl = (*message_2_impl).**j;

                    if(!field_spec.equals(field_impl))
                    {
                        match = FUZZY_MATCH;
                    }
                    
                }
            }
        }

        if((reaction_2.getError() == MISSING_REACTION)&&(reaction_1.getError() == UNEXPECTED_REACTION))
        {
            // Message from reaction 1
            const ::cpptesk::hw::Message *_message1Impl = reaction_1.getImplMessage();
            // Message from reaction 2
            const ::cpptesk::hw::Message *_message2Spec = reaction_2.getSpecMessage();
            
            if(_message2Spec->size() == _message1Impl->size())
            {
                match = EXACT_MATCH;

                for(::cpptesk::hw::Message::const_iterator i = _message2Spec->begin(), j = _message1Impl->begin();
                    i != _message2Spec->end(); i++, j++)
                {
                    const Field<> &field_spec = (*_message2Spec).**i;
                    const Field<> &field_impl = (*_message1Impl).**j;

                    if(!field_spec.equals(field_impl))
                    {
                        match = FUZZY_MATCH;
                    }
                    
                }
            }
        }
        
        return match;
    }

}}}
