/*
 * Copyright 2010-2012 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Point event is a triple <message, interface, time>.
//           Prolonged event has an additional time dimension (duration).
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/assertion.hpp>
#include <hw/errors.hpp>
#include <hw/message.hpp>
#include <hw/time.hpp>
#include <hw/utils/shared_ptr.hpp>

#include <list>
#include <string>
#include <sstream>

namespace cpptesk {
namespace hw {

class Interface;

///////////////////////////////////////////////////////////////////////////////
// Instant Event
///////////////////////////////////////////////////////////////////////////////
class PointEvent;

typedef SharedPtr<PointEvent>  event_ptr_t;
typedef std::list<event_ptr_t> event_list_t;

class PointEvent
{
    friend class EventHistory;
public:
    PointEvent(const Interface &iface, const Message &message, cycle_t start_time):
        _id(0), _iface(&iface), _message(message.clone()), _start_time(start_time) {}

    PointEvent(const Interface &iface, cycle_t start_time):
        _id(0), _iface(&iface), _message(0), _start_time(start_time) {}
        
    PointEvent():
        _id(0), _iface(0), _message(0), _start_time(CPPTESK_INFINITY_TIME) {}
        
    virtual ~PointEvent() {}

    unsigned id() const { return _id; }

    const Interface *where() const { return _iface; }
    const Message *what() const { return _message.get(); }
    cycle_t when() const { return _start_time; }

    bool isInput() const;
    bool isOutput() const;

    bool has_cause() const { return !_cause.isNull(); }
    const event_ptr_t& cause() const { return _cause; }
    void cause(const event_ptr_t &event) { _cause = event; }

protected:
    void id(unsigned v) { _id = v; }

protected:
    unsigned _id;
    const Interface *_iface;
    SharedPtr<Message> _message;
    cycle_t _start_time;
    event_ptr_t _cause;
};

////////////////////////////////////////////////////////////////////////////////
// Prolonged Event
////////////////////////////////////////////////////////////////////////////////

class ProlongedEvent: public PointEvent
{
public:
    ProlongedEvent(const Interface &iface, const Message &message, cycle_t start_time, cycle_t stop_time):
        PointEvent(iface, message, start_time), _stop_time(stop_time) {}

    ProlongedEvent(const Interface &iface, cycle_t start_time, cycle_t stop_time):
        PointEvent(iface, start_time), _stop_time(stop_time) {}

    ProlongedEvent():
        PointEvent(), _stop_time(CPPTESK_INFINITY_TIME) {}

    virtual ~ProlongedEvent() {}

    cycle_t whenStarted() const { return _start_time; }
    cycle_t whenStopped() const { return _stop_time; }

protected:
    cycle_t _stop_time;
};

///////////////////////////////////////////////////////////////////////////////
// Stimulus
///////////////////////////////////////////////////////////////////////////////

class Stimulus: public ProlongedEvent 
{
public:
    Stimulus(const Interface &iface, const Message &message, cycle_t start_time, cycle_t stop_time):
        ProlongedEvent(iface, message, start_time, stop_time) {}

    Stimulus(const Interface &iface, cycle_t start_time, cycle_t stop_time):
        ProlongedEvent(iface, start_time, stop_time) {}

    Stimulus():
        ProlongedEvent() {}

    virtual ~Stimulus() {}
};

///////////////////////////////////////////////////////////////////////////////
// Reaction
///////////////////////////////////////////////////////////////////////////////

class Reaction: public ProlongedEvent 
{
public:
    Reaction(error_type_t error, const Interface &iface, const Message &spec_msg, const Message &impl_msg,
        cycle_t start_time, cycle_t stop_time, const std::string &info, unsigned type = -1u):
        ProlongedEvent(iface, spec_msg, start_time, stop_time), _impl_msg(impl_msg.clone()), _error(error), _info(info), _type(type) {}

    Reaction(error_type_t error, const Interface &iface, const Message &msg,
        cycle_t start_time, cycle_t stop_time, const std::string &info, unsigned type = -1u):
        ProlongedEvent(iface, msg, start_time, stop_time),
        _impl_msg(msg.clone()),
        _error(error), _info(info), _type(type) {}

    Reaction(error_type_t error, const Interface &iface,
        cycle_t start_time, cycle_t stop_time, const std::string &info, unsigned type = -1u):
        ProlongedEvent(iface, start_time, stop_time), _impl_msg(0), _error(error), _info(info), _type(type) {}

    Reaction():
        ProlongedEvent(), _impl_msg(0), _error(NORMAL_REACTION), _info(""), _type(-1u) {}
        
    virtual ~Reaction() {}

    const Message *getSpecMessage() const { return _message.get();  }
    const Message *getImplMessage() const { return _impl_msg.get(); }

    /// Returns true iff the reaction is erroneous.
    bool isError() const { return _error != NORMAL_REACTION; }
    /// Returns the error type.
    error_type_t getError() const { return _error; }
    /// Returns the reaction's type (@depricated).
    uint32_t getType() const { return _type; }
    /// Returns the reaction's info.
    const std::string& getInfo() const { return _info; }
    /// Returns if the reaction is expected but has not been received yet.
    bool isExpected() const { return _error == MISSING_REACTION && _info.empty(); }
    /// Returns true iff the reaction is erroneous and wasn't expected.
    bool isRealError() const { return isError() && !isExpected(); }

    /// Returns the short description of the reaction.
    std::string getShortDescription() const;
    /// Returns the full description of the reaction.
    std::string getFullDescription() const;

protected:
    SharedPtr<Message> _impl_msg;
    error_type_t _error;
    std::string _info;
    unsigned _type;
};

inline std::ostream& operator << (std::ostream& out, const Reaction &reaction)
{
    return out << reaction.getFullDescription();
}

}} // namespace cpptesk::hw

