/*
 * Copyright 2010-2012 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Model of a hardware interface
///////////////////////////////////////////////////////////////////////////////

#include <hw/assertion.hpp>
#include <hw/debug.hpp>
#include <hw/interface.hpp>
#include <hw/message.hpp>
#include <hw/process/process.hpp>

#include <algorithm>
#include <ostream>
#include <vector>

#define __CPPTESK_CHECK_ARBITER() \
   do { \
        if(!_arbiter) { \
            CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_WARN, "Reaction arbiter is not set for the interface " << *this); \
        } \
   } while(false)

namespace cpptesk {
namespace hw {

void Interface::request(const Process &process, const Message &message)
{
    CPPTESK_DEBUG_PRINT(DEBUG_MORE, "[IFACE:REQUEST](iface=" << *this << ")");

    Component::request(process);
    _message = (Message*)&message;

    // Increment number of stimuli
    incrementEventNumber();
}

void Interface::capture()
{
    CPPTESK_DEBUG_PRINT(DEBUG_MORE, "[IFACE:CAPTURE](iface=" << *this << ")");

    Component::capture();
 }

void Interface::release()
{
    CPPTESK_DEBUG_PRINT(DEBUG_MORE, "[IFACE:RELEASE](iface=" << *this << ")");

    Component::release();
    _message = 0;
}

void Interface::appendRequest(const Process &process, const Message &message)
{
    getArbiter()->appendRequest(process, message);
    CPPTESK_DEBUG_PRINT(DEBUG_MORE, "[ARBITER:APPEND](iface=" << *this << "): " << process);

    _request_count++;
}

bool Interface::chooseRequest(const Process &process, const Message &message)
{
    bool choose = getArbiter()->chooseRequest(process, message);
    CPPTESK_DEBUG_PRINT(DEBUG_MORE, "[ARBITER:CHOOSE](iface=" << *this << "): " << process << ": " << (choose ? "YES" : "NO"));

    return choose;
}

bool Interface::preChooseRequest(const Process &process, const Message &message)
{
    bool choose = getArbiter()->preChooseRequest(process, message);
    CPPTESK_DEBUG_PRINT(DEBUG_MORE, "[ARBITER:PRECHOOSE](iface=" << *this << "): " << process << ": " << (choose ? "YES" : "NO"));

    return choose;
}

void Interface::removeRequest(const Process &process, const Message &message)
{
    CPPTESK_ASSERTION(_request_count > 0, "There are no requests to remove (probably, WAIT_REACTION() is missing)");
    CPPTESK_DEBUG_PRINT(DEBUG_MORE, "[ARBITER:REMOVE](iface=" << *this << "): " << process);

    getArbiter()->removeRequest(process, message);

    _request_count--;
}

void Interface::removeRequest(const Process &process, const Message &message, time_t timemark)
{
    removeRequest(process, message);

    _remove_reaction_timemark = timemark;
}

bool Interface::can_adapter_start(const Process &process, const Message &message)
{
    bool choose = isModelBased() ? getArbiter()->chooseRequest(process, message) : getArbiter()->preChooseRequest(process, message);
    CPPTESK_DEBUG_PRINT(DEBUG_MORE, "[ARBITER:CAN_ADAPTER_START](iface=" << *this << "): " << process << ": " << (choose ? "YES" : "NO"));
    return choose;
}

bool Interface::takeReaction(time_t timemark)
{
    CPPTESK_ASSERTION(_request_count > 0, "There are no requests to take");

    // Increment number of reactions
    incrementEventNumber();

    // If a request has been taken, returns false
    if(_take_reaction_timemark == timemark)
        { return false; }

    _take_reaction_timemark = timemark;

    // Otherwise returns true
    return true;
}

void Interface::takeOldReaction()
{
    // Increment number of reactions
    incrementEventNumber();
}

}} // namespace cpptesk::hw
