/*
 * Copyright 2010-2012 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Macroses to define a mediator for a hardware model
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/errors.hpp>
#include <hw/model.hpp>

#include <sstream>

////////////////////////////////////////////////////////////////////////////////
// Declaration of Mediator (Adapter) Classes
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_MEDIA
 * Declares a mediator (adapter) of a hardware model.
 */
#define CPPTESK_MEDIA(UserMedia, UserModel) \
    struct UserMedia: public UserModel

/** @def CPPTESK_SIGNAL_MEDIA
 * Declares a signal mediator (adapter) of a hardware model.
 */
#define CPPTESK_SIGNAL_MEDIA(UserMedia, UserModel, Inputs, Outputs) \
    struct UserMedia: public UserModel, public ::cpptesk::hw::IOSignals<Inputs, Outputs>

/** @def CPPTESK_ADAPTER
 * Alias of CPPTESK_MEDIA
 */
#define CPPTESK_ADAPTER(UserMedia, UserModel) CPPTESK_MEDIA(UserMedia, UserModel)

/** @def CPPTESK_SIGNAL_ADAPTER
 * Alias of CPPTESK_SIGNAL_MEDIA
 */
#define CPPTESK_SIGNAL_ADAPTER(UserMedia, UserModel, Inputs, Outputs) CPPTESK_SIGNAL_MEDIA(UserMedia, UserModel, Inputs, Outputs)

#ifdef HW_SHORT_NAMES
    #define MEDIA(UserMedia, UserModel)                           CPPTESK_MEDIA(UserMedia, UserModel)
    #define ADAPTER(UserMedia, UserModel)                         CPPTESK_ADAPTER(UserMedia, UserModel)
    #define SIGNAL_MEDIA(UserMedia, UserModel, Inputs, Outputs)   CPPTESK_SIGNAL_MEDIA(UserMedia, UserModel, Inputs, Outputs)
    #define SIGNAL_ADAPTER(UserMedia, UserModel, Inputs, Outputs) CPPTESK_SIGNAL_ADAPTER(UserMedia, UserModel, Inputs, Outputs)
#endif // HW_SHORT_NAMES

///////////////////////////////////////////////////////////////////////////////
// Reaction Arbitration
///////////////////////////////////////////////////////////////////////////////

#define __CPPTESK_DETECT_REACTION_LABEL $CPPTESK_WAIT_REACTION_REACTION

#define __CPPTESK_CHECK_DETECT_REACTION() \
    do { \
        if(false) goto __CPPTESK_DETECT_REACTION_LABEL; \
    } while(false)

#define __CPPTESK_DETECT_REACTION(cond) \
    do { \
        if(_process.listener) { \
            CPPTESK_WAIT(cond); \
            if(_iface.isModelBased()) { \
                CPPTESK_CALL_PARALLEL_EX(model_t::singleListener, _iface, _message, _START_MODE_OF_LISTENER); \
            } \
        } else if(!_message.isDetected()) { \
            /* Wait until the interface is ready (not captured by some other process) */ \
            /*CPPTESK_WAIT_TIMEOUT(_iface.isReady() || _iface.getOwner() == &_process, getReactionTimeout(_iface));*/ \
            /* Requst the interface arbiter */ \
            _iface.appendRequest(_process, _message); \
            /* Only one DETECT_REACTION() (WAIT_REACTION()) is allowed within one deserializer */ \
            __CPPTESK_DETECT_REACTION_LABEL: \
            CPPTESK_GET_MESSAGE() = *__cpptesk_original_reference_model_message; \
            \
            CPPTESK_WAIT_TIMEOUT((cond) && _iface.can_adapter_start(_process, _message) && \
                                 _iface.getRemoveReactionTimemark() < (time_t)time(), getReactionTimeout(_iface)); \
            \
            if(some_unexpected_reaction_matches(CPPTESK_GET_PARAMETERS())) { \
                _process.timeout = false; \
                _iface.takeOldReaction(); _iface.removeRequest(_process, _message); goto __CPPTESK_NEXT_REACTION_STOP_LABEL; } \
            \
            if(time() >= (getReactionTimeout(_iface) + _process.start_time())) { \
                _process.timeout = true; \
                onTimeoutViolation(_process); \
            } \
            /* Missing specification reaction */ \
            if(_process.timeout) \
                goto __CPPTESK_NEXT_REACTION_LABEL; \
        } \
    } while(false)

/**
 * @def CPPTESK_DETECT_REACTION
 * Detects a reaction by using a single-cycle condition but do not take it.
 * It is used when a detector is multi-cycle.
 */
#define CPPTESK_DETECT_REACTION(cond) \
    ::cpptesk::hw::Process &_process = CPPTESK_GET_PROCESS(); \
    ::cpptesk::hw::Interface &_iface = CPPTESK_GET_IFACE(); \
    ::cpptesk::hw::Message &_message = CPPTESK_GET_MESSAGE(); \
    ::cpptesk::hw::Message *__cpptesk_original_reference_model_message = (cpptesk::hw::Message*)_message.clone(); \
    bool __user_used_expect_reaction = false; \
    __CPPTESK_DETECT_REACTION(cond)

#define __CPPTESK_TAKE_REACTION() \
    do { \
        /*--------------------------------------------------------------------*/ \
        /* Unexpected Reaction Listener                                       */ \
        /*--------------------------------------------------------------------*/ \
        if(_process.listener) { \
            /* Do nothing */ \
        } \
        /*--------------------------------------------------------------------*/ \
        /* Reaction Deserializer                                              */ \
        /*--------------------------------------------------------------------*/ \
        else if(!_message.isDetected()) { \
            /* Paranoid check */ \
            if(_process.timeout) { \
                CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_WARN, \
                    "Time limit in WAIT_TIMEOUT() has been reached"); \
            } \
            /* Error if a reaction has been taken by someone else in case of modelBased arbiters */ \
            if(_iface.isModelBased()) { \
                if(!_iface.takeReaction(time())) { \
                    CPPTESK_ASSERTION(false, __CPPTESK_COMMON_INFO() << ": " << \
                    "Conflict of model reactions (use SET_ARBITER() to set an arbiter)"); \
                } \
                _iface.removeRequest(_process, _message, time()); \
            } \
        } \
    } while(false)

/**
 * @def CPPTESK_TAKE_REACTION
 * Takes a reaction (says to the interface that the reaction is taken).
 */
#define CPPTESK_TAKE_REACTION() \
    do { \
        ::cpptesk::hw::Process &_process = CPPTESK_GET_PROCESS(); \
        ::cpptesk::hw::Interface &_iface = CPPTESK_GET_IFACE(); \
        ::cpptesk::hw::Message &_message = CPPTESK_GET_MESSAGE(); \
        __CPPTESK_TAKE_REACTION(); \
    } while(false)

/** @def CPPTESK_WAIT_REACTION
 * Detects a reaction by using a single-cycle condition and takes it.
 */
#define CPPTESK_WAIT_REACTION(cond) \
    ::cpptesk::hw::Process &_process = CPPTESK_GET_PROCESS(); \
    ::cpptesk::hw::Interface &_iface = CPPTESK_GET_IFACE(); \
    ::cpptesk::hw::Message &_message = CPPTESK_GET_MESSAGE(); \
    ::cpptesk::hw::Message *__cpptesk_original_reference_model_message = (cpptesk::hw::Message*)_message.clone(); \
    bool __user_used_expect_reaction = false; \
    \
    if(!_process.listener && !_iface.isModelBased()) { /* Put the message into spec-reaction-buffer to let impl-unexp-reactions be compared with */ \
        ::cpptesk::hw::Reaction spec_reaction = \
           ::cpptesk::hw::Reaction(::cpptesk::hw::MISSING_REACTION, CPPTESK_GET_IFACE(), CPPTESK_GET_MESSAGE(), time(), time(), ""); \
        _iface.spec_reactions.push_back(spec_reaction); \
    } \
    \
    do { \
        __CPPTESK_DETECT_REACTION(cond); \
        /* Reaction is taken only if time limit is not reached */ \
        if(!_process.timeout) { \
            __CPPTESK_TAKE_REACTION(); \
        } \
    } while(false)

/** @def CPPTESK_EXPECT_REACTION
 * Enable next reactions to come and be deserialized by the adapter.
 */
#define CPPTESK_EXPECT_REACTION() \
    do { \
        if(CPPTESK_GET_PROCESS().listener) { \
            CPPTESK_ASSERTION(!__user_used_expect_reaction, "Macro CPPTESK_EXPECT_REACTION() can be used only once in one deserializer."); \
            CPPTESK_GET_IFACE().setDetectionFinished(true); \
            __user_used_expect_reaction = true; \
        } \
    } while(false)

/** @def __CPPTESK_EXPECT_REACTION
 * Enable next reactions to come and be deserialized by the adapter (to be used in NEXT_REACTION macro.
 */
#define __CPPTESK_EXPECT_REACTION() \
    do { \
        if(CPPTESK_GET_PROCESS().listener && !__user_used_expect_reaction) \
            CPPTESK_GET_IFACE().setDetectionFinished(true); \
        __user_used_expect_reaction = false; \
    } while(false)

/** @def CPPTESK_NEXT_REACTION
 * Finishes processing of a reaction and switches to the next one.
 */
#define CPPTESK_NEXT_REACTION() \
    do { \
        __CPPTESK_NEXT_REACTION_LABEL: \
        \
        /*--------------------------------------------------------------------*/ \
        /* Unexpected Reaction Listener                                       */ \
        /*--------------------------------------------------------------------*/ \
        if(_process.listener) \
        { \
            if(!_iface.isModelBased()) \
            { \
                CPPTESK_CALL_PARALLEL_EX(model_t::singleListener, _iface, _message, _START_MODE_OF_LISTENER); \
            } \
            \
            __CPPTESK_EXPECT_REACTION(); \
            \
            if(getUnexpectedReactionTimeout(_iface) != 0) \
                CPPTESK_DELAY(getUnexpectedReactionTimeout(_iface) + 1); \
            else \
                CPPTESK_CYCLE(); \
        } \
        /*--------------------------------------------------------------------*/ \
        /* Reaction Deserializer                                              */ \
        /*--------------------------------------------------------------------*/ \
        \
        /* The function "isDetected" always returns false but it supports earlier versions of adapters? */ \
        \
        else if(!_message.isDetected()) \
        { \
            /* Remember: the following section is started only for non-listeners! */ \
            \
            /* Check that DETECT_REACTION() (WAIT_REACTION()) exists */ \
            __CPPTESK_CHECK_DETECT_REACTION(); \
            \
            /* If the adapter uses the compare arbiter and the timeout is not still violated, */ \
            /* one should try to find appropriate message in the arbiter msg-list. If the message is found, Adapter tries to take it. */ \
            /* If the one didn't succeed in taking, it means that some other adapter had already taken the message */ \
            /* and this adapter should start checking again */ \
            \
            if(!_iface.isModelBased() && !_process.timeout) \
            { \
                if(!_iface.chooseRequest(_process, _message)) \
                { \
                    CPPTESK_CYCLE(); \
                    goto __CPPTESK_DETECT_REACTION_LABEL; \
                } \
                else if(!_iface.takeReaction(time())) \
                { \
                    CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_WARN, __CPPTESK_COMMON_INFO() << ": " << \
                    "The compare arbiter cannot make a decision (two model messages are the same)"); \
                    CPPTESK_CYCLE(); \
                    goto __CPPTESK_DETECT_REACTION_LABEL; \
                } \
                \
                _iface.removeRequest(_process, _message, time()); \
            } \
        } \
        __CPPTESK_NEXT_REACTION_STOP_LABEL: \
        delete(__cpptesk_original_reference_model_message); \
    } while(false)

#ifdef HW_SHORT_NAMES
    #define DETECT_REACTION(cond) CPPTESK_DETECT_REACTION(cond)
    #define TAKE_REACTION()       CPPTESK_TAKE_REACTION()
    #define WAIT_REACTION(cond)   CPPTESK_WAIT_REACTION(cond)
    #define NEXT_REACTION()       CPPTESK_NEXT_REACTION()
    #define EXPECT_REACTION()     CPPTESK_EXPECT_REACTION()
#endif // HW_SHORT_NAMES

///////////////////////////////////////////////////////////////////////////////
// Declaration of Listeners
///////////////////////////////////////////////////////////////////////////////

/*-----------------------------------------------------------------------------
// Example of a User-Defined Listener
//-----------------------------------------------------------------------------
DECLARE_LISTENER(listen_iface1)
{
    CPPTESK_START_LISTENER(msgtype);
    CPPTESK_CYCLE_BEFORE_CHECHING();
    if(!CPPTESK_IS_REACTION_TAKEN()) {
        CPPTESK_DESERIALIZE_UNEXPECTED_REACTION();
    }
    CPPTESK_CYCLE();
    STOP_LISTENING();
}
------------------------------------------------------------------------------*/

#if CPPTESKHW_CHECK_VERSION(111010)

#define CPPTESK_DECLARE_LISTENER(listener_name) \
    CPPTESK_DECLARE_PROCESS(listener_name)

#define CPPTESK_DEFINE_LISTENER(listener_name) \
    CPPTESK_DEFINE_PROCESS(listener_name)

#define CPPTESK_START_LISTENER(msgtype) \
    msgtype impl_msg; \
    ::cpptesk::hw::cycle_t start_time, stop_time; \
    ::cpptesk::hw::pid_t pid = CPPTESK_INCORRECT_PID; \
    CPPTESK_UNUSED_ARG(CPPTESK_GET_PROCESS()); \
    CPPTESK_UNUSED_ARG(CPPTESK_GET_MESSAGE()); \
    CPPTESK_START_PROCESS(); \
    while (true) {

#define CPPTESK_CYCLE_BEFORE_CHECKING() \
    adapters_map_t::iterator i = _adapters.find(&iface); \
    start_time = time(); \
    if(i != _adapters.end()) { \
        impl_msg.setDetected(true); \
        impl_msg.invalidateFields(); \
        (i->second->call)(CPPTESK_GET_PROCESS(), CPPTESK_GET_IFACE(), impl_msg /* input/output */); \
    }

#define CPPTESK_IS_REACTION_TAKEN() \
    CPPTESK_GET_IFACE().isReactionTaken(time())

#define CPPTESK_DESERIALIZE_UNEXPECTED_REACTIONS() \
        do { \
            std::stringstream buffer; \
            CPPTESK_WAIT_PROCESS_FINISHED(pid); \
            stop_time = time() - 1; \
            buffer << "Unexpected reaction: " << impl_msg; \
            ::cpptesk::hw::Reaction reaction(::cpptesk::hw::UNEXPECTED_REACTION, CPPTESK_GET_IFACE(), impl_msg, start_time, stop_time, buffer.str()); \
            CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_ERROR, reaction); \
            if(!(CPPTESK_GET_IFACE().getFailureIgnoreTypes() & ::cpptesk::hw::UNEXPECTED_REACTION)) \
                registerReaction(reaction); \
        } while(false)

#define CPPTESK_STOP_LISTENER() \
    } CPPTESK_STOP_PROCESS()

#ifdef HW_SHORT_NAMES
#define DECLARE_LISTENER(listener_name) CPPTESK_DECLARE_LISTENER(listener_name)
#define DEFINE_LISTENER(listener_name) CPPTESK_DEFINE_LISTENER(listener_name)
#define START_LISTENER(iface) CPPTESK_START_LISTENER(iface)
#define LISTENING() CPPTESK_LISTENING()
#define CYCLE_BEFORE_CHECKING() CPPTESK_CYCLE_BEFORE_CHECKING()
#define IS_REACTION_TAKEN() CPPTESK_IS_REACTION_TAKEN()
#define DESERIALIZE_UNEXPECTED_REACTIONS() CPPTESK_DESERIALIZE_UNEXPECTED_REACTIONS()
#define STOP_LISTENER() CPPTESK_STOP_LISTENER()
#endif

#endif //CPPTESKHW_CHECK_VERSION(111010)

#ifdef CPPTESK_DESERIALIZE_UNEXPECTED_REACTION
#define CPPTESK_DEFINE_BASIC_OUTPUT_LISTENER(listener_name, output_iface, message_type, condition) \
    CPPTESK_DECLARE_PROCESS(listener_name) { \
        message_type impl_msg; \
        ::cpptesk::hw::cycle_t start_time, stop_time; \
        ::cpptesk::hw::pid_t pid = CPPTESK_INCORRECT_PID; \
        CPPTESK_UNUSED_ARG(CPPTESK_GET_PROCESS()); \
        CPPTESK_UNUSED_ARG(CPPTESK_GET_MESSAGE()); \
        \
        CPPTESK_START_PROCESS(); \
        while(true) { \
            CPPTESK_WAIT(condition); \
            adapters_map_t::iterator i = _adapters.find(&iface); \
            \
            start_time = time(); \
            \
            if(i != _adapters.end()) { \
                impl_msg.setDetected(true); \
                impl_msg.invalidateFields(); \
                (i->second->call)(CPPTESK_GET_PROCESS(), CPPTESK_GET_IFACE(), impl_msg); \
            } \
            \
            if(!CPPTESK_GET_IFACE().isReactionTaken(time())) { \
                std::stringstream buffer; \
                \
                CPPTESK_WAIT_PROCESS_FINISHED(pid); \
                stop_time = time() - 1; \
                \
                buffer << "Unexpected reaction: " << impl_msg; \
                \
                ::cpptesk::hw::Reaction reaction(::cpptesk::hw::UNEXPECTED_REACTION, \
                    CPPTESK_GET_IFACE(), impl_msg, start_time, stop_time, buffer.str()); \
                CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_ERROR, reaction); \
                if(!(CPPTESK_GET_IFACE().getFailureIgnoreTypes() & ::cpptesk::hw::UNEXPECTED_REACTION)) \
                    registerReaction(reaction); \
            } \
            CPPTESK_CYCLE(); \
        } \
        CPPTESK_STOP_PROCESS(); \
    }
#else
#define CPPTESK_DEFINE_BASIC_OUTPUT_LISTENER(listener_name, output_iface, condition) \
    CPPTESK_DECLARE_PROCESS(listener_name) { \
        CPPTESK_UNUSED_ARG(CPPTESK_GET_PROCESS()); \
        CPPTESK_UNUSED_ARG(CPPTESK_GET_MESSAGE()); \
        CPPTESK_START_PROCESS(); \
        while(true) { \
            CPPTESK_WAIT(condition); \
            \
            if(!iface.isReactionTaken(time())) { \
                ::cpptesk::hw::Reaction reaction(::cpptesk::hw::UNEXPECTED_REACTION, \
                    CPPTESK_GET_IFACE(), time() - 1, time() - 1, "Unexpected reaction"); \
                CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_ERROR, reaction); \
                if(!(CPPTESK_GET_IFACE().getFailureIgnoreTypes() & ::cpptesk::hw::UNEXPECTED_REACTION)) \
                    registerReaction(reaction); \
             } \
             CPPTESK_CYCLE(); \
          } \
          CPPTESK_STOP_PROCESS(); \
      }
#endif

#define CPPTESK_CALL_OUTPUT_LISTENER(full_listener_name, output_iface) \
    CPPTESK_CALL_LOOP_EX(full_listener_name,  output_iface, ::cpptesk::hw::LOWEST_PRIORITY)

#ifdef HW_SHORT_NAMES
    #define DEFINE_BASIC_OUTPUT_LISTENER CPPTESK_DEFINE_BASIC_OUTPUT_LISTENER
    #define CALL_OUTPUT_LISTENER         CPPTESK_CALL_OUTPUT_LISTENER
#endif // HW_SHORT_NAMES

namespace cpptesk {
namespace hw {

}} // namespace cpptesk::hw
