/*
 * Copyright 2010-2012 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Basic model of a synchronous hardware design
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/assertion.hpp>
#include <hw/component.hpp>
#include <hw/debug.hpp>
#include <hw/history.hpp>
#include <hw/interface.hpp>
#include <hw/message.hpp>
#include <hw/params.hpp>
#include <hw/priority.hpp>
#include <hw/random.hpp>
#include <hw/signals.hpp>
#include <hw/veritool/signals.hpp>
#include <hw/synchronizer.hpp>
#include <hw/time.hpp>
#include <hw/test_tracer.hpp>
#include <hw/unused.hpp>
#include <hw/arbiter/arbiter.hpp>
#include <hw/process/context_switcher.hpp>
#include <hw/process/process.hpp>
#include <hw/process/process_scheduler.hpp>
#include <hw/process/process_stack.hpp>
#include <hw/process/process_table.hpp>
#include <hw/sequence/manager.hpp>

#include <tracer/tracer.hpp>
#include <ts/coverage.hpp>

#include <algorithm>
#include <map>
#include <iomanip>
#include <iostream>
#include <sstream>
#include <list>
#include <vector>

#include <assert.h>
#include <stdint.h>
#include <time.h>

#if defined(WIN32)
    #include <hw/utils/win32/ucontext.h>
#elif defined(__GNUC__)
    #include <ucontext.h>
#else
    #error "unknown platform"
#endif

////////////////////////////////////////////////////////////////////////////////
// Return Types of Special-Purpose Methods
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_PROCESS
 * Keywork marking a model process.
 */
#define CPPTESK_PROCESS void

/** @def CPPTESK_STIMULUS
 * Keywork marking a stimulus.
 */
#define CPPTESK_STIMULUS CPPTESK_PROCESS

/** @def CPPTESK_REACTION
 * Keywork marking a reaction.
 */
#define CPPTESK_REACTION CPPTESK_PROCESS

#ifdef HW_SHORT_NAMES
    #define PROCESS  CPPTESK_PROCESS
    #define STIMULUS CPPTESK_STIMULUS
    #define REACTION CPPTESK_REACTION
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Declaration of Model Classes
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_MODEL
 * Declares a model of a hardware component.
 */
#define CPPTESK_MODEL(UserModel) \
    struct UserModel##Name { static const char* name() { return #UserModel; } }; \
    struct UserModel: public ::cpptesk::hw::Model<UserModel, UserModel##Name>

#define CPPTESK_SIGNAL_MODEL(UserModel, Inputs, Outputs) \
    CPPTESK_MODEL(UserModel), public ::cpptesk::hw::IOSignals<Inputs, Outputs>

#ifdef HW_SHORT_NAMES
    #define MODEL        CPPTESK_MODEL
    #define SIGNAL_MODEL CPPTESK_SIGNAL_MODEL
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Parameters of Special-Purpose Methods
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_GET_PROCESS
 * Returns a reference to a process parameter.
 */
#define CPPTESK_GET_PROCESS() process

/** @def CPPTESK_GET_IFACE
 * Returns a reference to an interface parameter.
 */
#define CPPTESK_GET_IFACE() iface

/** @def CPPTESK_GET_MESSAGE
 * Returns a reference to a message parameter.
 */
#define CPPTESK_GET_MESSAGE() message

/** @def CPPTESK_CAST_MESSAGE
 * Casts a message parameter to a given type and returns a reference.
 */
#define CPPTESK_CAST_MESSAGE(MessageType) \
    ::cpptesk::hw::_cpptesk_cast_message<MessageType>(CPPTESK_GET_MESSAGE(), __FILE__, __LINE__)

#define CPPTESK_CONST_CAST_MESSAGE(MessageType) \
    ::cpptesk::hw::_cpptesk_const_cast_message<MessageType>(CPPTESK_GET_MESSAGE(), __FILE__, __LINE__)

/** @def CPPTESK_GET_PARAMETERS
 * Returns a comma-separated list of parameters.
 */
#define CPPTESK_GET_PARAMETERS() \
    CPPTESK_GET_PROCESS(), CPPTESK_GET_IFACE(), CPPTESK_GET_MESSAGE()

#ifdef HW_SHORT_NAMES
    #define GET_PROCESS()                   CPPTESK_GET_PROCESS()
    #define GET_IFACE()                     CPPTESK_GET_IFACE()
    #define GET_MESSAGE()                   CPPTESK_GET_MESSAGE()
    #define CAST_MESSAGE(MessageType)       CPPTESK_CAST_MESSAGE(MessageType)
    #define CONST_CAST_MESSAGE(MessageType) CPPTESK_CONST_CAST_MESSAGE(MessageType)
    #define GET_PARAMETERS()                CPPTESK_GET_PARAMETERS()
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Declaration of Special-Purpose Methods
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_DECLARE_PROCESS
 * Declares a process within a model class.
 */
#define CPPTESK_DECLARE_PROCESS(process_name) \
    virtual CPPTESK_PROCESS process_name(::cpptesk::hw::Process &CPPTESK_GET_PROCESS(), \
        ::cpptesk::hw::Interface &CPPTESK_GET_IFACE(), ::cpptesk::hw::Message &CPPTESK_GET_MESSAGE())

/** @def CPPTESK_DECLARE_NON_VIRTUAL_PROCESS
 * Declares a non-virtal process within a model class.
 */
#define CPPTESK_DECLARE_NON_VIRTUAL_PROCESS(process_name) \
    CPPTESK_PROCESS process_name(::cpptesk::hw::Process &CPPTESK_GET_PROCESS(), \
        ::cpptesk::hw::Interface &CPPTESK_GET_IFACE(), ::cpptesk::hw::Message &CPPTESK_GET_MESSAGE())

/** @def CPPTESK_DECLARE_STIMULUS
 * Declares a stimulus within a model class.
 */
#define CPPTESK_DECLARE_STIMULUS(stimulus_name) \
    virtual CPPTESK_STIMULUS stimulus_name(::cpptesk::hw::Process &CPPTESK_GET_PROCESS(), \
        ::cpptesk::hw::Interface &CPPTESK_GET_IFACE(), ::cpptesk::hw::Message &CPPTESK_GET_MESSAGE())

/** @def CPPTESK_DECLARE_PRECONDITION
 * Declares a stimulus precondition within a model class.
 */
#define CPPTESK_DECLARE_PRECONDITION(stimulus_name) \
    virtual bool pre_##stimulus_name(::cpptesk::hw::Interface &CPPTESK_GET_IFACE(), \
        ::cpptesk::hw::Message &CPPTESK_GET_MESSAGE())

/** @def CPPTESK_DECLARE_STIMULUS_WITH_PRE \
 * Declares a stimulus and its precondition.
 */
#define CPPTESK_DECLARE_STIMULUS_WITH_PRE(stimulus_name) \
    CPPTESK_DECLARE_PRECONDITION(stimulus_name); \
    CPPTESK_DECLARE_STIMULUS(stimulus_name)

/** @def CPPTESK_DECLARE_REACTION
 * Declares a reaction within a model class.
 */
#define CPPTESK_DECLARE_REACTION(reaction_name) \
    virtual CPPTESK_REACTION reaction_name(::cpptesk::hw::Process &CPPTESK_GET_PROCESS(), \
        ::cpptesk::hw::Interface &CPPTESK_GET_IFACE(), ::cpptesk::hw::Message &CPPTESK_GET_MESSAGE())

#define CPPTESK_DECLARE_INPUT_ADAPTER(adapter_name)  CPPTESK_DECLARE_PROCESS(adapter_name)
#define CPPTESK_DEFINE_INPUT_ADAPTER(adapter_name)   CPPTESK_DEFINE_PROCESS(adapter_name)

#define CPPTESK_DECLARE_OUTPUT_ADAPTER(adapter_name) CPPTESK_DECLARE_PROCESS(adapter_name)
#define CPPTESK_DEFINE_OUTPUT_ADAPTER(adapter_name)  CPPTESK_DEFINE_PROCESS(adapter_name)

#ifdef HW_SHORT_NAMES
    #define DECLARE_PROCESS             CPPTESK_DECLARE_PROCESS
    #define DECLARE_NON_VIRTUAL_PROCESS CPPTESK_DECLARE_NON_VIRTUAL_PROCESS
    #define DECLARE_STIMULUS            CPPTESK_DECLARE_STIMULUS
    #define DECLARE_PRECONDITION        CPPTESK_DECLARE_PRECONDITION
    #define DECLARE_STIMULUS_WITH_PRE   CPPTESK_DECLARE_STIMULUS_WITH_PRE
    #define DECLARE_REACTION            CPPTESK_DECLARE_REACTION
    #define DECLARE_INPUT_ADAPTER       CPPTESK_DECLARE_INPUT_ADAPTER
    #define DEFINE_INPUT_ADAPTER        CPPTESK_DEFINE_INPUT_ADAPTER
    #define DECLARE_OUTPUT_ADAPTER      CPPTESK_DECLARE_OUTPUT_ADAPTER
    #define DEFINE_OUTPUT_ADAPTER       CPPTESK_DEFINE_OUTPUT_ADAPTER
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Declaration of Special-Purpose Fields
////////////////////////////////////////////////////////////////////////////////

#if CPPTESKHW_CHECK_VERSION(111010)

// IMPORTANT: 'iface' should start the default adapter name
#define CPPTESK_DEFAULT_INPUT_ADAPTER(iface)  iface##_serialize
#define CPPTESK_DEFAULT_OUTPUT_ADAPTER(iface) iface##_deserialize

#define CPPTESK_GET_IFACE_MESSAGE_TYPE(iface) MessageType_##iface
// IMPORTANT: 'iface' should end the message name to support interface arrays.
#define CPPTESK_GET_IFACE_MESSAGE(iface)      message_##iface

#define CPPTESK_DECLARE_DEFAULT_INPUT_ADAPTER(iface) \
    /* Check presence of the iface */ \
    typedef CPPTESK_GET_IFACE_MESSAGE_TYPE(iface) __default_input_##iface; \
    CPPTESK_DECLARE_INPUT_ADAPTER(CPPTESK_DEFAULT_INPUT_ADAPTER(iface))
#define CPPTESK_DECLARE_DEFAULT_OUTPUT_ADAPTER(iface) \
    /* Check presence of the iface */ \
    typedef CPPTESK_GET_IFACE_MESSAGE_TYPE(iface) __default_output_##iface; \
    CPPTESK_DECLARE_OUTPUT_ADAPTER(CPPTESK_DEFAULT_OUTPUT_ADAPTER(iface))

#define CPPTESK_DEFINE_DEFAULT_INPUT_ADAPTER(AdapterClass_iface) \
    CPPTESK_DEFINE_INPUT_ADAPTER(CPPTESK_DEFAULT_INPUT_ADAPTER(AdapterClass_iface))
#define CPPTESK_DEFINE_DEFAULT_OUTPUT_ADAPTER(AdapterClass_iface) \
    CPPTESK_DEFINE_OUTPUT_ADAPTER(CPPTESK_DEFAULT_OUTPUT_ADAPTER(AdapterClass_iface))

#define __CPPTESK_DECLARE_IFACE(iface_class, iface, msg_type) \
    typedef msg_type CPPTESK_GET_IFACE_MESSAGE_TYPE(iface); \
    CPPTESK_GET_IFACE_MESSAGE_TYPE(iface) CPPTESK_GET_IFACE_MESSAGE(iface); \
    iface_class iface

#define __CPPTESK_DECLARE_IFACE_ARRAY(iface_class, iface, msg_type, array_size) \
    typedef msg_type CPPTESK_GET_IFACE_MESSAGE_TYPE(iface); \
    CPPTESK_GET_IFACE_MESSAGE_TYPE(iface) CPPTESK_GET_IFACE_MESSAGE(iface)[array_size]; \
    iface_class iface[array_size]

#define __CPPTESK_DECLARE_IFACE_ARRAY_2D(iface_class, iface, msg_type, i_size, j_size) \
    typedef msg_type CPPTESK_GET_IFACE_MESSAGE_TYPE(iface); \
    CPPTESK_GET_IFACE_MESSAGE_TYPE(iface) CPPTESK_GET_IFACE_MESSAGE(iface)[i_size][j_size]; \
    iface_class iface[i_size][j_size]

#define CPPTESK_DECLARE_IFACE(iface, msg_type) \
    __CPPTESK_DECLARE_IFACE(::cpptesk::hw::Interface, iface, msg_type)
#define CPPTESK_DECLARE_IFACE_ARRAY(iface, msg_type, array_size) \
    __CPPTESK_DECLARE_IFACE_ARRAY(::cpptesk::hw::Interface, iface, msg_type, array_size)
#define CPPTESK_DECLARE_IFACE_ARRAY_2D(iface, msg_type, i_size, j_size) \
    __CPPTESK_DECLARE_IFACE_ARRAY_2D(::cpptesk::hw::Interface, iface, msg_type, i_size, j_size)

#define CPPTESK_DECLARE_INPUT(iface, msg_type) \
    __CPPTESK_DECLARE_IFACE(::cpptesk::hw::InputInterface, iface, msg_type); \
    CPPTESK_DECLARE_DEFAULT_INPUT_ADAPTER(iface) { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); }
#define CPPTESK_DECLARE_INPUT_ARRAY(iface, msg_type, array_size) \
    __CPPTESK_DECLARE_IFACE_ARRAY(::cpptesk::hw::InputInterface, iface, msg_type, array_size); \
    CPPTESK_DECLARE_DEFAULT_INPUT_ADAPTER(iface) { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); }
#define CPPTESK_DECLARE_INPUT_ARRAY_2D(iface, msg_type, i_size, j_size) \
    __CPPTESK_DECLARE_IFACE_ARRAY_2D(::cpptesk::hw::InputInterface, iface, msg_type, i_size, j_size); \
    CPPTESK_DECLARE_DEFAULT_INPUT_ADAPTER(iface) { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); }

#define CPPTESK_DECLARE_OUTPUT(iface, msg_type, arbiter_type) \
    __CPPTESK_DECLARE_IFACE(::cpptesk::hw::OutputInterface, iface, msg_type); \
    CPPTESK_DECLARE_DEFAULT_OUTPUT_ADAPTER(iface) { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); } \
    CPPTESK_DECLARE_ARBITER(arbiter_type, iface##_matcher);
#define CPPTESK_DECLARE_OUTPUT_ARRAY(iface, msg_type, arbiter_type, array_size) \
    __CPPTESK_DECLARE_IFACE_ARRAY(::cpptesk::hw::OutputInterface, iface, msg_type, array_size); \
    CPPTESK_DECLARE_DEFAULT_OUTPUT_ADAPTER(iface) { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); } \
    CPPTESK_DECLARE_ARBITER_ARRAY(arbiter_type, iface##_matcher, array_size);
#define CPPTESK_DECLARE_OUTPUT_ARRAY_2D(iface, msg_type, arbiter_type, i_size, j_size) \
    __CPPTESK_DECLARE_IFACE_ARRAY_2D(::cpptesk::hw::OutputInterface, iface, msg_type, i_size, j_size); \
    CPPTESK_DECLARE_DEFAULT_OUTPUT_ADAPTER(iface) { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); } \
    CPPTESK_DECLARE_ARBITER_ARRAY_2D(arbiter_type, iface##_matcher, i_size, j_size);

#define CPPTESK_DECLARE_INOUT(iface, msg_type, arbiter_type) \
    __CPPTESK_DECLARE_IFACE(::cpptesk::hw::InoutInterface, iface, msg_type); \
    CPPTESK_DECLARE_DEFAULT_INPUT_ADAPTER(iface)  { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); } \
    CPPTESK_DECLARE_DEFAULT_OUTPUT_ADAPTER(iface) { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); } \
    CPPTESK_DECLARE_ARBITER(arbiter_type, iface##_matcher);
#define CPPTESK_DECLARE_INOUT_ARRAY(iface, msg_type, arbiter_type, array_size) \
    __CPPTESK_DECLARE_IFACE_ARRAY(::cpptesk::hw::InoutInterface, iface, msg_type, array_size); \
    CPPTESK_DECLARE_DEFAULT_INPUT_ADAPTER(iface)  { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); } \
    CPPTESK_DECLARE_DEFAULT_OUTPUT_ADAPTER(iface) { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); } \
    CPPTESK_DECLARE_ARBITER_ARRAY(arbiter_type, iface##_matcher, array_size);
#define CPPTESK_DECLARE_INOUT_ARRAY_2D(iface, msg_type, arbiter_type, i_size, j_size) \
    __CPPTESK_DECLARE_IFACE_ARRAY_2D(::cpptesk::hw::InoutInterface, iface, msg_type, i_size, j_size); \
    CPPTESK_DECLARE_DEFAULT_INPUT_ADAPTER(iface)  { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); } \
    CPPTESK_DECLARE_DEFAULT_OUTPUT_ADAPTER(iface) { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); } \
    CPPTESK_DECLARE_ARBITER_ARRAY_2D(arbiter_type, iface##_matcher, i_size, j_size);

#else

/** @def CPPTESK_DECLARE_IFACE
 * Declares a model interface.
 */
#define CPPTESK_DECLARE_IFACE(iface) \
    ::cpptesk::hw::Interface iface
#define CPPTESK_DECLARE_IFACE_ARRAY(iface, array_size) \
    ::cpptesk::hw::Interface iface[array_size]
#define CPPTESK_DECLARE_IFACE_ARRAY_2D(iface, i_size, j_size) \
    ::cpptesk::hw::Interface iface[i_size][j_size]

/** @def CPPTESK_DECLARE_INPUT
 * Declares an input interface.
 */
#define CPPTESK_DECLARE_INPUT(iface) \
    ::cpptesk::hw::InputInterface iface
#define CPPTESK_DECLARE_INPUT_ARRAY(iface, array_size) \
    ::cpptesk::hw::InputInterface iface[array_size]
#define CPPTESK_DECLARE_INPUT_ARRAY_2D(iface, i_size, j_size) \
    ::cpptesk::hw::InputInterface iface[i_size][j_size]

/** @def CPPTESK_DECLARE_OUTPUT
 * Declares an output interface.
 */
#define CPPTESK_DECLARE_OUTPUT(iface) \
    ::cpptesk::hw::OutputInterface iface
#define CPPTESK_DECLARE_OUTPUT_ARRAY(iface, array_size) \
    ::cpptesk::hw::OutputInterface iface[array_size]
#define CPPTESK_DECLARE_OUTPUT_ARRAY_2D(iface, i_size, j_size) \
    ::cpptesk::hw::OutputInterface iface[i_size][j_size]

/** @def CPPTESK_DECLARE_INOUT
 * Declares an input/output interface.
 */
#define CPPTESK_DECLARE_INOUT(iface) \
    ::cpptesk::hw::InoutInterface iface
#define CPPTESK_DECLARE_INOUT_ARRAY(iface, array_size) \
    ::cpptesk::hw::InoutInterface iface[array_size]
#define CPPTESK_DECLARE_INOUT_ARRAY_2D(iface, i_size, j_size) \
    ::cpptesk::hw::InoutInterface iface[i_size][j_size]

#endif // CPPTESKHW_CHECK_VERSION(111010)

/** @def CPPTESK_DECLARE_ARBITER
 * Declares a primary reaction arbiter (matcher).
 */
#define CPPTESK_DECLARE_ARBITER(type, arbiter) \
    type arbiter

#define CPPTESK_DECLARE_ARBITER_ARRAY(type, arbiter, array_size) \
    type arbiter[array_size]

#define CPPTESK_DECLARE_ARBITER_ARRAY_2D(type, arbiter, i_size, j_size) \
    type arbiter[i_size][j_size]

/** @def CPPTESK_DECLARE_TWO_LEVEL_ARBITER
 * Declares two-level reaction arbiter (matcher).
 */
#define CPPTESK_DECLARE_TWO_LEVEL_ARBITER(A1, A2, arbiter) \
    ::cpptesk::hw::TwoLevelArbiter<A1, A2> arbiter

#ifdef HW_SHORT_NAMES
    #define GET_IFACE_MESSAGE_TYPE         CPPTESK_GET_IFACE_MESSAGE_TYPE
    #define GET_IFACE_MESSAGE              CPPTESK_GET_IFACE_MESSAGE
    #define DEFAULT_INPUT_ADAPTER          CPPTESK_DEFAULT_INPUT_ADAPTER
    #define DEFAULT_OUTPUT_ADAPTER         CPPTESK_DEFAULT_OUTPUT_ADAPTER
    #define DECLARE_DEFAULT_INPUT_ADAPTER  CPPTESK_DECLARE_DEFAULT_INPUT_ADAPTER
    #define DECLARE_DEFAULT_OUTPUT_ADAPTER CPPTESK_DECLARE_DEFAULT_OUTPUT_ADAPTER
    #define DEFINE_DEFAULT_INPUT_ADAPTER   CPPTESK_DEFINE_DEFAULT_INPUT_ADAPTER
    #define DEFINE_DEFAULT_OUTPUT_ADAPTER  CPPTESK_DEFINE_DEFAULT_OUTPUT_ADAPTER
    #define DECLARE_IFACE                  CPPTESK_DECLARE_IFACE
    #define DECLARE_IFACE_ARRAY            CPPTESK_DECLARE_IFACE_ARRAY
    #define DECLARE_IFACE_ARRAY_2D         CPPTESK_DECLARE_IFACE_ARRAY_2D
    #define DECLARE_INPUT                  CPPTESK_DECLARE_INPUT
    #define DECLARE_INPUT_ARRAY            CPPTESK_DECLARE_INPUT_ARRAY
    #define DECLARE_INPUT_ARRAY_2D         CPPTESK_DECLARE_INPUT_ARRAY_2D
    #define DECLARE_OUTPUT                 CPPTESK_DECLARE_OUTPUT
    #define DECLARE_OUTPUT_ARRAY           CPPTESK_DECLARE_OUTPUT_ARRAY
    #define DECLARE_OUTPUT_ARRAY_2D        CPPTESK_DECLARE_OUTPUT_ARRAY_2D
    #define DECLARE_INOUT                  CPPTESK_DECLARE_INOUT
    #define DECLARE_INOUT_ARRAY            CPPTESK_DECLARE_INOUT_ARRAY
    #define DECLARE_INOUT_ARRAY_2D         CPPTESK_DECLARE_INOUT_ARRAY_2D
    #define DECLARE_ARBITER                CPPTESK_DECLARE_ARBITER
    #define DECLARE_ARBITER_ARRAY          CPPTESK_DECLARE_ARBITER_ARRAY
    #define DECLARE_ARBITER_ARRAY_2D       CPPTESK_DECLARE_ARBITER_ARRAY_2D
    #define DECLARE_TWO_LEVEL_ARBITER      CPPTESK_DECLARE_TWO_LEVEL_ARBITER
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Definition of Special-Purpose Methods
////////////////////////////////////////////////////////////////////////////////

#define CPPTESK_PROCESS_SIGNATURE() \
    ::cpptesk::hw::Process   &CPPTESK_GET_PROCESS(), \
    ::cpptesk::hw::Interface &CPPTESK_GET_IFACE(), \
    ::cpptesk::hw::Message   &CPPTESK_GET_MESSAGE()

/** @def CPPTESK_DEFINE_PROCESS
 * Defines a process.
 */
#define CPPTESK_DEFINE_PROCESS(process_name) \
    CPPTESK_PROCESS process_name(CPPTESK_PROCESS_SIGNATURE())

/** @def CPPTESK_DEFINE_STIMULUS
 * Defines a stimulus.
 */
#define CPPTESK_DEFINE_STIMULUS(stimulus_name) \
    CPPTESK_STIMULUS stimulus_name(CPPTESK_PROCESS_SIGNATURE())

/** @def CPPTESK_DEFINE_PRECONDITION
 * Defines a stimulus precondition.
 */
#define CPPTESK_DEFINE_PRECONDITION(pre_name) \
    bool pre_name(::cpptesk::hw::Interface &CPPTESK_GET_IFACE(), \
        ::cpptesk::hw::Message &CPPTESK_GET_MESSAGE())

/** @def CPPTESK_DEFINE_REACTION
 * Defines a reaction within a model class.
 */
#define CPPTESK_DEFINE_REACTION(reaction_name) \
    CPPTESK_REACTION reaction_name(CPPTESK_PROCESS_SIGNATURE())

#ifdef HW_SHORT_NAMES
    #define PROCESS_SIGNATURE   CPPTESK_PROCESS_SIGNATURE
    #define DEFINE_PROCESS      CPPTESK_DEFINE_PROCESS
    #define DEFINE_STIMULUS     CPPTESK_DEFINE_STIMULUS
    #define DEFINE_PRECONDITION CPPTESK_DEFINE_PRECONDITION
    #define DEFINE_REACTION     CPPTESK_DEFINE_REACTION
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Registration of Special-Purpose Components
////////////////////////////////////////////////////////////////////////////////

// Component
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_ADD_COMPONENT
 * Registers a component in a model.
 */
#define CPPTESK_ADD_COMPONENT(component) \
    do { \
        CPPTESK_ASSERTION(!(component).getParent(), "Component " << #component << " has been already registered"); \
        (component).updateName(#component); \
        (component).setParent(this); \
        registerComponent(component); \
    } while(false)

// Interface
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_ADD_IFACE
 * Registers an interface in a model.
 */
#define CPPTESK_ADD_IFACE(iface) \
    do { \
        CPPTESK_ADD_COMPONENT(iface); \
        /* Update interface coverage */ \
        __CPPTESK_IFACE_COVERAGE::add(&iface, iface.getName()); \
    } while(false)

#define CPPTESK_ADD_IFACE_ARRAY(iface, array_size) \
    do { \
        for(int __cpptesk_i = 0; __cpptesk_i < array_size; __cpptesk_i++) \
        { \
            std::stringstream __cpptesk_buffer; \
            __cpptesk_buffer << #iface << "[" << std::dec << __cpptesk_i << std::hex << "]"; \
            iface[__cpptesk_i].setName(__cpptesk_buffer.str()); \
            CPPTESK_ADD_IFACE(iface[__cpptesk_i]); \
        } \
    } while(false)

#define CPPTESK_ADD_IFACE_ARRAY_2D(iface, i_size, j_size) \
    do { \
        for(int __cpptesk_i = 0; __cpptesk_i < i_size; __cpptesk_i++) \
        for(int __cpptesk_j = 0; __cpptesk_j < j_size; __cpptesk_j++) \
        { \
            std::stringstream __cpptesk_buffer; \
            __cpptesk_buffer << #iface << "[" << std::dec << __cpptesk_i << std::hex << "][" << __cpptesk_j << "]"; \
            iface[__cpptesk_i][__cpptesk_j].setName(__cpptesk_buffer.str()); \
            CPPTESK_ADD_IFACE(iface[__cpptesk_i][__cpptesk_j]); \
        } \
    } while(false)

// Interface Adapter
////////////////////////////////////////////////////////////////////////////////

#define __CPPTESK_SET_IFACE_ARRAY_ADAPTER(IFACE_ADAPTER, iface, full_method_name, array_size) \
    do { \
        for(int __cpptesk_i = 0; __cpptesk_i < array_size; __cpptesk_i++) \
        { \
            CPPTESK_SET_##IFACE_ADAPTER(iface[__cpptesk_i], full_method_name); \
        } \
    } while(false)

#define __CPPTESK_SET_IFACE_ARRAY_2D_ADAPTER(IFACE_ADAPTER, iface, full_method_name, i_size, j_size) \
    do { \
        for(int __cpptesk_i = 0; __cpptesk_i < i_size; __cpptesk_i++) \
        for(int __cpptesk_j = 0; __cpptesk_j < j_size; __cpptesk_j++) \
        { \
            CPPTESK_SET_##IFACE_ADAPTER(iface[__cpptesk_i][__cpptesk_j], full_method_name); \
        } \
    } while(false)

#define __CPPTESK_SET_IFACE_ARRAY_ADAPTER_USER_LISTENER(IFACE_ADAPTER, iface, full_method_name, full_listener_name, array_size) \
    do { \
        for(int __cpptesk_i = 0; __cpptesk_i < array_size; __cpptesk_i++) \
        { \
            CPPTESK_SET_##IFACE_ADAPTER(iface[__cpptesk_i], full_method_name, full_listener_name); \
        } \
    } while(false)

#define __CPPTESK_SET_IFACE_ARRAY_2D_ADAPTER_USER_LISTENER(IFACE_ADAPTER, iface, full_method_name, full_listener_name, i_size, j_size) \
    do { \
        for(int __cpptesk_i = 0; __cpptesk_i < i_size; __cpptesk_i++) \
        for(int __cpptesk_j = 0; __cpptesk_j < j_size; __cpptesk_j++) \
        { \
            CPPTESK_SET_##IFACE_ADAPTER(iface[__cpptesk_i][__cpptesk_j], full_method_name, full_listener_name); \
        } \
    } while(false)

/** @def CPPTESK_SET_INPUT_ADAPTER
 * Registers an input interface adapter (aka serializer).
 */
#define CPPTESK_SET_INPUT_ADAPTER(iface, full_method_name) \
    registerInputAdapter(iface, (model_t::process_t)&full_method_name, #full_method_name)

#define CPPTESK_SET_INPUT_ARRAY_ADAPTER(iface, full_method_name, array_size) \
    __CPPTESK_SET_IFACE_ARRAY_ADAPTER(INPUT_ADAPTER, iface, full_method_name, array_size)
#define CPPTESK_SET_INPUT_ARRAY_2D_ADAPTER(iface, full_method_name, i_size, j_size) \
    __CPPTESK_SET_IFACE_ARRAY_2D_ADAPTER(INPUT_ADAPTER, iface, full_method_name, i_size, j_size)

#if CPPTESKHW_CHECK_VERSION(111010)

#define CPPTESK_SET_DEFAULT_INPUT_ADAPTER(iface) \
    CPPTESK_SET_INPUT_ADAPTER(iface, CPPTESK_DEFAULT_INPUT_ADAPTER(model_t::iface))
#define CPPTESK_SET_DEFAULT_INPUT_ARRAY_ADAPTER(iface, array_size) \
    CPPTESK_SET_INPUT_ARRAY_ADAPTER(iface, CPPTESK_DEFAULT_INPUT_ADAPTER(model_t::iface), array_size)
#define CPPTESK_SET_DEFAULT_INPUT_ARRAY_2D_ADAPTER(iface, i_size, j_size) \
    CPPTESK_SET_INPUT_ARRAY_2D_ADAPTER(iface, CPPTESK_DEFAULT_INPUT_ADAPTER(model_t::iface), i_size, j_size)

#define CPPTESK_SET_OUTPUT_ADAPTER(iface, full_method_name) \
    do { \
        registerOutputAdapter(iface, (model_t::process_t)&full_method_name, #full_method_name); \
        CPPTESK_CALL_PARALLEL_EX(model_t::listener, iface, CPPTESK_GET_IFACE_MESSAGE(iface), ::cpptesk::hw::EMPTY_MODEL::START_DONT_UPDATE_IFACE_COVERAGE | \
                                                                                             ::cpptesk::hw::HIGHEST_PRIORITY); \
    } while(false)

#define CPPTESK_SET_DEFAULT_OUTPUT_ADAPTER(iface) \
    CPPTESK_SET_OUTPUT_ADAPTER(iface, CPPTESK_DEFAULT_OUTPUT_ADAPTER(model_t::iface))
#define CPPTESK_SET_DEFAULT_OUTPUT_ARRAY_ADAPTER(iface, array_size) \
    CPPTESK_SET_OUTPUT_ARRAY_ADAPTER(iface, CPPTESK_DEFAULT_OUTPUT_ADAPTER(model_t::iface), array_size)
#define CPPTESK_SET_DEFAULT_OUTPUT_ARRAY_2D_ADAPTER(iface, i_size, j_size) \
    CPPTESK_SET_OUTPUT_ARRAY_2D_ADAPTER(iface, CPPTESK_DEFAULT_OUTPUT_ADAPTER(model_t::iface), i_size, j_size)

#define CPPTESK_SET_OUTPUT_ADAPTER_NO_LISTENER(iface, full_method_name) \
    do { \
        registerOutputAdapter(iface, (model_t::process_t)&full_method_name, #full_method_name); \
    } while(false)

#define CPPTESK_SET_OUTPUT_ARRAY_ADAPTER_NO_LISTENER(iface, full_method_name, array_size) \
    __CPPTESK_SET_IFACE_ARRAY_ADAPTER(OUTPUT_ADAPTER_NO_LISTENER, iface, full_method_name, array_size)
#define CPPTESK_SET_OUTPUT_ARRAY_2D_ADAPTER_NO_LISTENER(iface, full_method_name, i_size, j_size) \
    __CPPTESK_SET_IFACE_ARRAY_2D_ADAPTER(OUTPUT_ADAPTER_NO_LISTENER, iface, full_method_name, i_size, j_size)

#define CPPTESK_SET_DEFAULT_OUTPUT_ADAPTER_NO_LISTENER(iface) \
    CPPTESK_SET_OUTPUT_ADAPTER_NO_LISTENER(iface, CPPTESK_DEFAULT_OUTPUT_ADAPTER(model_t::iface))
#define CPPTESK_SET_DEFAULT_OUTPUT_ARRAY_ADAPTER_NO_LISTENER(iface, array_size) \
    CPPTESK_SET_OUTPUT_ARRAY_ADAPTER_NO_LISTENER(iface, CPPTESK_DEFAULT_OUTPUT_ADAPTER(model_t::iface), array_size)
#define CPPTESK_SET_DEFAULT_OUTPUT_ARRAY_2D_ADAPTER_NO_LISTENER(iface, i_size, j_size) \
    CPPTESK_SET_OUTPUT_ARRAY_2D_ADAPTER_NO_LISTENER(iface, CPPTESK_DEFAULT_OUTPUT_ADAPTER(model_t::iface), i_size, j_size)

#define CPPTESK_SET_OUTPUT_ADAPTER_USER_LISTENER(iface, full_method_name, full_listener_name) \
    do { \
        registerOutputAdapter(iface, (model_t::process_t)&full_method_name, #full_method_name); \
        CPPTESK_CALL_LOOP_EX(full_listener_name, iface, ::cpptesk::hw::EMPTY_MODEL::START_DONT_UPDATE_IFACE_COVERAGE | ::cpptesk::hw::LOWEST_PRIORITY); \
    } while(false)

#define CPPTESK_SET_OUTPUT_ARRAY_ADAPTER_USER_LISTENER(iface, full_method_name, full_listener_name, array_size) \
    __CPPTESK_SET_IFACE_ARRAY_ADAPTER_USER_LISTENER(OUTPUT_ADAPTER_USER_LISTENER, iface, full_method_name, full_listener_name, array_size)
#define CPPTESK_SET_OUTPUT_ARRAY_2D_ADAPTER_USER_LISTENER(iface, full_method_name, full_listener, i_size, j_size) \
    __CPPTESK_SET_IFACE_ARRAY_2D_ADAPTER_USER_LISTENER(OUTPUT_ADAPTER_USER_LISTENER, iface, full_method_name, full_listener_name, i_size, j_size)

#define CPPTESK_SET_DEFAULT_OUTPUT_ADAPTER_USER_LISTENER(iface, full_listener_name) \
    CPPTESK_SET_OUTPUT_ADAPTER_USER_LISTENER(iface, CPPTESK_DEFAULT_OUTPUT_ADAPTER(model_t::iface), full_listener_name)
#define CPPTESK_SET_DEFAULT_OUTPUT_ARRAY_ADAPTER_USER_LISTENER(iface, full_listener_name, array_size) \
    CPPTESK_SET_OUTPUT_ARRAY_ADAPTER_USER_LISTENER(iface, CPPTESK_DEFAULT_OUTPUT_ADAPTER(model_t::iface), array_size)
#define CPPTESK_SET_DEFAULT_OUTPUT_ARRAY_2D_ADAPTER_USER_LISTENER(iface, full_listener, i_size, j_size) \
    CPPTESK_SET_OUTPUT_ARRAY_2D_ADAPTER_USER_LISTENER(iface, CPPTESK_DEFAULT_OUTPUT_LISTENER(model_t::iface), full_listener_name, i_size, j_size)

#else

/** @def CPPTESK_SET_OUTPUT_ADAPTER
 * Registers an output interface adapter (aka deserializer).
 */
#define CPPTESK_SET_OUTPUT_ADAPTER(iface, full_method_name) \
    registerOutputAdapter(iface, (model_t::process_t)&full_method_name, #full_method_name)

#endif // CPPTESKHW_VERSION_CHECK

#define CPPTESK_SET_OUTPUT_ARRAY_ADAPTER(iface, full_method_name, array_size) \
    __CPPTESK_SET_IFACE_ARRAY_ADAPTER(OUTPUT_ADAPTER, iface, full_method_name, array_size)
#define CPPTESK_SET_OUTPUT_ARRAY_2D_ADAPTER(iface, full_method_name, i_size, j_size) \
    __CPPTESK_SET_IFACE_ARRAY_2D_ADAPTER(OUTPUT_ADAPTER, iface, full_method_name, i_size, j_size)

/** @def CPPTESK_SET_INPUT_ADAPTER_OF
 * Registers an input interface adapter for an inner unit's interface.
 */
#define CPPTESK_SET_INPUT_ADAPTER_OF(UnitClass, unit, iface, full_method_name) \
    do { \
        unit.registerInputAdapter(iface, reinterpret_cast<UnitClass::model_t&>(*this), \
            (UnitClass::model_t::process_t)&full_method_name, #full_method_name); \
    } while(false)

/** @def CPPTESK_SET_OUTPUT_ADAPTER_OF
 * Registers an output interface adapter for an inner unit's interface.
 */
#define CPPTESK_SET_OUTPUT_ADAPTER_OF(UnitClass, unit, iface, full_method_name) \
    do { \
        unit.registerOutputAdapter(iface, reinterpret_cast<UnitClass::model_t&>(*this), \
            (UnitClass::model_t::process_t)&full_method_name, #full_method_name); \
    } while(false)

#if CPPTESKHW_CHECK_VERSION(111010)

// Input: No Adapter
////////////////////////////////////////////////////////////////////////////////

#define CPPTESK_ADD_INPUT(iface) \
    CPPTESK_ADD_IFACE(iface)

#define CPPTESK_ADD_INPUT_ARRAY(iface, array_size) \
    CPPTESK_ADD_IFACE_ARRAY(iface, array_size)

#define CPPTESK_ADD_INPUT_ARRAY_2D(iface, i_size, j_size) \
    CPPTESK_ADD_IFACE_ARRAY_2D(iface, i_size, j_size)

// Input: Default Adapter
////////////////////////////////////////////////////////////////////////////////

#define CPPTESK_ADD_INPUT_WITH_DEFAULT_ADAPTER(iface) \
    do { \
        CPPTESK_ADD_IFACE(iface); \
        CPPTESK_SET_INPUT_ADAPTER(iface, model_t::CPPTESK_DEFAULT_INPUT_ADAPTER(iface)); \
    } while(false)

#define CPPTESK_ADD_INPUT_ARRAY_WITH_DEFAULT_ADAPTER(iface, array_size) \
    do { \
        CPPTESK_ADD_IFACE_ARRAY(iface, array_size); \
        CPPTESK_SET_INPUT_ARRAY_ADAPTER(iface, model_t::CPPTESK_DEFAULT_INPUT_ADAPTER(iface), array_size); \
    } while(false)

#define CPPTESK_ADD_INPUT_ARRAY_2D_WITH_DEFAULT_ADAPTER(iface, i_size, j_size) \
    do { \
        CPPTESK_ADD_IFACE_ARRAY_2D(iface, i_size, j_size); \
        CPPTESK_SET_INPUT_ARRAY_2D_ADAPTER(iface, model_t::CPPTESK_DEFAULT_INPUT_ADAPTER(iface), i_size, j_size); \
    } while(false)

// Output: No Adapter
////////////////////////////////////////////////////////////////////////////////

#define CPPTESK_ADD_OUTPUT(iface) \
    do { \
        CPPTESK_ADD_IFACE(iface); \
        CPPTESK_SET_ARBITER(iface, iface##_matcher); \
    } while(false)

#define CPPTESK_ADD_OUTPUT_ARRAY(iface, array_size) \
    do { \
        CPPTESK_ADD_IFACE_ARRAY(iface, array_size); \
        CPPTESK_SET_ARBITER_ARRAY(iface, iface##_matcher, array_size); \
    } while(false)

#define CPPTESK_ADD_OUTPUT_ARRAY_2D(iface, i_size, j_size) \
    do { \
        CPPTESK_ADD_IFACE_ARRAY_2D(iface, i_size, j_size); \
        CPPTESK_SET_ARBITER_ARRAY_2D(iface, iface##_matcher, i_size, j_size); \
    } while(false)

// Output: Default Adapter (No Listener)
////////////////////////////////////////////////////////////////////////////////

#define CPPTESK_ADD_OUTPUT_WITH_DEFAULT_ADAPTER(iface) \
    do { \
        CPPTESK_ADD_IFACE(iface); \
        CPPTESK_SET_OUTPUT_ADAPTER_NO_LISTENER(iface, model_t::CPPTESK_DEFAULT_OUTPUT_ADAPTER(iface)); \
        CPPTESK_SET_ARBITER(iface, iface##_matcher); \
    } while(false)

#define CPPTESK_ADD_OUTPUT_ARRAY_WITH_DEFAULT_ADAPTER(iface, array_size) \
    do { \
        CPPTESK_ADD_IFACE_ARRAY(iface, array_size); \
        CPPTESK_SET_OUTPUT_ARRAY_ADAPTER_NO_LISTENER(iface, model_t::CPPTESK_DEFAULT_OUTPUT_ADAPTER(iface), array_size); \
        CPPTESK_SET_ARBITER_ARRAY(iface, iface##_matcher, array_size); \
    } while(false)

#define CPPTESK_ADD_OUTPUT_ARRAY_2D_WITH_DEFAULT_ADAPTER(iface, i_size, j_size) \
    do { \
        CPPTESK_ADD_IFACE_ARRAY_2D(iface, i_size, j_size); \
        CPPTESK_SET_OUTPUT_ARRAY_2D_ADAPTER_NO_LISTENER(iface, model_t::CPPTESK_DEFAULT_OUTPUT_ADAPTER(iface), i_size, j_size); \
        CPPTESK_SET_ARBITER_ARRAY_2D(iface, iface##_matcher, i_size, j_size); \
    } while(false)

#else

#define CPPTESK_ADD_INPUT(iface) \
    CPPTESK_ADD_IFACE(iface)
#define CPPTESK_ADD_INPUT_ARRAY(iface, array_size) \
    CPPTESK_ADD_IFACE_ARRAY(iface, array_size)
#define CPPTESK_ADD_INPUT_ARRAY_2D(iface, i_size, j_size) \
    CPPTESK_ADD_IFACE_ARRAY_2D(iface, i_size, j_size)

#define CPPTESK_ADD_OUTPUT(iface) \
    CPPTESK_ADD_IFACE(iface)
#define CPPTESK_ADD_OUTPUT_ARRAY(iface, array_size) \
    CPPTESK_ADD_IFACE_ARRAY(iface, array_size)
#define CPPTESK_ADD_OUTPUT_ARRAY_2D(iface, i_size, j_size) \
    CPPTESK_ADD_IFACE_ARRAY_2D(iface, i_size, j_size)

#endif // CPPTESKHW_CHECK_VERSION

/** @def CPPTESK_SET_ARBITER
 * Registers a primary arbiter for a given output interface.
 */
#define CPPTESK_SET_ARBITER(iface, arbiter) \
    iface.registerArbiter(arbiter)

#define CPPTESK_SET_ARBITER_ARRAY(iface, arbiter, array_size) \
    do { \
        for(int __cpptesk_i = 0; __cpptesk_i < array_size; __cpptesk_i++) \
        { \
            CPPTESK_SET_ARBITER(iface[__cpptesk_i], arbiter[__cpptesk_i]); \
        } \
    } while(false)

#define CPPTESK_SET_ARBITER_ARRAY_2D(iface, arbiter, i_size, j_size) \
    do { \
        for(int __cpptesk_i = 0; __cpptesk_i < i_size; __cpptesk_i++) \
        for(int __cpptesk_j = 0; __cpptesk_j < j_size; __cpptesk_j++) \
        { \
            CPPTESK_SET_ARBITER(iface[__cpptesk_i][__cpptesk_j], arbiter[__cpptesk_i][__cpptesk_j]); \
        } \
    } while(false)

#ifdef HW_SHORT_NAMES
    #define ADD_COMPONENT                                     CPPTESK_ADD_COMPONENT
    #define ADD_IFACE                                         CPPTESK_ADD_IFACE
    #define ADD_IFACE_ARRAY                                   CPPTESK_ADD_IFACE_ARRAY
    #define ADD_IFACE_ARRAY_2D                                CPPTESK_ADD_IFACE_ARRAY_2D
    #define ADD_INPUT                                         CPPTESK_ADD_INPUT
    #define ADD_INPUT_ARRAY                                   CPPTESK_ADD_INPUT_ARRAY
    #define ADD_INPUT_ARRAY_2D                                CPPTESK_ADD_INPUT_ARRAY_2D
    #define ADD_OUTPUT                                        CPPTESK_ADD_OUTPUT
    #define ADD_OUTPUT_ARRAY                                  CPPTESK_ADD_OUTPUT_ARRAY
    #define ADD_OUTPUT_ARRAY_2D                               CPPTESK_ADD_OUTPUT_ARRAY_2D
    #define SET_ARBITER                                       CPPTESK_SET_ARBITER
    #define SET_ARBITER_ARRAY                                 CPPTESK_SET_ARBITER_ARRAY
    #define SET_ARBITER_ARRAY_2D                              CPPTESK_SET_ARBITER_ARRAY_2D
    #define SET_INPUT_ADAPTER                                 CPPTESK_SET_INPUT_ADAPTER
    #define SET_INPUT_ARRAY_ADAPTER                           CPPTESK_SET_INPUT_ARRAY_ADAPTER
    #define SET_INPUT_ARRAY_2D_ADAPTER                        CPPTESK_SET_INPUT_ARRAY_2D_ADAPTER
    #define SET_OUTPUT_ADAPTER                                CPPTESK_SET_OUTPUT_ADAPTER
    #define SET_OUTPUT_ARRAY_ADAPTER                          CPPTESK_SET_OUTPUT_ARRAY_ADAPTER
    #define SET_OUTPUT_ARRAY_2D_ADAPTER                       CPPTESK_SET_OUTPUT_ARRAY_2D_ADAPTER
    #define SET_DEFAULT_INPUT_ADAPTER                         CPPTESK_SET_DEFAULT_INPUT_ADAPTER
    #define SET_DEFAULT_INPUT_ARRAY_ADAPTER                   CPPTESK_SET_DEFAULT_INPUT_ARRAY_ADAPTER
    #define SET_DEFAULT_INPUT_ARRAY_2D_ADAPTER                CPPTESK_SET_DEFAULT_INPUT_ARRAY_2D_ADAPTER
    #define SET_DEFAULT_OUTPUT_ADAPTER                        CPPTESK_SET_DEFAULT_OUTPUT_ADAPTER
    #define SET_DEFAULT_OUTPUT_ARRAY_ADAPTER                  CPPTESK_SET_DEFAULT_OUTPUT_ARRAY_ADAPTER
    #define SET_DEFAULT_OUTPUT_ARRAY_2D_ADAPTER               CPPTESK_SET_DEFAULT_OUTPUT_ARRAY_2D_ADAPTER
    #define SET_INPUT_ADAPTER_OF                              CPPTESK_SET_INPUT_ADAPTER_OF
    #define SET_OUTPUT_ADAPTER_OF                             CPPTESK_SET_OUTPUT_ADAPTER_OF
#if CPPTESKHW_CHECK_VERSION(111010)
    #define ADD_INPUT_WITH_DEFAULT_ADAPTER                    CPPTESK_ADD_INPUT_WITH_DEFAULT_ADAPTER
    #define ADD_INPUT_ARRAY_WITH_DEFAULT_ADAPTER              CPPTESK_ADD_INPUT_ARRAY_WITH_DEFAULT_ADAPTER
    #define ADD_INPUT_ARRAY_2D_WITH_DEFAULT_ADAPTER           CPPTESK_ADD_INPUT_ARRAY_2D_WITH_DEFAULT_ADAPTER
    #define ADD_OUTPUT_WITH_DEFAULT_ADAPTER                   CPPTESK_ADD_OUTPUT_WITH_DEFAULT_ADAPTER
    #define ADD_OUTPUT_ARRAY_WITH_DEFAULT_ADAPTER             CPPTESK_ADD_OUTPUT_ARRAY_WITH_DEFAULT_ADAPTER
    #define ADD_OUTPUT_ARRAY_2D_WITH_DEFAULT_ADAPTER          CPPTESK_ADD_OUTPUT_ARRAY_2D_WITH_DEFAULT_ADAPTER
    #define SET_OUTPUT_ADAPTER_NO_LISTENER                    CPPTESK_SET_OUTPUT_ADAPTER_NO_LISTENER
    #define SET_OUTPUT_ARRAY_ADAPTER_NO_LISTENER              CPPTESK_SET_OUTPUT_ARRAY_ADAPTER_NO_LISTENER
    #define SET_OUTPUT_ARRAY_ADAPTER_2D_NO_LISTENER           CPPTESK_SET_OUTPUT_ARRAY_ADAPTER_2D_NO_LISTENER
    #define SET_OUTPUT_ADAPTER_USER_LISTENER                  CPPTESK_SET_OUTPUT_ADAPTER_USER_LISTENER
    #define SET_OUTPUT_ARRAY_ADAPTER_USER_LISTENER            CPPTESK_SET_OUTPUT_ARRAY_ADAPTER_USER_LISTENER
    #define SET_OUTPUT_ARRAY_ADAPTER_2D_USER_LISTENER         CPPTESK_SET_OUTPUT_ARRAY_ADAPTER_2D_USER_LISTENER
    #define SET_DEFAULT_OUTPUT_ADAPTER_NO_LISTENER            CPPTESK_SET_DEFAULT_OUTPUT_ADAPTER_NO_LISTENER
    #define SET_DEFAULT_OUTPUT_ARRAY_ADAPTER_NO_LISTENER      CPPTESK_SET_DEFAULT_OUTPUT_ARRAY_ADAPTER_NO_LISTENER
    #define SET_DEFAULT_OUTPUT_ARRAY_ADAPTER_2D_NO_LISTENER   CPPTESK_SET_DEFAULT_OUTPUT_ARRAY_ADAPTER_2D_NO_LISTENER
    #define SET_DEFAULT_OUTPUT_ADAPTER_USER_LISTENER          CPPTESK_SET_DEFAULT_OUTPUT_ADAPTER_USER_LISTENER
    #define SET_DEFAULT_OUTPUT_ARRAY_ADAPTER_USER_LISTENER    CPPTESK_SET_DEFAULT_OUTPUT_ARRAY_ADAPTER_USER_LISTENER
    #define SET_DEFAULT_OUTPUT_ARRAY_ADAPTER_2D_USER_LISTENER CPPTESK_SET_DEFAULT_OUTPUT_ARRAY_ADAPTER_2D_USER_LISTENER

#endif // CPPTESKHW_CHECK_VERSION
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Parameters
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_SET_REACTION_TIMEOUT
 * Sets reaction timeout.
 */
#define CPPTESK_SET_REACTION_TIMEOUT(timeout) \
    setReactionTimeout(timeout)

#define CPPTESK_SET_UNEXPECTED_REACTION_TIMEOUT(timeout) \
    setUnexpectedReactionTimeout(timeout)

#ifdef HW_SHORT_NAMES
    #define SET_REACTION_TIMEOUT(timeout) CPPTESK_SET_REACTION_TIMEOUT(timeout)
    #define SET_UNEXPECTED_REACTION_TIMEOUT(timeout) CPPTESK_SET_UNEXPECTED_REACTION_TIMEOUT(timeout)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Control Statements
////////////////////////////////////////////////////////////////////////////////

/**
 * Process status is reassigned to EXECUTING to handle nested calls.
 */
#define __CPPTESK_SYNPOINT(type, name) \
    { \
        CPPTESK_GET_PROCESS().set_point(::cpptesk::hw::Process::PointDescriptor(type, name, time(), __FILE__, __LINE__)); \
        /* IMPORTANT: Timeout flag is not reset by CYCLE() and DELAY() */ \
        if(type == ::cpptesk::hw::GUARD) { CPPTESK_GET_PROCESS().timeout = false; } \
        CPPTESK_GET_PROCESS().set_status(::cpptesk::hw::EXECUTING); \
    }

#define __CPPTESK_RETURN(result) \
    do { \
        /* To check presense of CPPTESK_START_PROCESS() */ \
        __CPPTESK_CHECK_START(); \
        CPPTESK_ASSERTION(CPPTESK_GET_PROCESS().call_stack_size() > 0, "There is no an opening START for the current STOP"); \
        CPPTESK_GET_PROCESS().set_status(result); \
        /* External parallel process jumps to the parent process */ \
        CPPTESK_GET_PROCESS().stop_call(); \
        if(CPPTESK_GET_PROCESS().call_stack_size() == 0) { \
            suspend_process(CPPTESK_GET_PROCESS(), CPPTESK_INFINITY_TIME); \
        } \
        return; \
    } while(false)

////////////////////////////////////////////////////////////////////////////////
// Precondition and Postcondition
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_CHECK_CONDITION
 * Returns false if a condition is not satistied. The macros is typically used
 * in precondition methods.
 */
#define CPPTESK_CHECK_CONDITION(cond) \
    if(!(cond)) \
    { \
        CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_INFO, "Condition " << #cond << " is violated"); \
        return false; \
    }

/** @def CPPTESK_PRE
 * Precondition assertion. The macros is typically used in stimulus methods.
 */
#define CPPTESK_PRE(cond) \
    if(!(cond)) \
    { \
        CPPTESK_GET_PROCESS().set_point(::cpptesk::hw::OTHER, \
            ::cpptesk::hw::Process::PointDescriptor(type, "PRE(" #cond ")", time(), __FILE__, __LINE__)); \
        CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_INFO, "Pre-condition " << #cond << " is violated"); \
        __CPPTESK_RETURN(::cpptesk::hw::UNPREDICTED); \
    }

/** @def CPPTESK_POST
 * Postcondition assertion. The macros is typically used in stimulus methods.
 */
#define CPPTESK_POST(cond) \
    if(!(cond)) \
    { \
        CPPTESK_GET_PROCESS().set_point(::cpptesk::hw::OTHER, \
            ::cpptesk::hw::Process::PointDescriptor(type, "POST(" #cond ")", time(), __FILE__, __LINE__)); \
        CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_INFO, "Post-condition " << #cond << " is violated"); \
        __CPPTESK_RETURN(::cpptesk::hw::VIOLATION); \
    }

#ifdef HW_SHORT_NAMES
    #define CHECK_CONDITION(cond) CPPTESK_CHECK_CONDITION(cond)
    #define PRE(cond)             CPPTESK_PRE(cond)
    #define POST(cond)            CPPTESK_POST(cond)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Call Process
////////////////////////////////////////////////////////////////////////////////

#define __CPPTESK_PROCESS_PTR() ((::cpptesk::hw::Process*)__cpptesk_process)

#define __CPPTESK_CHECK_IFACE(iface) \
    ::cpptesk::hw::__cpptesk_assertion_file_line((iface).getStatus() == ::cpptesk::hw::Interface::IDLE, \
        "Interface " + (iface).getName() + " has been already requested", __FILE__, __LINE__)

#define __CPPTESK_CHECK_START_MODE(model, full_method_name, mode) \
    ::cpptesk::hw::__cpptesk_assertion_file_line((model).inside_cycle() || \
        (mode & ::cpptesk::hw::EMPTY_MODEL::START_THIS_CYCLE) == 0, \
        "Cannot start " #full_method_name " between two cycles", __FILE__, __LINE__)

#define __CPPTESK_PREPARE_CALL(method_name) \
    prepare_call(::cpptesk::hw::Process::CallDescriptor(#method_name, __FILE__, __LINE__))

#define CPPTESK_CALL_PARALLEL(method_name, iface, message) \
    ( \
        /* IMPORTANT: Do not check existance of START() */ \
        __CPPTESK_PREPARE_CALL(method_name), \
        start((model_t::process_t)&model_t::method_name, iface, message, __CPPTESK_PROCESS_PTR(), START_IMMEDIATELY) \
    )

#define CPPTESK_CALL_SEQUENTIAL(method_name, iface, message) \
    ( \
        __CPPTESK_CHECK_START(), \
        __CPPTESK_PREPARE_CALL(method_name), \
        start((model_t::process_t)&model_t::method_name, CPPTESK_GET_PROCESS(), iface, message) \
    )

#define CPPTESK_CALL_PARALLEL_EX(full_method_name, iface, message, mode) \
    ( \
        /* IMPORTANT: Do not check existence of START() */ \
        __CPPTESK_PREPARE_CALL(full_method_name), \
        start((model_t::process_t)&full_method_name, iface, message, __CPPTESK_PROCESS_PTR(), mode) \
    )

#define CPPTESK_CALL_SEQUENTIAL_EX(full_method_name, iface, message) \
    ( \
        /* IMPORANT: Do not check existence of START() */ \
        __CPPTESK_PREPARE_CALL(full_method_name), \
        start((model_t::process_t)&full_method_name, CPPTESK_GET_PROCESS(), iface, message) \
    )

#define CPPTESK_CALL_PROCESS(call_mode, method_name, iface, message) \
    ((call_mode) == PARALLEL ? \
        CPPTESK_CALL_PARALLEL(method_name, iface, message) : \
        CPPTESK_CALL_SEQUENTIAL(method_name, iface, message))

#define CPPTESK_CALL_PROCESS_EX(call_mode, full_method_name, iface, message, mode) \
    ((call_mode) == PARALLEL ? \
        /* IMPORTANT: Parallel call does not use the process parameter */ \
        CPPTESK_CALL_PARALLEL_EX(full_method_name, iface, message, mode) : \
        /* IMPORTANT: Sequential call does not use the mode parameter */ \
        CPPTESK_CALL_SEQUENTIAL_EX(full_method_name, iface, message))

#define CPPTESK_RECV_STIMULUS(call_mode, iface, message) \
    CPPTESK_CALL_PROCESS(call_mode, onStimulus, iface, message)

#define CPPTESK_SEND_REACTION(call_mode, iface, message) \
    CPPTESK_CALL_PROCESS(call_mode, onReaction, iface, \
   _cpptesk_set_explanation(message, __FILE__, __LINE__))

#define CPPTESK_SEND_REACTION_WITH_COMMENT(call_mode, iface, message, comment) \
    CPPTESK_CALL_PROCESS(call_mode, onReaction, iface, \
    _cpptesk_set_explanation(message, __FILE__, __LINE__, comment))

#define CPPTESK_CALL_PROCESS_OF(model, full_method_name, iface, message) \
    ( \
        (model).__CPPTESK_PREPARE_CALL(full_method_name), \
        /* The interface is not locked */ \
        (model).start(&full_method_name, iface, message, __CPPTESK_PROCESS_PTR()) \
    )

#define CPPTESK_CALL_PROCESS_OF_EX(model, full_method_name, iface, message, mode) \
    ( \
        (model).__CPPTESK_PREPARE_CALL(full_method_name), \
        /* Check the start mode */ \
        __CPPTESK_CHECK_START_MODE(model, full_method_name, mode), \
        /* Go */ \
        (model).start(&full_method_name, iface, message, __CPPTESK_PROCESS_PTR(), mode) \
    )

#define CPPTESK_CALL_STIMULUS_OF(model, full_method_name, iface, message) \
    ( \
        /* Trace the call */ \
        ::cpptesk::hw::Tracer::get().trace(model.time(), #model, #full_method_name, #iface, #message, message), \
        /* Prepare the call */ \
        (model).__CPPTESK_PREPARE_CALL(full_method_name), \
        /* Check the interface */ \
        __CPPTESK_CHECK_IFACE(iface), \
        /* Go and lock the interface */ \
        (model).start(&full_method_name, iface, message, __CPPTESK_PROCESS_PTR(), ::cpptesk::hw::EMPTY_MODEL::START_REQUEST_IFACE) \
    )

#define CPPTESK_CALL_STIMULUS_OF_EX(model, full_method_name, iface, message, mode) \
    ( \
        /* Trace the call */ \
        ::cpptesk::hw::Tracer::get().trace(model.time(), #model, #full_method_name, #iface, #message, message), \
        /* Prepare the call */ \
        (model).__CPPTESK_PREPARE_CALL(full_method_name), \
         /* Check the interface */ \
        __CPPTESK_CHECK_IFACE(iface), \
        /* Check the start mode */ \
        __CPPTESK_CHECK_START_MODE(model, full_method_name, mode), \
        /* Go and lock the interface */ \
        (model).start(&full_method_name, iface, message, __CPPTESK_PROCESS_PTR(), ::cpptesk::hw::EMPTY_MODEL::START_REQUEST_IFACE | (mode)) \
    )

#define CPPTESK_CALL_REACTION_OF(model, full_method_name, iface, message) \
    ( \
        (model).__CPPTESK_PREPARE_CALL(full_method_name), \
        /* Go */ \
        (model).start(&full_method_name, iface, message, __CPPTESK_PROCESS_PTR()) \
    )

#define CPPTESK_CALL_REACTION_OF_EX(model, full_method_name, iface, message, mode) \
    ( \
        (model).__CPPTESK_PREPARE_CALL(full_method_name), \
        /* Check the start mode */ \
        __CPPTESK_CHECK_START_MODE(model, full_method_name, mode), \
        /* Go */ \
        (model).start(&full_method_name, iface, message, __CPPTESK_PROCESS_PTR(), mode) \
    )

/** @def CPPTESK_CALL_LOOP
 * Launches an infinite loop.
 */
#define CPPTESK_CALL_LOOP(method_name) \
    CPPTESK_CALL_PARALLEL(method_name, CPPTESK_DEFAULT_IFACE, ::cpptesk::hw::CPPTESK_DEFAULT_MESSAGE)

/** @def CPPTESK_CALL_LOOP_EX
 * Launches an infinite loop (usually, a listener of an output interface).
 */
#define CPPTESK_CALL_LOOP_EX(full_method_name, iface, mode) \
    CPPTESK_CALL_PARALLEL_EX(full_method_name, iface, ::cpptesk::hw::CPPTESK_DEFAULT_MESSAGE, mode)

#ifdef HW_SHORT_NAMES
    #define CALL_PARALLEL              CPPTESK_CALL_PARALLEL
    #define CALL_SEQUENTIAL            CPPTESK_CALL_SEQUENTIAL
    #define CALL_PROCESS               CPPTESK_CALL_PROCESS
    #define CALL_PARALLEL_EX           CPPTESK_CALL_PARALLEL_EX
    #define CALL_SEQUENTIAL_EX         CPPTESK_CALL_SEQUENTIAL_EX
    #define CALL_PROCESS_EX            CPPTESK_CALL_PROCESS_EX
    #define RECV_STIMULUS              CPPTESK_RECV_STIMULUS
    #define SEND_REACTION              CPPTESK_SEND_REACTION
    #define SEND_REACTION_WITH_COMMENT CPPTESK_SEND_REACTION_WITH_COMMENT
    #define CALL_PROCESS_OF            CPPTESK_CALL_PROCESS_OF
    #define CALL_PROCESS_OF_EX         CPPTESK_CALL_PROCESS_OF_EX
    #define CALL_STIMULUS_OF           CPPTESK_CALL_STIMULUS_OF
    #define CALL_STIMULUS_OF_EX        CPPTESK_CALL_STIMULUS_OF_EX
    #define CALL_REACTION_OF           CPPTESK_CALL_REACTION_OF
    #define CALL_REACTION_OF_EX        CPPTESK_CALL_REACTION_OF_EX
    #define CALL_LOOP                  CPPTESK_CALL_LOOP
    #define CALL_LOOP_EX               CPPTESK_CALL_LOOP_EX
#endif // HW_SHORT_NAMES

#define CPPTESK_CALL_STACK_OF(model) \
    ((model).process_stack())

#define CPPTESK_CALL_STACK() \
    CPPTESK_CALL_STACK_OF(*this)

#ifdef HW_SHORT_NAMES
    #define CALL_STACK_OF(model) CPPTESK_CALL_STACK_OF(model)
    #define CALL_STACK()         CPPTESK_CALL_STACK()
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Common part of model debug messages
////////////////////////////////////////////////////////////////////////////////

#define __CPPTESK_WHERE_AND_WHEN(iface, time) \
    "(" << iface << ":" << std::dec << time << std::hex << ")"

#define __CPPTESK_COMMON_INFO() \
    __CPPTESK_WHERE_AND_WHEN(CPPTESK_GET_IFACE(), time())

#define __CPPTESK_COMMON_START_INFO(start_time) \
    __CPPTESK_WHERE_AND_WHEN(CPPTESK_GET_IFACE(), start_time)

#define __CPPTESK_COMMON_STOP_INFO(start_time, stop_time) \
    __CPPTESK_WHERE_AND_WHEN(CPPTESK_GET_IFACE(), "[" << start_time << ".." << stop_time << "]")

////////////////////////////////////////////////////////////////////////////////
// Start and Stop
////////////////////////////////////////////////////////////////////////////////

#define __CPPTESK_START_LABEL $CPPTESK_START_PROCESS
#define __CPPTESK_START_VAR   _CPPTESK_START_PROCESS

#define __CPPTESK_CHECK_START() (void)__CPPTESK_START_VAR

#define __CPPTESK_INSIDE_PROCESS(process_stack, process) \
    /* Inform macroses that they are inside a process */ \
    bool  __cpptesk_inside_process = true; \
    void *__cpptesk_process_stack  = &(process_stack); \
    void *__cpptesk_process = &process; \
    CPPTESK_UNUSED_ARG(__cpptesk_inside_process); \
    CPPTESK_UNUSED_ARG(__cpptesk_process_stack); \
    CPPTESK_UNUSED_ARG(__cpptesk_process)

/** @def CPPTESK_START
 * Starts a model process.
 */
#define CPPTESK_START() \
    /* Inform macroses that they are inside a process */ \
    __CPPTESK_INSIDE_PROCESS(process_stack(), CPPTESK_GET_PROCESS()); \
    \
    /* To check that CPPTESK_START() preceedes any process call */ \
    int __CPPTESK_START_VAR; \
    \
    /* Only one CPPTESK_START() is allowed within a specification function */ \
    __CPPTESK_START_LABEL: \
    \
    do { \
        /* To avoid the 'unused variable' warning */ \
        __CPPTESK_CHECK_START(); \
        /* To avoid the 'unused label' warning */ \
        if(false) { goto __CPPTESK_START_LABEL; } \
        \
        CPPTESK_GET_PROCESS().start_call(); \
        CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_MORE, (CPPTESK_GET_PROCESS().call_stack_size() > 1 ? "[SEQUENTIAL:START]" : "[PARALLEL:START]") << \
            __CPPTESK_COMMON_START_INFO(time()) << ": " << CPPTESK_GET_PROCESS()); \
        /* Check the message invariant */ \
        CPPTESK_ASSERTION(CPPTESK_GET_MESSAGE().invariant(), "Message invariant has failed: " << CPPTESK_GET_MESSAGE()); \
    } while(false)

/** @def CPPTESK_START_PROCESS
 * Starts a process.
 */
#define CPPTESK_START_PROCESS() \
    CPPTESK_START()

/** @def CPPTESK_START_STIMULUS
 * Starts a stimulus and launches the input interface adapter.
 */
#define CPPTESK_START_STIMULUS(call_mode) \
    CPPTESK_START(); \
    CPPTESK_RECV_STIMULUS(call_mode, CPPTESK_GET_IFACE(), CPPTESK_GET_MESSAGE())

/** @def CPPTESK_START_REACTION
 * Starts a reaction and launches the output interface adapter.
 */
#define CPPTESK_START_REACTION(call_mode) \
    CPPTESK_START(); \
    CPPTESK_SEND_REACTION(call_mode, CPPTESK_GET_IFACE(), CPPTESK_GET_MESSAGE())

/** @def CPPTESK_STOP
 * Stops a model process.
 */
#define CPPTESK_STOP() \
    do { \
        /* To check presense of CPPTESK_START() */ \
        __CPPTESK_CHECK_START(); \
         CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_MORE, (CPPTESK_GET_PROCESS().call_stack_size() > 1 ? "[SEQUENTIAL:STOP]" : "[PARALLEL:STOP]") << \
             __CPPTESK_COMMON_START_INFO(time()) << ": " << CPPTESK_GET_PROCESS()); \
        __CPPTESK_RETURN(::cpptesk::hw::COMPLETED); \
    } while(false)

/** @def CPPTESK_STOP_PROCESS
 * Stops a process.
 */
#define CPPTESK_STOP_PROCESS()  CPPTESK_STOP()

/** @def CPPTESK_STOP_STIMULUS
 * Stops a stimulus.
 */
#define CPPTESK_STOP_STIMULUS() CPPTESK_STOP()

/** @def CPPTESK_STOP_REACTION
 * Stops a reaction.
 */
#define CPPTESK_STOP_REACTION() CPPTESK_STOP()

#ifdef HW_SHORT_NAMES
    #define START()                   CPPTESK_START()
    #define START_PROCESS()           CPPTESK_START_PROCESS()
    #define START_STIMULUS(call_mode) CPPTESK_START_STIMULUS(call_mode)
    #define START_REACTION(call_mode) CPPTESK_START_REACTION(call_mode)
    #define STOP()                    CPPTESK_STOP()
    #define STOP_PROCESS()            CPPTESK_STOP_PROCESS()
    #define STOP_STIMULUS()           CPPTESK_STOP_STIMULUS()
    #define STOP_REACTION()           CPPTESK_STOP_REACTION()
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Switch, Cycle, Delay, and Guard
////////////////////////////////////////////////////////////////////////////////

#define __CPPTESK_CHECK_STACK(unit) \
    CPPTESK_ASSERTION((unit).process_stack().size() > 1, \
    "Cannot interrupt a process if the process stack is empty"); \

/** @def CPPTESK_SWITCH
 * Emulates an inter-cycle context switching of a unit.
 */
#define CPPTESK_SWITCH_OF(unit) \
    do { \
        __CPPTESK_SYNPOINT(::cpptesk::hw::OTHER, "SWITCH()"); \
        __CPPTESK_CHECK_STACK(unit); \
        (unit).suspend_process(CPPTESK_GET_PROCESS()); \
    } while(false)

/** @def CPPTESK_SWITCH
 * Emulates an inter-cycle context switching.
 */
#define CPPTESK_SWITCH() \
    CPPTESK_SWITCH_OF(*this)

/** @def CPPTESK_CYCLE_OF
 * Emulates a clock cycle of a unit.
 */
#define CPPTESK_CYCLE_OF(unit) \
    do { \
        __CPPTESK_SYNPOINT(::cpptesk::hw::CYCLE, "CYCLE()"); \
        __CPPTESK_CHECK_STACK(unit); \
        (unit).suspend_process(CPPTESK_GET_PROCESS(), (unit).time() + 1); \
    } while(false)

/** @def CPPTESK_CYCLE
 * Emulates a clock cycle.
 */
#define CPPTESK_CYCLE() \
    CPPTESK_CYCLE_OF(*this)

////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_DELAY_OF
 * Emulates several clock cycles of a unit.
 */
#define CPPTESK_DELAY_OF(unit, delay) \
    do { \
        __CPPTESK_SYNPOINT(::cpptesk::hw::CYCLE, "CYCLE()"); \
        while((time() - CPPTESK_GET_PROCESS().get_point().time) < (::cpptesk::hw::cycle_t)(delay)) \
        { \
            __CPPTESK_CHECK_STACK(unit); \
            (unit).suspend_process(CPPTESK_GET_PROCESS(), (unit).time() + 1); \
        } \
    } while(false)

/** @def CPPTESK_DELAY
 * Emulates several clock cycles.
 */
#define CPPTESK_DELAY(delay) \
    CPPTESK_DELAY_OF(*this, delay)

////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_GUARD_OF
 * Waits unit a condition becomes true.
 */
#define CPPTESK_GUARD_OF(unit, condition) \
    do { \
        __CPPTESK_SYNPOINT(::cpptesk::hw::GUARD, "WAIT(" #condition ")"); \
        while(!(condition)) \
        { \
            __CPPTESK_CHECK_STACK(unit); \
            (unit).suspend_process(CPPTESK_GET_PROCESS(), (unit).time() + 1); \
        } \
        /* Do not wait anymore */ \
        __CPPTESK_SYNPOINT(::cpptesk::hw::CYCLE, "CYCLE()"); \
    } while(false)

/** @def CPPTESK_GUARD
 * Waits until a condition becomes true.
 */
#define CPPTESK_GUARD(condition) \
    CPPTESK_GUARD_OF(*this, condition)

////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_GUARD_TIMEOUT_OF
 * Waits until a condition becomes true or time is over.
 */
#define CPPTESK_GUARD_TIMEOUT_OF(unit, condition, limit) \
    do { \
        __CPPTESK_SYNPOINT(::cpptesk::hw::GUARD, "WAIT(" #condition ")"); \
        while(!(condition)) \
        { \
            /* Timeout violation */ \
            if((time() - CPPTESK_GET_PROCESS().get_point().time) >= (::cpptesk::hw::cycle_t)(limit)) \
            { \
                CPPTESK_GET_PROCESS().timeout = true; \
                onTimeoutViolation(CPPTESK_GET_PROCESS()); \
                break; \
            } \
            __CPPTESK_CHECK_STACK(unit); \
            (unit).suspend_process(CPPTESK_GET_PROCESS(), (unit).time() + 1); \
        } \
        /* Do not wait anymore */ \
        __CPPTESK_SYNPOINT(::cpptesk::hw::CYCLE, "CYCLE()"); \
    } while(false)

/** @def CPPTESK_GUARD_TIMEOUT
 * Waits until a condition becomes true or time is over.
 */
#define CPPTESK_GUARD_TIMEOUT(condition, limit) \
    CPPTESK_GUARD_TIMEOUT_OF(*this, condition, limit)

////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_WAIT_OF
 * Alias of CPPTESK_GUARD_OF
 */
#define CPPTESK_WAIT_OF(unit, condition) \
    CPPTESK_GUARD_OF(unit, condition)

/** @def CPPTESK_WAIT
 * Alias of CPPTESK_GUARD
 */
#define CPPTESK_WAIT(condition) \
    CPPTESK_GUARD(condition)

/** @def CPPTESK_WAIT_TIMEOUT_OF
 * Alias of CPPTESK_GUARD_TIMEOUT_OF
 */
#define CPPTESK_WAIT_TIMEOUT_OF(unit, condition, limit) \
    CPPTESK_GUARD_TIMEOUT_OF(unit, condition, limit)

/** @def CPPTESK_WAIT_TIMEOUT
 * Alias of CPPTESK_GUARD_TIMEOUT
 */
#define CPPTESK_WAIT_TIMEOUT(condition, limit) \
    CPPTESK_GUARD_TIMEOUT(condition, limit)

////////////////////////////////////////////////////////////////////////////////

#ifdef HW_SHORT_NAMES
    #define SWITCH_OF(unit)                          CPPTESK_SWITCH_OF(unit)
    #define SWITCH()                                 CPPTESK_SWITCH()
    #define CYCLE_OF(unit)                           CPPTESK_CYCLE_OF(unit)
    #define CYCLE()                                  CPPTESK_CYCLE()
    #define DELAY_OF(unit, delay)                    CPPTESK_DELAY_OF(unit, delay)
    #define DELAY(delay)                             CPPTESK_DELAY(delay)
    #define GUARD_OF(unit, condition)                CPPTESK_GUARD_OF(unit, condition)
    #define GUARD(condition)                         CPPTESK_GUARD(condition)
    #define GUARD_TIMEOUT_OF(unit, condition, limit) CPPTESK_GUARD_TIMEOUT_OF(unit, condition, limit)
    #define GUARD_TIMEOUT(condition, limit)          CPPTESK_GUARD_TIMEOUT(condition, limit)
    #define WAIT_OF(unit, condition)                 CPPTESK_WAIT_OF(unit, condition)
    #define WAIT(condition)                          CPPTESK_WAIT(condition)
    #define WAIT_TIMEOUT_OF(unit, condition, limit)  CPPTESK_WAIT_TIMEOUT_OF(unit, condition, limit)
    #define WAIT_TIMEOUT(condition, limit)           CPPTESK_WAIT_TIMEOUT(condition, limit)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Waiting Process
////////////////////////////////////////////////////////////////////////////////

#define CPPTESK_WAIT_PROCESS_FINISHED(pid) \
    CPPTESK_WAIT(!is_process_alive(pid))

#define CPPTESK_WAIT_PROCESS_FINISHED_TIMEOUT(pid, limit) \
    CPPTESK_WAIT_TIMEOUT(!is_process_alive(pid), limit)

#ifdef HW_SHORT_NAMES
    #define WAIT_PROCESS_FINISHED(pid)                CPPTESK_WAIT_PROCESS_FINISHED(pid)
    #define WAIT_PROCESS_FINISHED_TIMEOUT(pid, limit) CPPTESK_WAIT_PROCESS_FINISHED_TIMEOUT(pid, limit)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Process Attributes
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_GET_PRIORITY
 * Gets a process priority.
 */
#define CPPTESK_GET_PRIORITY() \
    ((unsigned)(CPPTESK_GET_PROCESS().get_priority()))

/** @def CPPTESK_SET_PRIORITY
 * Sets a process priority.
 */
#define CPPTESK_SET_PRIORITY(priority) \
    do { \
        CPPTESK_GET_PROCESS().set_priority((unsigned)priority); \
    } while(false)

/** @def CPPTESK_GET_INPUT_IFACE
 * Gets a process input interface.
 */
#define CPPTESK_GET_INPUT_IFACE() \
    (CPPTESK_GET_PROCESS().input_iface)

/** @def CPPTESK_SET_INPUT_IFACE
 * Sets a process input interface.
 */
#define CPPTESK_SET_INPUT_IFACE(iface) \
    do { \
        CPPTESK_GET_PROCESS().input_iface = &iface; \
    } while(false)

#ifdef HW_SHORT_NAMES
    #define GET_PRIORITY()         CPPTESK_GET_PRIORITY()
    #define SET_PRIORITY(priority) CPPTESK_SET_PRIORITY(priority)
    #define GET_INPUT_IFACE()      CPPTESK_GET_INPUT_IFACE()
    #define SET_INPUT_IFACE(iface) CPPTESK_SET_INPUT_IFACE(iface)
#endif // HW_SHORT_NAMES

///////////////////////////////////////////////////////////////////////////////
// Model Events
///////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_RAISE_EVENT
 * Raises an event.
 */
#define CPPTESK_RAISE_EVENT(id) \
    raise(sequence::ModelEvent(id, _model_event_manager))

/** @def CPPTESK_CHECK_EVENT
 * Checks whether an event occures.
 */
#define CPPTESK_CHECK_EVENT(id) \
    check(sequence::ModelEvent(id, _model_event_manager))

/** @def CPPTESK_WAIT_EVENT
 * Waits until an event occures.
 */
#define CPPTESK_WAIT_EVENT(id) \
    CPPTESK_WAIT(CPPTESK_CHECK_EVENT(id))

/** @def CPPTESK_WAIT_EVENT_TIMEOUT
 * Waits until an event occures but not longer than an amount of time specified.
 */
#define CPPTESK_WAIT_EVENT_TIMEOUT(id, timeout) \
    CPPTESK_WAIT_TIMEOUT(CPPTESK_CHECK_EVENT(id), timeout)

#ifdef HW_SHORT_NAMES
    #define RAISE_EVENT(id)                 CPPTESK_RAISE_EVENT(id)
    #define CHECK_EVENT(id)                 CPPTESK_CHECK_EVENT(id)
    #define WAIT_EVENT(id)                  CPPTESK_WAIT_EVENT(id)
    #define WAIT_EVENT_TIMEOUT(id, timeout) CPPTESK_WAIT_EVENT_TIMEOUT(id, timeout)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Interface Coverage
////////////////////////////////////////////////////////////////////////////////

#define __CPPTESK_DEFINE_IFACE_COVERAGE(cov_name) \
    struct __CPPTESK_IFACE_COVERAGE_$_enum \
    { \
        typedef cpptesk::hw::Interface *Id; \
        typedef cpptesk::ts::coverage::CoverageElement<Id> Element; \
        typedef std::vector<Element> ElementVector; \
        typedef ElementVector::iterator ElementIterator; \
        typedef cpptesk::utils::Range<ElementIterator> ElementRange; \
        \
        enum Command { ADD, GET }; \
        \
        static ElementRange main(Command cmd, const Id &id, const std::string &label) \
        { \
            static ElementVector _elements; \
            \
            if(cmd == ADD) \
                { _elements.push_back(Element(id, label)); } \
            \
            return ElementRange(_elements.begin(), _elements.end(), _elements.size()); \
        } \
        \
        static ElementRange get() { return main(GET, Id(), ""); } \
        static void add(const Id &id, const std::string &label) { main(ADD, id, label); } \
        \
        struct Matcher \
        { \
            Id id; \
            inline bool operator()(const Id& other) const { return id == other; } \
            inline Matcher& operator= (const Id& _id) { id = _id; return *this; } \
        }; \
    }; \
    struct __CPPTESK_IFACE_COVERAGE_$_name { static inline const char* name() { return cov_name; } }; \
    typedef cpptesk::ts::coverage::Coverage<__CPPTESK_IFACE_COVERAGE_$_enum, __CPPTESK_IFACE_COVERAGE_$_name> __CPPTESK_IFACE_COVERAGE;

namespace cpptesk {
namespace hw {

////////////////////////////////////////////////////////////////////////////////
// Basic Model Class
////////////////////////////////////////////////////////////////////////////////

class BaseModel: public Synchronizer, public ProcessTable
{
protected:
    BaseModel():
        ProcessTable() {}
    BaseModel(ProcessScheduler &process_scheduler, ContextSwitcher &context_switcher):
        ProcessTable(process_scheduler, context_switcher) {}
    virtual ~BaseModel() {}
};

template<class UserModel /* Model's sub-class */, class UserModelName> class Model: public BaseModel
{
public:
    typedef UserModel model_t;

    const static int DEFAULT_REACTION_TIMEOUT = 10;
    const static int DEFAULT_UNEXPECTED_REACTION_TIMEOUT = 0;

    static const char* name() { return UserModelName::name();}

    typedef void (UserModel::*method_t)(Process &process, Interface &iface, Message &message);

    typedef method_t process_t;
    typedef method_t stimulus_t;
    typedef method_t reaction_t;

    enum call_mode_t
    {
        PARALLEL,
        SEQUENTIAL
    };

    /// [7:0] - Process priority
    /// [*:8] - Control flags
    enum start_mode_t
    {
        //---------------------------------------------------------------------
        // Basic Start Modes
        //---------------------------------------------------------------------

        /// Calling start() adds a new process into the pipeline; the process will be executed when cycle() is called.
        START_NORMAL                     = 0x0000,
        /// Calling start() adds a new process into the pipeline and executes the first stage of the process within the current cycle.
        START_IMMEDIATELY                = 0x0100,
        /// Request interface when starting a process.
        START_REQUEST_IFACE              = 0x0200,
        /// Do not clone the message parameter.
        START_DONT_CLONE_MESSAGE         = 0x0400,
        /// Do not update the interface coverage.
        START_DONT_UPDATE_IFACE_COVERAGE = 0x0800,
        /// Start process at the end of the current cycle.
        START_THIS_CYCLE                 = 0x1000,

        //---------------------------------------------------------------------
        // System Start Modes
        //---------------------------------------------------------------------

        /// Set the listener flag of the process.
        _START_LISTENER                  = 0x2000,

        //---------------------------------------------------------------------
        // Start Modes for Particular Processes
        //---------------------------------------------------------------------

        _START_MODE_OF_LISTENER = (  START_DONT_UPDATE_IFACE_COVERAGE
                                  |  START_DONT_CLONE_MESSAGE
                                  |  START_THIS_CYCLE
                                  | _START_LISTENER
                                  | LOWEST_PRIORITY )

    };

    __CPPTESK_DEFINE_IFACE_COVERAGE((std::string(model_t::name()) + " Interface Coverage").c_str());
    cpptesk::tracer::coverage::CoverageTracker coverageTracker;

    Model();
    virtual ~Model();

    void registerComponent(const Component &component);

    void registerInterfaceAdapter(Interface &iface, process_t adapter, const std::string &name);

    void registerInputAdapter(Interface &input, process_t serializer, const std::string &name)
        { registerInterfaceAdapter(input, serializer, name); }

    void registerOutputAdapter(Interface &output, process_t deserializer, const std::string &name)
        { registerInterfaceAdapter(output, deserializer, name); }

    void registerInterfaceAdapter(Interface &iface, model_t &model, process_t adapter, const std::string &name);

    void registerInputAdapter(Interface &input, model_t &model, process_t adapter, const std::string &name)
        { registerInterfaceAdapter(input, model, adapter, name); }

    void registerOutputAdapter(Interface &output, model_t &model, process_t adapter, const std::string &name)
        { registerInterfaceAdapter(output, model, adapter, name); }

public:

///////////////////////////////////////////////////////////////////////////////
// Process creation and destruction
///////////////////////////////////////////////////////////////////////////////

    /// Starts a new process for a given stimulus.
    pid_t start(stimulus_t stimulus, Interface &iface, Message &message, Process *parent, int mode = START_NORMAL);
    /// Starts a new process for a given stimulus with the empty message.
    pid_t start(stimulus_t stimulus, Interface &iface, Process *parent, int mode = START_NORMAL);
    /// Starts a new process for a given stimulus.
    pid_t start(stimulus_t stimulus, Interface &iface, Message &message, int mode = START_NORMAL);
    /// Starts a new process for a given stimulus with the empty message.
    pid_t start(stimulus_t stimulus, Interface &iface, int mode = START_NORMAL);
    /// Fully equivalent to calling stimulus(process, iface, message).
    pid_t start(stimulus_t stimulus, Process &process, Interface &iface, Message &message = CPPTESK_DEFAULT_MESSAGE);
    /// Stops a process.
    void stop(pid_t pid);

///////////////////////////////////////////////////////////////////////////////
// System clock
///////////////////////////////////////////////////////////////////////////////
public:
    /// Returns the current time of the model.
    cycle_t time() const { return _clock; }
    /// Executes a single process quantum (performs one context switch).
    void quantum();
    /// Executes a single cycle (executes current stages of all active processes).
    void cycle();
    /// Executes several cycles.
    void delay(cycle_t n);
    /// Returns true iff the model is processing a cycle.
    bool inside_cycle() const { return _inside_cycle; }

private:
    void start_cycle();
    void end_cycle();

///////////////////////////////////////////////////////////////////////////////
// Event history
///////////////////////////////////////////////////////////////////////////////
public:

    /// Registers the stimulus.
    event_ptr_t registerStimulus(const Stimulus &stimulus);
    event_ptr_t registerStimulus(Process &process, const Stimulus &stimulus);
    /// Registers the reaction.
    event_ptr_t registerReaction(const Reaction &reaction);
    event_ptr_t registerReaction(Process &process, const Reaction &reaction);

    /**
     * Returns the first error (if there is any) having occured starting from the previous call of verdict().
     * Do not clear the set of reactions.
     */
    std::string error(error_mode_t mode = ALL_ERRORS) const;
    std::string error(int mode) const;

    /**
     * Returns a verdict for a set of reactions having occured starting from the previous call.
     * Clears the set of reactions to be analyzed.
     */
    bool verdict();

protected:
    CPPTESK_DECLARE_PROCESS(_add_reaction_to_impl_reactions);
    EventHistory _history;

////////////////////////////////////////////////////////////////////////////////
// Model Events Management (for Temporal Sequence Recognition)
////////////////////////////////////////////////////////////////////////////////

public:

    void raise(const sequence::ModelEvent &event) { _model_event_manager.add(event); }
    bool check(const sequence::ModelEvent &event) const { return _model_event_manager.contains(event); }

protected:

    sequence::ModelEventManager _model_event_manager;

////////////////////////////////////////////////////////////////////////////////
// Working with input and output interfaces
////////////////////////////////////////////////////////////////////////////////

// IMPORTANT: Should be public.
public:
#if CPPTESKHW_CHECK_VERSION(111010)
    CPPTESK_DECLARE_INOUT(CPPTESK_DEFAULT_IFACE, ::cpptesk::hw::EmptyMessage, CPPTESK_TRIVIAL_ARBITER);
#else
    /// Pseudo interface for reactions that are not assosiated with existing interfaces.
    CPPTESK_DECLARE_INOUT(CPPTESK_DEFAULT_IFACE);
#endif

protected:
    /// Processes a stimulus: serializes a message and applies it to an input interface.
    CPPTESK_DECLARE_PROCESS(onStimulus);
    /// Processes a reaction: detects a reaction, deserializes a message from an output interface and calls a comparator.
    CPPTESK_DECLARE_PROCESS(onReaction);
    /// Processes two messages and registers reactions with errors.
    void reviewMessages(::cpptesk::hw::Process &process,  ::cpptesk::hw::Interface &iface,
                        ::cpptesk::hw::Message &spec_msg, ::cpptesk::hw::Message &impl_msg, time_t &stop_time);

public:
    /// Sets the reaction waiting time.
    void setReactionTimeout(time_t timeout) { _reaction_timeout = timeout; }
    /// Returns the reaction waiting time.
    time_t getReactionTimeout(::cpptesk::hw::Interface &iface) const
        { return iface.isReactionTimeoutValid() ? iface.getReactionTimeout() : _reaction_timeout; }
    /// Sets the implementation reaction waiting time (it is used when the implementation reaction is received earlier than the model reaction).
    void setUnexpectedReactionTimeout(time_t timeout) { _unexpected_reaction_timeout = timeout; }
    /// Returns the implementation reaction waiting time (it is used when the implementation reaction is received earlier than the model reaction).
    time_t getUnexpectedReactionTimeout(::cpptesk::hw::Interface &iface) const
        { return iface.isUnexpectedReactionTimeoutValid() ? iface.getUnexpectedReactionTimeout() : _unexpected_reaction_timeout; }
    /// To maintain compatibility with test systems made earlier.
    time_t getReactionTimeout() const { return _reaction_timeout; }
    time_t getUnexpectedReactionTimeout() const { return _unexpected_reaction_timeout; }

public:

    CPPTESK_DECLARE_PROCESS(listener);
    CPPTESK_DECLARE_PROCESS(singleListener);
    bool some_unexpected_reaction_matches(CPPTESK_PROCESS_SIGNATURE());
    bool some_spec_reaction_matches(CPPTESK_PROCESS_SIGNATURE());

////////////////////////////////////////////////////////////////////////////////
// Callback methods
//
// They are not constant to allow users to use side-effect.
////////////////////////////////////////////////////////////////////////////////
protected:
    /// Precondition violation callback.
    virtual void onPreconditionViolation(Process &process);
    /// Postcondition violation callback.
    virtual void onPostconditionViolation(Process &process);
    /// Timeout violation callback.
    virtual void onTimeoutViolation(Process &process);
    /// Normal completion callback.
    virtual void onNormalCompletion(Process &process);
    /// Every cycle callback.
    virtual void onEveryCycle(Process &process);

    /// Every cycle callbacks.
    virtual void onCycleBegin() {}
    virtual void onEveryCycle() {}
    virtual void onCycleEnd()   {}

    /// Invariant checking callback.
    virtual bool invariant() { return true; }

    /// User condition for forced termination of simulation.
    virtual bool isExitCondition() { return false; }

////////////////////////////////////////////////////////////////////////////////
// Synchronization methods
//
// Most of them are virtual, because they can be overloaded in a model adapter.
// The typical use-case, however, is to implement them in a synchronizer class.
////////////////////////////////////////////////////////////////////////////////
public:
    /// Initializes simulation using command line parameters.
    void initialize(int argc, char **argv);
    /// Initializes simulation.
    virtual void initialize();
    /// Finalizes simulation.
    virtual void finalize();
    /// Inputs synchronizer: sets input signals of a design.
    virtual void setInputs()  { if(_synchronizer) { _synchronizer->setInputs();  } }
    /// Outputs synchronizer: gets output signals of a design.
    virtual void getOutputs() { if(_synchronizer) { _synchronizer->getOutputs(); } }
    /// Thread synchronizer: synchronizes execution threads of a model and simulator.
    virtual void simulate()   { if(_synchronizer) { _synchronizer->simulate();   } }
    /// Registers the synchronizer.
    void setSynchronizer(Synchronizer *synchronizer) { _synchronizer = synchronizer; }

private:
    /// The synchronizer.
    Synchronizer *_synchronizer;

    /// Simulation start time.
    time_t _sim_start_time;
    /// Simulation end time.
    time_t _sim_end_time;

protected:
////////////////////////////////////////////////////////////////////////////////
// Data
////////////////////////////////////////////////////////////////////////////////

    typedef std::vector<Component*> components_list_t;
    components_list_t _components;

    class ProcessWrapper
    {
    public:
        ProcessWrapper(model_t &model, process_t process, const std::string name):
            _model(model), _process(process), _name(name) {}

        /// Start the adapter sequentially
        CPPTESK_DECLARE_NON_VIRTUAL_PROCESS(call)
        {
            // Since it is an internal call, it is transparent to an user
            Process::CallDescriptor call = process.call_stack_top();
	    call.name = _name;
            _model.prepare_call(call);
            _model.start(_process, CPPTESK_GET_PARAMETERS());
        }

        /// Start the adapter in a new process
        pid_t call_parallel(Interface &iface, Message &message)
        {
            // Hide the call
            _model._call.reset();
            return _model.start(_process, iface, message, (START_IMMEDIATELY | START_DONT_CLONE_MESSAGE));
        }

    private:
        model_t &_model;
        process_t _process;
        std::string _name;
    };

    typedef std::map<Interface*, ProcessWrapper*> adapters_map_t;
    adapters_map_t _adapters;

    cycle_t _reaction_timeout;
    cycle_t _unexpected_reaction_timeout;

    cycle_t _clock;
    bool _inside_cycle;

public:

    // TODO:
    std::vector<Process::PointDescriptor> getPipelineSlice() const
    {
        std::vector<Process::PointDescriptor> slice; //(_process_table.size());
	/* TODO:

        for(process_table_t::const_iterator i = _process_table.begin(); i != _process_table.end(); i++)
            { slice.push_back(i->second->point);  }
        */
        return slice;
    }

////////////////////////////////////////////////////////////////////////////////
// Command-Line Parameters
////////////////////////////////////////////////////////////////////////////////
private:

    ModelParameters _parameters;
};

template<class UserModel, class UserModelName>
Model<UserModel, UserModelName>::Model():
    CPPTESK_DEFAULT_IFACE("<default>"),
    _synchronizer(0),
    _reaction_timeout(DEFAULT_REACTION_TIMEOUT),
    _unexpected_reaction_timeout(DEFAULT_UNEXPECTED_REACTION_TIMEOUT),
    _clock(0),
    _inside_cycle(false)
{
    // Register the default interface
    CPPTESK_ADD_IFACE(CPPTESK_DEFAULT_IFACE);
}

template<class UserModel, class UserModelName>
Model<UserModel, UserModelName>::~Model()
{
    // Delete the interface adapters
    for(typename adapters_map_t::iterator i = _adapters.begin(); i != _adapters.end(); i++)
        { delete i->second; }
    _adapters.clear();

    // Delete the synchronizer
    if(_synchronizer)
        { delete _synchronizer; }
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::registerComponent(const Component &component)
{
    CPPTESK_DEBUG_PRINT(DEBUG_INFO, "Register component " << component.getName());
    _components.push_back(const_cast<Component*>(&component));
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::registerInterfaceAdapter(Interface &iface, process_t adapter, const std::string &name)
{
    registerInterfaceAdapter(iface, dynamic_cast<model_t&>(*this), adapter, name);
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::registerInterfaceAdapter(Interface &iface, model_t &model, process_t adapter, const std::string &name)
{
    typename adapters_map_t::iterator i = _adapters.find(&iface);

    // Deallocate the previously registered adapter
    if(i != _adapters.end() && !(i->second))
        { delete i->second; }	

    // Register the new one
    _adapters[&iface] = new ProcessWrapper(model, adapter, name);
}

///////////////////////////////////////////////////////////////////////////////
// Process creation and descruction
///////////////////////////////////////////////////////////////////////////////

template<class UserModel, class UserModelName>
pid_t Model<UserModel, UserModelName>::start(stimulus_t stimulus, Interface &iface, Message &message, Process *parent, int mode)
{
    const unsigned type = 0;
    const unsigned priority = get_real_priority(mode & __CPPTESK_PRIORITY_MASK);
    const bool clone_message = (mode & START_DONT_CLONE_MESSAGE) == 0;

    Process::Attributes attrs(type, priority, clone_message);

    Process &process = make_process(*this, (ProcessTable::method_t)stimulus, iface, message, attrs);

    mode &= ~__CPPTESK_PRIORITY_MASK;

    if((mode & START_REQUEST_IFACE) != 0 && iface.isInput())
        { iface.request(process, *(process.input_message)); }

    if((mode & START_DONT_UPDATE_IFACE_COVERAGE) == 0)
        { coverageTracker << __CPPTESK_IFACE_COVERAGE(&iface); }

    if((mode & _START_LISTENER) != 0)
        { process.listener = true; }

    CPPTESK_DEBUG_PRINT(DEBUG_TOOL, "Setting cause for " << process);

    if(parent)
        { process.caused_by(*parent); }
    else
        { process.set_cause(new Stimulus(iface, message, time(), time())); }

    if((mode & START_IMMEDIATELY) != 0)
        { start_immediately(process); }
    else if((mode & START_THIS_CYCLE) != 0)
        { start_this_cycle(process); }
    else
        { start_next_cycle(process); }

    return process.get_pid();
}

template<class UserModel, class UserModelName>
pid_t Model<UserModel, UserModelName>::start(stimulus_t stimulus, Interface &iface, Message &message, int mode)
{
    return start(stimulus, iface, message, 0, mode);
}

template<class UserModel, class UserModelName>
pid_t Model<UserModel, UserModelName>::start(stimulus_t stimulus, Interface &iface, Process *parent, int mode)
{
    return start(stimulus, iface, CPPTESK_DEFAULT_MESSAGE, parent, mode);
}

template<class UserModel, class UserModelName>
pid_t Model<UserModel, UserModelName>::start(stimulus_t stimulus, Interface &iface, int mode)
{
    return start(stimulus, iface, 0, mode);
}

template<class UserModel, class UserModelName>
pid_t Model<UserModel, UserModelName>::start(stimulus_t stimulus, Process &process, Interface &iface, Message &message)
{
    process.prepare_call(get_call()); reset_call();

    ((UserModel*)this->*stimulus)(process, iface, message);

    return process.get_pid();
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::stop(pid_t pid)
{
    kill_process(get_process(pid));
}

///////////////////////////////////////////////////////////////////////////////
// System clock
///////////////////////////////////////////////////////////////////////////////

template <class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::start_cycle()
{
    CPPTESK_ASSERTION(!_inside_cycle, "Cannot start a new cycle inside an other one");

    // Forced termination of simulation
    if((_parameters.exit_at_cycle && _clock >= _parameters.exit_at_cycle_value) ||
       (_parameters.exit_if_condition && isExitCondition()))
    {
        ::aspectrace_finishTracing();
        ::exit(0);
    }

    _inside_cycle = true;

    // Update the clock
    _clock++;

    // Invoke the callbacks
    onCycleBegin();
    onEveryCycle();

    // Check the model's invariant
    CPPTESK_ASSERTION(invariant(), "Model invariant violation");

    // Process the model event manager (reset the set of events)
    _model_event_manager.cycle();

    // Process the registered components
    for(components_list_t::iterator i = _components.begin(); i != _components.end(); i++)
        { (*i)->cycle(); }
}

template <class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::end_cycle()
{
    // Set input signals
    setInputs();
    // Simulate a cycle
    simulate();
    // Invoke the callback
    onCycleEnd();
    // Get output signals
    getOutputs();

    _inside_cycle = false;
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::quantum()
{
    if(!inside_cycle())
        { start_cycle(); }

    if(execute(time(), EXECUTE_QUANTUM) != time())
        { end_cycle(); }
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::cycle()
{
    start_cycle();
    execute(time(), EXECUTE_CYCLE);
    end_cycle();
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::delay(cycle_t n)
{
    for(cycle_t i = 0; i < n; i++)
        { cycle(); }
}

////////////////////////////////////////////////////////////////////////////////
// Working with input and output interfaces
////////////////////////////////////////////////////////////////////////////////

template<class UserModel, class UserModelName>
CPPTESK_PROCESS Model<UserModel, UserModelName>::onStimulus(CPPTESK_PROCESS_SIGNATURE())
{
    CPPTESK_START_PROCESS();

    time_t start_time = time();

    CPPTESK_DEBUG_PRINT(DEBUG_INFO, "[STIMULUS:START]" << __CPPTESK_COMMON_START_INFO(start_time) << ": " << CPPTESK_GET_MESSAGE());

    // If a model does not have interface adapters, then do nothing
    if(_adapters.empty())
    {
        // Do nothing
    }
    else
    {
        // Otherwise, try to find an appropriate one
        typename adapters_map_t::iterator i = _adapters.find(&CPPTESK_GET_IFACE());

        // Ensure that an interface adapter has been found
        CPPTESK_ASSERTION(i != _adapters.end(), "Adapter has not been found for interface " << CPPTESK_GET_IFACE().getName());
        CPPTESK_ASSERTION(i->second != 0, "Adapter is null");

        // Call the serializer having been found
        (i->second->call)(CPPTESK_GET_PARAMETERS());

        // Release the interface if it has not been freed
        if(!CPPTESK_GET_IFACE().isReady() && CPPTESK_GET_IFACE().getOwner() == &CPPTESK_GET_PROCESS())
        {
	    CPPTESK_DEBUG_PRINT(DEBUG_WARN, "Interface " << CPPTESK_GET_IFACE().getName() << " has not been released");
            CPPTESK_RELEASE_IFACE();
        }
    }

    time_t stop_time = time();

    Stimulus stimulus(CPPTESK_GET_IFACE(), start_time, stop_time);
    registerStimulus(CPPTESK_GET_PROCESS(), stimulus);

    CPPTESK_DEBUG_PRINT(DEBUG_INFO, "[STIMULUS:STOP]" << __CPPTESK_COMMON_STOP_INFO(start_time, stop_time));

    CPPTESK_STOP_PROCESS();
}

template<class UserModel, class UserModelName>
CPPTESK_PROCESS Model<UserModel, UserModelName>::onReaction(CPPTESK_PROCESS_SIGNATURE())
{
    CPPTESK_START_PROCESS();

    time_t start_time = time();
    time_t stop_time  = time();

    const char *REACTION_START = CPPTESK_GET_PROCESS().listener ? "[LISTENER:START]" : "[REACTION:START]";
    const char *REACTION_STOP  = CPPTESK_GET_PROCESS().listener ? "[LISTENER:STOP]"  : "[REACTION:STOP]";

    CPPTESK_DEBUG_PRINT(DEBUG_INFO, REACTION_START << __CPPTESK_COMMON_START_INFO(start_time) << ": " << CPPTESK_GET_MESSAGE());

    // Emulate the data transmission
    if(CPPTESK_GET_MESSAGE().getPriority() != NULL_PRIORITY)
        CPPTESK_SET_PRIORITY(CPPTESK_GET_MESSAGE().getPriority());

    if((CPPTESK_GET_MESSAGE().getInterlock() & LOCK_HOLD)  != 0)
        { CPPTESK_WAIT_PROCESS_FINISHED(CPPTESK_GET_MESSAGE().getPID()); }
    if((CPPTESK_GET_MESSAGE().getInterlock() & LOCK_DELAY) != 0)
        { CPPTESK_DELAY(CPPTESK_GET_MESSAGE().getDelay()); }

    // If a model does not have interface adapters,
    // then register a normal (correct) reaction
    if(_adapters.empty())
    {
        Reaction reaction(NORMAL_REACTION, CPPTESK_GET_IFACE(), CPPTESK_GET_MESSAGE(), start_time, stop_time, "");
        CPPTESK_DEBUG_PRINT(DEBUG_INFO, reaction);
        registerReaction(CPPTESK_GET_PROCESS(), reaction);
    }
    else
    {
#ifndef CPPTESK_CLONE_INPUT_MESSAGE
        // The default handler requies clonning of input messages
        CPPTESK_ASSERTION(false, "Default reaction handler does not support non-cloning messages");
#endif
        // Otherwise (there are interface adapters), try to find an appropriate one
        typename adapters_map_t::iterator i = _adapters.find(&CPPTESK_GET_IFACE());

        // Ensure that an interface adapter has been found
        CPPTESK_ASSERTION(i != _adapters.end(), "Adapter has not been found for interface " << CPPTESK_GET_IFACE().getName());
        CPPTESK_ASSERTION(i->second != 0, "Adapter is null");

        Message &spec_msg =  (CPPTESK_GET_MESSAGE());
        Message &impl_msg = *(CPPTESK_GET_MESSAGE().clone());

        // Invalidate the message's fields to be able to check the deserializer
        impl_msg.invalidateFields();
        impl_msg.setUnexpected(false);
        // Call the deserializer having been found
        (i->second->call)(CPPTESK_GET_PROCESS(), CPPTESK_GET_IFACE(), impl_msg /* input/output */);

        reviewMessages(CPPTESK_GET_PROCESS(), CPPTESK_GET_IFACE(), spec_msg, impl_msg, stop_time);

        // Delete the clone
        delete &impl_msg;
    }

    CPPTESK_DEBUG_PRINT(DEBUG_INFO, REACTION_STOP << __CPPTESK_COMMON_STOP_INFO(start_time, stop_time));

    CPPTESK_STOP_PROCESS();
}

template<class UserModel, class UserModelName>
CPPTESK_PROCESS Model<UserModel, UserModelName>::reviewMessages(::cpptesk::hw::Process &process,  ::cpptesk::hw::Interface &iface,
                                                                ::cpptesk::hw::Message &spec_msg, ::cpptesk::hw::Message &impl_msg,
                                                                time_t &stop_time)
{
    std::stringstream buffer;
    bool valid_reaction = true;
    error_type_t error = NORMAL_REACTION;
    time_t reaction_start_time = time();
    stop_time = time();

    if(CPPTESK_GET_PROCESS().listener)
    {
        if(impl_msg.isUnexpected() && !(CPPTESK_GET_IFACE().getFailureIgnoreTypes() & ::cpptesk::hw::UNEXPECTED_REACTION))
        {
            stop_time = reaction_start_time = time() - 1;
            error = ::cpptesk::hw::UNEXPECTED_REACTION;
            buffer << "Unexpected reaction: " << impl_msg;
        }
        else
        {
            // Normal reactions and masked reactions from listeners are skipped
            valid_reaction = false;
        }
     }
     else if(CPPTESK_GET_PROCESS().timeout)
     {
         if(!(CPPTESK_GET_IFACE().getFailureIgnoreTypes() & MISSING_REACTION))
         {
             error = MISSING_REACTION;
             buffer << "Missing reaction: " << spec_msg;
         }
         else
         {
            // Masked reactions are skipped
            valid_reaction = false;
         }
     }
     else
     {
         std::string info = spec_msg.compare(impl_msg);

         if(!info.empty())
         {
             if(!(CPPTESK_GET_IFACE().getFailureIgnoreTypes() & INCORRECT_REACTION))
             {
                 error = INCORRECT_REACTION;
                 buffer << "Incorrect reaction: " << info;
             }
             else
             {
                // Masked reactions are skipped
                valid_reaction = false;
             }
         }
     }

     if(error != NORMAL_REACTION)
     {
         std::string explanation = spec_msg.getExplanation();

         if(!explanation.empty())
             { buffer << ": " << explanation; }
     }

     if(valid_reaction)
     {
         // Register the reaction
         Reaction reaction(error, CPPTESK_GET_IFACE(), spec_msg, impl_msg, reaction_start_time, stop_time, buffer.str());

         CPPTESK_DEBUG_PRINT((error != NORMAL_REACTION ? DEBUG_ERROR : DEBUG_INFO), reaction);

         if(error != MISSING_REACTION)
             { CPPTESK_DEBUG_PRINT(DEBUG_INFO, impl_msg); }

         registerReaction(process, reaction);
     }
}

// listener has the highest priority
template<class UserModel, class UserModelName>
CPPTESK_PROCESS Model<UserModel, UserModelName>::listener(CPPTESK_PROCESS_SIGNATURE())
{
    CPPTESK_START_PROCESS();
    CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_TOOL, "[LISTENER:START] iface=" << CPPTESK_GET_IFACE());

    while(true)
    {
        prepare_call(::cpptesk::hw::Process::CallDescriptor("<cpptesk_unexpected_reaction_listener>"));
        start(&Model<UserModel, UserModelName>::onReaction, CPPTESK_GET_IFACE(), CPPTESK_GET_MESSAGE(), __CPPTESK_PROCESS_PTR(), _START_MODE_OF_LISTENER);

        CPPTESK_WAIT(CPPTESK_GET_IFACE().isDetectionFinished());
        CPPTESK_GET_IFACE().setDetectionFinished(false);
        CPPTESK_CYCLE(); // to prevent two singleListeners from running at the same cycle
    }

    CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_TOOL, "[LISTENER:STOP] iface=" << CPPTESK_GET_IFACE());
    CPPTESK_STOP_PROCESS();
}

template<class UserModel, class UserModelName>
CPPTESK_PROCESS Model<UserModel, UserModelName>::_add_reaction_to_impl_reactions(CPPTESK_PROCESS_SIGNATURE())
{
    CPPTESK_UNUSED_ARG(CPPTESK_GET_PROCESS());
    Reaction impl_reaction = Reaction(UNEXPECTED_REACTION, CPPTESK_GET_IFACE(), CPPTESK_GET_MESSAGE(), time(), time(), "");
    CPPTESK_GET_IFACE().impl_reactions.push_back(impl_reaction);
    CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_INFO, "Add possibly unexpected reaction to " << CPPTESK_GET_IFACE() << ".impl_reactions: " << impl_reaction);
}

template<class UserModel, class UserModelName>
CPPTESK_PROCESS Model<UserModel, UserModelName>::singleListener(CPPTESK_PROCESS_SIGNATURE())
{
    CPPTESK_START_PROCESS();

    CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_TOOL, "[SINGLE_LISTENER:START] iface=" << CPPTESK_GET_IFACE());

    if(!CPPTESK_GET_IFACE().isReactionTaken(time()))
    {
        //time_t stop_time = time();

        _add_reaction_to_impl_reactions(CPPTESK_GET_PARAMETERS());

        CPPTESK_DELAY(getUnexpectedReactionTimeout(CPPTESK_GET_IFACE()));

        for(typename Interface::reactions_t::iterator i = CPPTESK_GET_IFACE().impl_reactions.begin(); i != CPPTESK_GET_IFACE().impl_reactions.end(); i++)
        {
            Reaction &r = *i;

            if(CPPTESK_GET_MESSAGE().compare(*(r.what())) == Message::COMPARE_OK)
            {
                if(some_spec_reaction_matches(CPPTESK_GET_PARAMETERS()))
                {
                    CPPTESK_GET_IFACE().takeOldReaction();
                    break;
                }

                CPPTESK_DEBUG_PRINT(DEBUG_INFO, "Unexpected reaction has been detected at cycle=" << std::dec << (r.when() - 1) <<
                                                "; its unexpectation was proven " << getUnexpectedReactionTimeout(CPPTESK_GET_IFACE()) <<
                                                " cycles later: " << std::hex << ": " << *(r.what()));

                CPPTESK_GET_PROCESS().listener = true;

                // Increment number of reactions
                CPPTESK_GET_IFACE().incrementEventNumber();
                CPPTESK_GET_MESSAGE().setUnexpected(true);

                //reviewMessages(CPPTESK_GET_PROCESS(), CPPTESK_GET_IFACE(), *(Message*)(r.what()), CPPTESK_GET_MESSAGE(), stop_time); //use reviewMessages in onReaction

                break;
            }
        }
    }

    CPPTESK_DEBUG_PRINT(DEBUG_TOOL, "[SINGLE_LISTENER:STOP] iface=" << CPPTESK_GET_IFACE());
    CPPTESK_STOP_PROCESS();
}

template<class UserModel, class UserModelName>
bool Model<UserModel, UserModelName>::some_unexpected_reaction_matches(CPPTESK_PROCESS_SIGNATURE())
{
    ::cpptesk::hw::Process &_process = CPPTESK_GET_PROCESS();
    ::cpptesk::hw::Interface &_iface = CPPTESK_GET_IFACE();
    ::cpptesk::hw::Message &_message = CPPTESK_GET_MESSAGE();

    /* Check the unexpected reaction storage where a listener puts found unexpected reactions which have not violated impl reaction timeout */
    if(!_iface.impl_reactions.empty()) {
        for(typename Interface::reactions_t::iterator i = _iface.impl_reactions.begin(); i != _iface.impl_reactions.end(); i++) {
            Reaction &r = *i;
            /* t_sr - first point of specification message, delta_sr - timeout of specification message
               t_ir - first point of implementation message, delta_it - timeout of implementation message */
            time_t t_sr = _process.start_time(), delta_sr = getReactionTimeout(_iface),
                   t_ir = r.when(), delta_ir = getUnexpectedReactionTimeout(_iface);

            /* If the current model reaction is the same as some impl reaction having been put into the storage earlier */
            if(_message.compare(*(r.what())) == cpptesk::hw::Message::COMPARE_OK &&
               /* SpecReaction   --*****----  ----*****--  ---***---  --*****--
                  ImplReaction   ----*****--  --*****----  --*****--  ---***--- */
               ((t_ir <= t_sr + delta_sr && t_sr + delta_sr <= t_ir + delta_ir) || (t_ir <= t_sr && t_sr <= t_ir + delta_ir) ||
                (t_ir <= t_sr && t_ir + delta_ir >= t_sr + delta_sr) || (t_ir >= t_sr && t_ir + delta_ir <= t_sr + delta_sr))) {
               /* the deserializer nevertheless uses iface arbiter to guarantee the fact that the spec and impl reactions are the same */
                if(_iface.chooseRequest(_process, *(r.what()))) {
                /* If so, the deserializer can remove the possibly unexpected reaction from the storage
                   as otherwise it would have been sent to the oracle as "proven to be unexpected" in singleListener() a bit later. */
                    _iface.impl_reactions.erase(i);
                    return true;
                }
            }
        }
    }

    return false;
}

template<class UserModel, class UserModelName>
bool Model<UserModel, UserModelName>::some_spec_reaction_matches(CPPTESK_PROCESS_SIGNATURE())
{
    CPPTESK_UNUSED_ARG(CPPTESK_GET_PROCESS());
    ::cpptesk::hw::Interface &_iface = CPPTESK_GET_IFACE();
    ::cpptesk::hw::Message &_message = CPPTESK_GET_MESSAGE();

    if(!_iface.spec_reactions.empty()) {
        for(typename Interface::reactions_t::iterator i = _iface.spec_reactions.begin(); i != _iface.spec_reactions.end(); i++) {
            Reaction &r = *i;
            if(_message.forceCompare(*(r.what())) == ::cpptesk::hw::Message::COMPARE_OK) {
                _iface.spec_reactions.erase(i);
                return true;
            }
        }
    }

    return false;
}

////////////////////////////////////////////////////////////////////////////////
// Stimuli & Reactions
////////////////////////////////////////////////////////////////////////////////

template<class UserModel, class UserModelName>
event_ptr_t Model<UserModel, UserModelName>::registerStimulus(const Stimulus &stimulus)
{
    return _history.register_stimulus(stimulus);
}

template<class UserModel, class UserModelName>
event_ptr_t Model<UserModel, UserModelName>::registerStimulus(Process &process, const Stimulus &stimulus)
{
    event_ptr_t stimulus_ptr = _history.register_stimulus(stimulus);
    process.add_effect(stimulus_ptr);
    return stimulus_ptr;
}

template<class UserModel, class UserModelName>
event_ptr_t Model<UserModel, UserModelName>::registerReaction(const Reaction &reaction)
{
    return _history.register_reaction(reaction);
}

template<class UserModel, class UserModelName>
event_ptr_t Model<UserModel, UserModelName>::registerReaction(Process &process, const Reaction &reaction)
{
    event_ptr_t reaction_ptr = _history.register_reaction(reaction);
    process.add_effect(reaction_ptr);
    return reaction_ptr;
}

template <class UserModel, class UserModelName>
std::string Model<UserModel, UserModelName>::error(error_mode_t mode) const
{
    return _history.error(mode);
}

template <class UserModel, class UserModelName>
std::string Model<UserModel, UserModelName>::error(int mode) const
{
    return error((error_mode_t)mode);
}

/**
 * Returns a verdict for a set of reactions having been occured starting from the previous call. 
 */
template <class UserModel, class UserModelName>
bool Model<UserModel, UserModelName>::verdict()
{
    return _history.verdict();
}

////////////////////////////////////////////////////////////////////////////////
// Callbacks
////////////////////////////////////////////////////////////////////////////////

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::onPreconditionViolation(Process &process)
{
    std::stringstream buffer;
    buffer << "(" << process.get_point().file << ":" << process.get_point().line << ")";

    Reaction reaction(ASSERTION_VIOLATION, CPPTESK_DEFAULT_IFACE, CPPTESK_DEFAULT_MESSAGE, time(), time(), buffer.str());
    registerReaction(reaction);
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::onPostconditionViolation(Process &process)
{
    std::stringstream buffer;
    buffer << "(" << process.get_point().file << ":" << process.get_point().line << ")";

    Reaction reaction(ASSERTION_VIOLATION, CPPTESK_DEFAULT_IFACE, CPPTESK_DEFAULT_MESSAGE, time(), time(), buffer.str());
    registerReaction(reaction);
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::onTimeoutViolation(Process &process)
{
    CPPTESK_UNUSED_ARG(process);
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::onNormalCompletion(Process &process)
{
    CPPTESK_UNUSED_ARG(process);
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::onEveryCycle(Process &process)
{
    CPPTESK_UNUSED_ARG(process);
}

////////////////////////////////////////////////////////////////////////////////
// Synchronization methods
////////////////////////////////////////////////////////////////////////////////

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::initialize(int argc, char **argv)
{
    initialize();

    _parameters.init(argc, argv);
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::initialize()
{
    _sim_start_time = ::time(NULL);

    if(_synchronizer) { _synchronizer->initialize(); }
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::finalize()
{
    _sim_end_time = ::time(NULL);

    if(_synchronizer) { _synchronizer->finalize(); }

    // Register missing reactions
    for(components_list_t::iterator i = _components.begin(); i != _components.end(); i++)
    {
        Interface *iface = dynamic_cast<Interface*>(*i);

        if(iface != 0 && iface->isOutput())
        {
            Arbiter *arbiter = iface->getArbiter();

            if(arbiter != 0)
            {
                // Get specification reactions for which implementation reactions are expected
                for(requests_t::iterator i = arbiter->begin(); i != arbiter->end(); i++)
                {
                    Reaction reaction(MISSING_REACTION, *iface, *(i->message), time(), time(), "");
                    registerReaction(reaction);
                }
            }
        }
    }

    // Append all reactions into the history
    _history.reset_current_set();
    // Print diagnosis
    _history.trace_diagnosis();
    // Print statistics on stimuli and reactions
    _history.trace_statistics();

    // Print simulation time and frequency
    uint64_t cycle_num = time();
    double   frequency = (double)cycle_num / (double)(_sim_end_time - _sim_start_time);

    std::string cycles = "cycles";
    std::string herz   = "Hz";

    if(frequency > 1000.0) { frequency /= 1000.0; herz = "KHz"; }
    if(frequency > 1000.0) { frequency /= 1000.0; herz = "MHz"; }
    if(frequency > 1000.0) { frequency /= 1000.0; herz = "G"; }

    CPPTESK_PRINT_SEPARATOR();
    CPPTESK_DEBUG_STREAM() << std::fixed << std::setprecision(2) << std::dec;
    CPPTESK_DEBUG_STREAM() << "Simulation time: " << cycle_num << " " << cycles << std::endl;
    CPPTESK_DEBUG_STREAM() << "Simulation freq: " << frequency << " " << herz   << std::endl;
}

CPPTESK_MODEL(EMPTY_MODEL) {};

}} // namespace cpptesk::hw
