/*
 * Copyright 2012 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Process context switcher
///////////////////////////////////////////////////////////////////////////////

#include <hw/process/context_switcher.hpp>

namespace cpptesk {
namespace hw {

void DefaultContextSwitcher::get_context(Process &process)
{
    // Allocate the process context.
    ucontext_t *context = (ucontext_t*)::malloc(sizeof(ucontext_t));
    CPPTESK_ASSERTION(context != NULL, "No memory to create a process context");

    ::getcontext(context);

    // Allocate the process stack.
    context->uc_stack.ss_size = _PROCESS_STACK_SIZE;
    context->uc_stack.ss_sp = (char *)::malloc(_PROCESS_STACK_SIZE);
    CPPTESK_ASSERTION(context->uc_stack.ss_sp != NULL, "No memory to create a process stack");

    set_process_context(process, context);
}

void DefaultContextSwitcher::make_context(BaseModel &model, method_t method, Process &process, Interface &iface, Message &message)
{
     get_context(process);

     ucontext_t *context = get_process_context<ucontext_t>(process);

    // The arguments structure is destroyed in the process wrapper.
    _MethodArguments *args = new _MethodArguments(*this, model, method, process, iface, message);

    _created++;

    // The makecontext() function supports only integer arguments.
    int addr_hi = (int)(((uint64_t)args) >> 32);
    int addr_lo = (int)(((uint64_t)args) & 0xffffFFFFull);

    ::makecontext(context, (void(*)(void))DefaultContextSwitcher::_process_wrapper, 2, addr_hi, addr_lo);

    set_process_context(process, context);
}

void DefaultContextSwitcher::free_context(Process &process)
{
    ucontext_t *context = get_process_context<ucontext_t>(process);

    if(context != 0)
    {
        ::free(context->uc_stack.ss_sp);
        ::free(context);
    }

    set_process_context(process, 0);
}

void DefaultContextSwitcher::swap_context(Process &from, Process &to)
{
    ucontext_t *lhs = get_process_context<ucontext_t>(from);
    ucontext_t *rhs = get_process_context<ucontext_t>(to);

    ::swapcontext(lhs, rhs);
}

void DefaultContextSwitcher::_process_wrapper(int addr_hi, int addr_lo)
{
    uint64_t addr = ((uint64_t)addr_hi << 32) | ((uint64_t)addr_lo & 0xffffFFFFull);
    _MethodArguments *args = (_MethodArguments*)addr;

    CPPTESK_ASSERTION(args != 0, "Pointer to method arguments is null");

    BaseModel &model = args->model;
    method_t method  = args->method;
    Process &process = args->process;
    Interface &iface = args->iface;
    Message &message = args->message;

    args->switcher._deleted++;

    // The process wrapper should destroy the arguments structure.
    delete args;

    // Start the process method.
    ((model).*(method))(process, iface, message);
}

}} // namespace cpptesk::hw
