/*
 * Copyright 2012 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Process scheduler
///////////////////////////////////////////////////////////////////////////////

#include <hw/process/process_scheduler.hpp>
#include <hw/unused.hpp>

#include <algorithm>

namespace cpptesk {
namespace hw {

void ProcessScheduler::schedule_process(Process &process)
{
    // Add the process to the current-cycle set of processes.
    activate_process(process);
    process.schedule_time(_clock);
}

void ProcessScheduler::schedule_process(Process &process, cycle_t time_slot)
{
    if(time_slot != CPPTESK_INFINITY_TIME)
    {
        // Postpone the process for the better time.
        _schedule[time_slot][process.get_pid()] = &process;
    }
    
    process.schedule_time(time_slot);
}

void ProcessScheduler::stop_process(Process &process)
{
    // For each time slot in the schedule
    for(_schedule_t::iterator i = _schedule.begin(); i != _schedule.end(); i++)
    {
        // try to find the process among the processes scheduled for that slot.
        // There are usually not many time slots in the schedule, so that's ok.
        _processes_t &processes = i->second;
        _processes_t::iterator j = processes.find(process.get_pid());

        // If the process is found, remove it from the schedule.
        if(j != processes.end())
        {
            // There can be at most one occurence of one process in the schedule.
            processes.erase(j);
            break;
        }
    }
    
    process.schedule_time(CPPTESK_INFINITY_TIME);
}

cycle_t ProcessScheduler::next_time() const
{
    if(_schedule.empty())
        { return CPPTESK_INFINITY_TIME; }

    return _schedule.begin()->first;
}

cycle_t ProcessScheduler::start_cycle(cycle_t now)
{
    _clock = now;

    // IMPORTANT: It is useful facility to immediately execute processes in scenarios.
    // However, there might be abnormal sitations, when there are two ocurences of the
    // same process in the current-cycle set of processes.
#if 0 
    CPPTESK_ASSERTION(is_cycle_completed(), "Attempt to execute a process in the hole between two cycles");
#endif

    // If there is nothing to schedule, returns infinity.
    if(_schedule.empty())
        { return CPPTESK_INFINITY_TIME; }

    // Get the nearest entry of the schedule.
    _schedule_t::iterator i = _schedule.begin();
    
    cycle_t first_time_slot = i->first;
    _processes_t &processes = i->second;

    if(first_time_slot == now)
    {
        // Move all processes of that entry into the current-cycle set of processes.
        for(_processes_t::iterator j = processes.begin(); j != processes.end(); j++)
        {
            Process &process = *(j->second);

            if(!process.is_completed())
                { activate_process(process); }
        }
    
        // Remove the entry.
        _schedule.erase(i);
    }
    else
    {
        CPPTESK_ASSERTION(first_time_slot > now, "Attempt to skip a scheduled process");
    }
    
    return first_time_slot;
}

///////////////////////////////////////////////////////////////////////////////
// Default process scheduler
///////////////////////////////////////////////////////////////////////////////

bool greater_process_priority(Process *lhs, Process *rhs)
{
    unsigned lp = lhs->get_priority();
    unsigned rp = rhs->get_priority();

    // PIDs are used to sort equivalent processes in order of increasing PIDs.
    return lp > rp || (lp == rp && lhs->get_pid() < rhs->get_pid());
}

bool DefaultProcessScheduler::is_cycle_completed() const
{
    // The delta cycle is completed iff there no current-cycle processes.
    return _current_set.empty();
}

Process& DefaultProcessScheduler::next_process()
{
    if(_modified)
    {
        // Sort the current-cycle processes in order of decreasing priorities.
        std::stable_sort(_current_set.begin(), _current_set.end(), greater_process_priority);
        _modified = false;
    }

    return *_current_set.front();
}

void DefaultProcessScheduler::activate_process(Process &process)
{
    // Set the process execution cycle.
    start_process_cycle(process);

    // IMPORTANT: The set type is used to avoid two occurences of the same process.
    // Such situation is possible when the process starts in the inter-cycle hole
    // with the START_IMMEDIATELY flag. It is activated, but not erased from the
    // current-cycle set of processes. Then, it is activated for the next cycle.
    if(_check_set.find(&process) == _check_set.end())
    {
        _check_set.insert(&process);

        _current_set.push_back(&process);
        _modified = true;
    }
}

void DefaultProcessScheduler::deactivate_process(Process &process)
{
    // Remove the process from the current-cycle set of processes.
    // The process should not be far from the beginning, so the loop is ok.
    for(_process_set_t::iterator i = _current_set.begin(); i != _current_set.end(); i++)
    {
        if(process.get_pid() == (*i)->get_pid())
        {
            _current_set.erase(i);
            break;
        }
    }

    _check_set.erase(&process);
}

void DefaultProcessScheduler::on_priority_change(Process &process)
{
    CPPTESK_UNUSED_ARG(process);
    _modified = true;
}

}} // namespace cpptesk::hw
