/*
 * Copyright 2012 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Process table
///////////////////////////////////////////////////////////////////////////////

#include <hw/process/process_table.hpp>
#include <hw/unused.hpp>

#include <set>

namespace cpptesk {
namespace hw {

ProcessTable::ProcessTable():
    _clock(0),
    _pid(0),
    _main_process(-1, *this),
    _delete_process_scheduler(true),
    _process_scheduler(*(new DefaultProcessScheduler())),
    _delete_context_switcher(true),
    _context_switcher(*(new DefaultContextSwitcher())),
    _process_table(HASH_TABLE_SIZE),
    _created(0),
    _deleted(0)
{
    _make_main_process();
}

ProcessTable::ProcessTable(ProcessScheduler &process_scheduler, ContextSwitcher &context_switcher):
    _clock(0),
    _pid(0),
    _main_process(-1, *this),
    _delete_process_scheduler(false),
    _process_scheduler(process_scheduler),
    _delete_context_switcher(false),
    _context_switcher(context_switcher),
    _process_table(HASH_TABLE_SIZE),
    _created(0),
    _deleted(0)
{
    _make_main_process();
}

ProcessTable::~ProcessTable()
{
    _context_switcher.free_context(_main_process);

    kill_all();

    if(_delete_process_scheduler)
        { delete &_process_scheduler; }
    if(_delete_context_switcher)
        { delete &_context_switcher; }
}

Process& ProcessTable::make_process(BaseModel &model, method_t method,
    Interface &iface, Message &message, Process::Attributes &attrs)
{
    // IMPORTANT: A newly created process should have a greater PID than the
    // previously launched ones. This guarantees that reset of a validity signal
    // (strobe) is done before a new process set the signal.
    Process *process = new Process(_pid++, _clock, iface, message, attrs, *this);
    CPPTESK_ASSERTION(process, "No memory to create a process");

    CPPTESK_DEBUG_PRINT_FILE_LINE(DEBUG_TOOL, "ProcessTable::make_process: new process: " << *process);

    // Make the process context.
    _context_switcher.make_context(model, method, *process, *process->input_iface, *process->input_message);
    // Insert the newly created process into the table.
    _process_table.put(process->get_pid(), process);

    // Set a call descriptor.
    process->prepare_call(get_call());
    reset_call();

    _created++;

    // The process is freezed. To schedule the process execution, call schedule_process().
    return *process;
}

void ProcessTable::kill_process(Process &process)
{
    // Remove the process from the schedule.
    _process_scheduler.stop_process(process);
    // Remove the proces from the table.
    _process_table.remove(process.get_pid());
    // Release the process context.
    _context_switcher.free_context(process);

    _deleted++;

    // Deallocate the process memory.
    delete &process;
}

void ProcessTable::kill_all()
{
    std::list<Process*> massacre;

    for(HashTable<pid_t, Process*>::iterator i = _process_table.begin(); i != _process_table.end(); i++)
        { massacre.push_back(*i); }

    for(std::list<Process*>::iterator i = massacre.begin(); i != massacre.end(); i++)
        { kill_process(**i); }
}

void ProcessTable::resume_process(Process &process)
{
    Process *from;
    Process *jump;

    CPPTESK_ASSERTION(!process.is_cycle_completed(), "Cannot resume a deactivated process");
    CPPTESK_ASSERTION(!_process_stack.empty(), "There is no process to return to");

    from = _process_stack.back();
    _process_stack.push_back(&process);
    jump = _process_stack.back();

    CPPTESK_DEBUG_PRINT_FILE_LINE(DEBUG_TOOL, "ProcessTable::resume_process: switch from " << *from << " to " << *jump);

    _context_switcher.swap_context(*from, *jump);
}

void ProcessTable::suspend_process(Process &process)
{
    Process *from;
    Process *jump;

    CPPTESK_ASSERTION(_process_stack.size() > 1, "There is no process to return to");

    from = _process_stack.back();
    _process_stack.pop_back();
    jump = _process_stack.back();

    CPPTESK_ASSERTION(from == &process, "The process stack has been corrupted");
    CPPTESK_DEBUG_PRINT_FILE_LINE(DEBUG_TOOL, "Model::suspend_process: switch from " << *from << " to " << *jump);

    _context_switcher.swap_context(*from, *jump);
}

void ProcessTable::suspend_process(Process &process, cycle_t next_time_slot)
{
    schedule_process(process, next_time_slot);
    suspend_process(process);
}

cycle_t ProcessTable::execute(cycle_t now, exec_mode_t mode)
{
    if(mode == EXECUTE_CYCLE || (mode == EXECUTE_QUANTUM && _clock != now))
    {
        cycle_t time_slot = _process_scheduler.start_cycle(_clock = now);

        // If the schedule does not contain anything for the time given,
        // do nothing and return the nearest non-empty time slot.
        if(time_slot > now)
            { return time_slot; }
    }

    std::list<Process*> massacre;
    bool quantum_executed = false;

    while(!_process_scheduler.is_cycle_completed())
    {
        Process &process = _process_scheduler.next_process();

        // The process or its cycle could have been completed.
        if(!process.is_completed() && !process.is_cycle_completed())
        {
            CPPTESK_ASSERTION(process.current_time() == _clock,
	        "Process " << process  << " has not been scheduled for the current time slot: "
		"current=" << std::dec << _clock << ", process=" << process.current_time());

            quantum_executed = true;
            // Execute a process quantum.
            resume_process(process);

            // Check the execution flow consistency.
            CPPTESK_ASSERTION(_process_stack.back() == &_main_process,
                "Process " << process << " has unexpectedly finished (probably, STOP_PROCESS() is missing)");
        }

        // Kill all completed processes.
        if(process.is_completed())
            { massacre.push_back(&process); }

        // If the process cycle is fully executed,
        if(process.is_completed() || process.is_cycle_completed())
        {
            // remove the process from the current-cycle set.
            _process_scheduler.deactivate_process(process);
        }

        if(mode == EXECUTE_QUANTUM && quantum_executed)
            { break; }
    }

    for(std::list<Process*>::iterator i = massacre.begin(); i != massacre.end(); i++)
        { kill_process(**i); }

    if(mode == EXECUTE_QUANTUM)
        { return _process_scheduler.is_cycle_completed() ? _process_scheduler.next_time() : _clock; }

    return _process_scheduler.next_time();
}

void ProcessTable::_make_main_process()
{
    _context_switcher.get_context(_main_process);

    // Pseudo call of the main process.
    _main_process.prepare_call(Process::CallDescriptor("<cpptesk_main_process>"));
    _main_process.start_call();

    _process_stack.push_back(&_main_process);
}

}} // namespace cpptesk::hw

