/*
 * Copyright 2010-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Range of unsigned integer values
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <iostream>

namespace cpptesk {
namespace hw {

/**
 * Range of unsigned integer values.
 */
template<typename T, T inf = (T)-1>
struct Range
{
    static const T INF = inf;

    Range(): min(0), max(0) {}
    Range(T val): min(val), max(val) {}
    Range(T min, T max): min(min), max(max) {}

    virtual ~Range() {}

    void set(T val)
    {
        this->min = this->max = val;
    }

    void set(T min, T max)
    {
        this->min = min;
        this->max = max;
    }

    void set(const Range<T> &range)
    {
        this->min = range.min;
        this->max = range.max;
    }

    /// Returns true iff the range consists of a single point.
    bool isPoint() const { return min == max; }
    /// Returns true iff the range is [0,0].
    bool isZero() const { return min == 0 && max == 0; }
    /// Returns true iff the range is [1,1].
    bool isOne() const { return min == 1 && max == 1; }
    /// Returns true iff the range is [inf,inf].
    bool isInfinity() const { return min == INF && max == INF; }

    /// Compares two ranges.
    bool operator == (const Range &rhs) const { return min == rhs.min && max == rhs.max; }
    /// Adds two ranges.
    Range operator + (const Range &rhs) const { return Range(min + rhs.min, max + rhs.max); }

    Range& operator += (const Range &rhs)
    {
        min += rhs.min;
        max += rhs.max;
        return *this;
    }

    T min;
    T max;
};

#define __CPPTESK_PRINT_RANGE_VALUE(out, val) \
    do { \
        if(val == Range<T>::INF) \
            { out << "inf"; } \
        else \
            { out << std::dec << (val) << std::hex; } \
    } while(false)

template<typename T>
inline std::ostream& operator << (std::ostream &out, const Range<T> &range)
{
    if(range.isPoint())
    {
        __CPPTESK_PRINT_RANGE_VALUE(out, range.min);
    }
    else
    {
        out << "[";
        __CPPTESK_PRINT_RANGE_VALUE(out, range.min);
        out << ",";
        __CPPTESK_PRINT_RANGE_VALUE(out, range.max);
        out << "]";
    }

    return out;
}

}} // cpptesk::hw

