/** @file ccp_model.h
 *
 * A model of the simple write-invalidate cache coherence protocol (CCP) used to
 * illustrate the C++TESK facilities on cache coherence (CC) verification using
 * the graph exploration (state enumeration) approach. The idea of the example
 * was borrowed from F. Pong and M. Dubois' paper "Verification Techniques for
 * Cache Coherence Protocols" (ACM Computing Surveys (CSUR), Volume 29 Issue 1,
 * March 1997).
 *
 * The CCP model is rather abstract. The goal of the example is to illustrate
 * state abstraction techniques and non-deterministic graph exploration.
 *
 * @author Alexander Kamkin (kamkin@ispras.ru)
 */

#pragma once

#include <hw/model.hpp>
#include <hw/state/state.hpp>
#include <ts/coverage.hpp>
#include <tracer/tracer.hpp>

#include <sstream>
#include <vector>

using namespace cpptesk::hw;
using namespace cpptesk::hw::state;
using namespace cpptesk::ts::coverage;
using namespace cpptesk::tracer::coverage;

namespace cpptesk {
namespace examples {
namespace ccp {

////////////////////////////////////////////////////////////////////////////////
// Cache
////////////////////////////////////////////////////////////////////////////////

/**
 * A single-block cache model.
 */
class Cache
{
public:
    /// Cache states.
    enum state_t
    {
        /// Invalid (I): invalid copy.
        INVALID,
        /// Shared (S): clean copy shared with other caches.
        SHARED,
        /// Dirty (D): modified and only cached copy.
        DIRTY
    };

    /// Cache operations.
    enum operation_t
    {
        /// Read (R).
        READ,
        /// Write (W).
        WRITE,
        /// Replacement (Z).
        REPLACE
    };

    /// Cache requesters.
    enum requester_t
    {
        /// Request from this cache.
        THIS,
        /// Request from an other cache.
        OTHER
    };

    Cache(): _state(INVALID) {}
    virtual ~Cache() {}

    /// Returns the cache state.
    state_t state() const { return _state; }
    /// Resets the cache state.
    void reset() { _state = INVALID; }

    /// Processes an operation (updates the cache state).
    void process(requester_t r, operation_t o);

protected:
    /// Cache state.
    state_t _state;
};

////////////////////////////////////////////////////////////////////////////////
// Processor
////////////////////////////////////////////////////////////////////////////////

/**
 * A processor model.
 */
class Processor
{
public:
    Processor(unsigned id): _id(id) {}
    virtual ~Processor() {}

    /// Processes an operation.
    virtual void process(unsigned id, Cache::operation_t o)
        { _cache.process((id == _id ? Cache::THIS : Cache::OTHER), o); }

    /// Returns a processor state.
    virtual std::string getState() const
    {
        switch(state())
        {
        case Cache::INVALID: { return "I"; }
        case Cache::SHARED:  { return "S"; }
        case Cache::DIRTY:   { return "D"; }
        }
        return "-";
    }

    /// Reads data.
    void read   (unsigned id) { process(id, Cache::READ);    }
    /// Writes data.
    void write  (unsigned id) { process(id, Cache::WRITE);   }
    /// Replaces data.
    void replace(unsigned id) { process(id, Cache::REPLACE); }

    /// Replace precondition.
    virtual bool pre_replace(unsigned id) const
    {
        if(state() == Cache::INVALID && id == _id) { return false; }
	if(state() == Cache::DIRTY   && id != _id) { return false; }
	return true;
    }

    /// Returns the cache state.
    virtual Cache::state_t state() const { return _cache.state(); }

    /// Resets the processor state.
    virtual void reset() { _cache.reset(); }

    /// Returns the processor id.
    unsigned id() const { return _id; }

protected:
    Processor(): _id(-1u) {}

    Cache _cache;
    unsigned _id;
};

////////////////////////////////////////////////////////////////////////////////
// Symmetric Multi-Processor
////////////////////////////////////////////////////////////////////////////////

/**
 * A multi-processor model.
 */
class MultiProcessor: public Processor
{
public:
    MultiProcessor(unsigned n)
    {
        for(unsigned i = 0; i < n; i++)
        {
	    Processor processor(i);
            _processors.push_back(processor);
        }
    }

    virtual ~MultiProcessor() {}

    /// Processes an operation.
    virtual void process(unsigned id, Cache::operation_t o)
    {
        for(std::vector<Processor>::iterator i = _processors.begin(); i != _processors.end(); i++)
            { i->process(id, o); }
    }

    /// Returns a composite state.
    virtual std::string getState() const
    {
        std::stringstream buffer;

        buffer << "(";
        for(std::vector<Processor>::const_iterator i = _processors.begin(); i != _processors.end(); i++)
            { buffer << (i == _processors.begin() ? "" : ",") << i->state(); }
        buffer << ")";

        return buffer.str();
    }

    /// Replace precondition.
    virtual bool pre_replace(unsigned id) const
    {
        return _processors[id].pre_replace(id);
    }

    /// Resets the multi-processor state.
    virtual void reset()
    {
        for(std::vector<Processor>::iterator i = _processors.begin(); i != _processors.end(); i++)
	    { i->reset(); }
    }

    /// Returns the number of processors.
    size_t size() const { return _processors.size(); }
    /// Returns the vector of processors.
    const std::vector<Processor>& processors() const { return _processors; }

protected:
    std::vector<Processor> _processors;
};

////////////////////////////////////////////////////////////////////////////////
// Cache Coherence Model
////////////////////////////////////////////////////////////////////////////////

class CacheCoherenceModel: public MultiProcessor
{
public:
    ////////////////////////////////////////////////////////////////////////////
    // Abstraction technique
    ////////////////////////////////////////////////////////////////////////////

    /// State abstraction techniques.
    enum state_abstraction_t { VECTOR, MULTISET, SET };
    /// Transition abstraction techniques.
    enum trans_abstraction_t { INDEX, KIND };

    ////////////////////////////////////////////////////////////////////////////
    // Test Coverage
    ////////////////////////////////////////////////////////////////////////////
    ENUM_COV(CCP_STATE_COVERAGE, "State Coverage",
    (
        (INVALID, "Invalid"),
        (SHARED,  "Shared"),
        (DIRTY,   "Dirty")
    ));

    CCP_STATE_COVERAGE stateCoverage(Cache::state_t s) const
    {
        switch(s)
	{
	case Cache::INVALID: { return CCP_STATE_COVERAGE::INVALID; }
	case Cache::SHARED:  { return CCP_STATE_COVERAGE::SHARED;  }
	default:             { return CCP_STATE_COVERAGE::DIRTY;   }
	}
    }

    ////////////////////////////////////////////////////////////////////////////

    ENUM_COV(CCP_OPERATION_COVERAGE, "Operation Coverage",
    (
        (READ,    "Read"),
        (WRITE,   "Write"),
        (REPLACE, "Replace")
    ));

    CCP_OPERATION_COVERAGE operationCoverage(Cache::operation_t o) const
    {
        switch(o)
	{
	case Cache::READ:  { return CCP_OPERATION_COVERAGE::READ;    }
	case Cache::WRITE: { return CCP_OPERATION_COVERAGE::WRITE;   }
	default:           { return CCP_OPERATION_COVERAGE::REPLACE; }
	}
    }

    ////////////////////////////////////////////////////////////////////////////

    ENUM_COV(CCP_REQUESTER_COVERAGE, "Requester Coverage",
    (
        (THIS,    "This"),
        (OTHER,   "Other")
    ));

    CCP_REQUESTER_COVERAGE requesterCoverage(Cache::requester_t r) const
    {
        switch(r)
	{
	case Cache::THIS: { return CCP_REQUESTER_COVERAGE::THIS;  }
	default:          { return CCP_REQUESTER_COVERAGE::OTHER; }
	}
    }

    ////////////////////////////////////////////////////////////////////////////

    COMP_COV(CCP_STATExOPERATION_COVERAGE, "StatexOperation Coverage",
        CCP_STATE_COVERAGE, CCP_OPERATION_COVERAGE);

    CCP_STATExOPERATION_COVERAGE stateOperationCoverage(Cache::state_t s, Cache::operation_t o) const
        { return stateCoverage(s) * operationCoverage(o); }

    ////////////////////////////////////////////////////////////////////////////

    COMP_COV(CCP_MAIN_COVERAGE, "Cache Coherence Coverage",
        CCP_STATExOPERATION_COVERAGE, CCP_REQUESTER_COVERAGE);

    CCP_MAIN_COVERAGE mainCoverage(Cache::state_t s, Cache::operation_t o, Cache::requester_t r) const
        { return stateOperationCoverage(s, o) * requesterCoverage(r); }

    ////////////////////////////////////////////////////////////////////////////

    CoverageTracker coverageTracker;

    void trackCoverage(unsigned id, Cache::operation_t o);

    ////////////////////////////////////////////////////////////////////////////
    // Initialization/Finalization
    ////////////////////////////////////////////////////////////////////////////

    CacheCoherenceModel(unsigned n,
        state_abstraction_t state_abstraction = VECTOR,
        trans_abstraction_t trans_abstraction = INDEX): MultiProcessor(n),
        _state_abstraction(state_abstraction), _trans_abstraction(trans_abstraction) {}

    virtual ~CacheCoherenceModel() {}

    virtual void initialize() {}
    virtual void finalize()   {}

    ////////////////////////////////////////////////////////////////////////////
    // Data Integrity Condition
    ////////////////////////////////////////////////////////////////////////////

    virtual bool invariant()
    {
        const std::vector<Processor> &cpu = processors();

        bool dirty  = false;
        bool shared = false;

        for(std::vector<Processor>::const_iterator i = cpu.begin(); i != cpu.end(); i++)
        {
	    if(i->state() == Cache::DIRTY)
	    {
                // D and D are mutually exclusive
	        if(dirty) { return false; }
                dirty = true;
	    }
            if(i->state() == Cache::SHARED)
            {
                // D and S are mutually exclusive
                if(dirty) { return false; }
                shared = true;
            }
        }

        return true;
    }

    ////////////////////////////////////////////////////////////////////////////
    // State Abstraction
    ////////////////////////////////////////////////////////////////////////////

    /// Returns the state depending on the abstraction type.
    std::string getState() const;

    ////////////////////////////////////////////////////////////////////////////
    // Transition Abstraction
    ////////////////////////////////////////////////////////////////////////////

    /// Returns the number of transitions. 
    unsigned getTransitionNumber() const;
    /// Returns the processor identifier or -1.
    unsigned getProcessorId(unsigned transitionId) const;

    /// Returns the number of processors.
    unsigned getIndexNumber() const;
    /// Returns the processor identifier.
    unsigned getIndexId(unsigned transitionId) const;

    /// Returns the number of transition kinds (cache states).
    unsigned getKindNumber() const;
    /// Returns the processor identifier.
    unsigned getKindId(unsigned transitionId) const;

protected:

    /// State abstraction type.
    state_abstraction_t _state_abstraction;
    /// Transition abstraction type.
    trans_abstraction_t _trans_abstraction;
};

}}} // namespace cpptesk::examples::ccp

