#include <stdio.h>
#include <string.h>
#include <limits.h>

#include <atl/list.h>

#include "cfg_parser.h"


/******************************************************************************/
/**                                 nodes.cfg                                **/
/******************************************************************************/
/* line format:
 *  address : mcc_port, sync_in_port
 */

typedef enum NodelineParserStates {
                                BEFORE_ADDRESS, ADDRESS, AFTER_ADDRESS
                              , BEFORE_FIRST_PORT, FIRST_PORT, AFTER_FIRST_PORT
                              , BEFORE_SECOND_PORT, SECOND_PORT, AFTER_SECOND_PORT
} NodelineParserStates;

/******************************************************************************/
/* returns:
 *   0 - wrong format
 *   1 - ok
 *  -1 - empty line
 */
int parse_nodelist_line( const char *line, char *address, char *mccPort, char *inPort )
{
    int i, size;
    int startAddr = 0, endAddr = 0, startMccPort = 0, endMccPort = 0, startInPort = 0, endInPort = 0;
    NodelineParserStates state = BEFORE_ADDRESS;

    assertion( line != NULL, "parse_nodelist_line: line == NULL" );
    assertion( address != NULL, "parse_nodelist_line: address == NULL" );
    assertion( mccPort != NULL, "parse_nodelist_line: mccPort == NULL" );
    assertion( inPort != NULL, "parse_nodelist_line: inPort == NULL" );

    size = strlen( line );
    for( i=0; i < size; i++ ) {

        if( isspace(line[i]) ) {
            switch( state ) {
                case BEFORE_ADDRESS:
                case AFTER_ADDRESS:
                case BEFORE_FIRST_PORT:
                case AFTER_FIRST_PORT:
                case BEFORE_SECOND_PORT:
                case AFTER_SECOND_PORT:
                    continue;
                case ADDRESS:
                    state = AFTER_ADDRESS;
                    endAddr = i - 1;
                    break;
                case FIRST_PORT:
                    state = AFTER_FIRST_PORT;
                    endMccPort = i - 1;
                    break;
                case SECOND_PORT:
                    state = AFTER_SECOND_PORT;
                    endInPort = i - 1;
                    break;
            }
        } else if( isalpha(line[i]) || line[i] == '_' || line[i] == '-' ) {
            switch( state ) {
                case BEFORE_ADDRESS:
                    state = ADDRESS;
                    startAddr = i;
                    break;
                case ADDRESS:
                    continue;
                case AFTER_ADDRESS:
                case BEFORE_FIRST_PORT:
                case FIRST_PORT:
                case AFTER_FIRST_PORT:
                case BEFORE_SECOND_PORT:
                case SECOND_PORT:
                case AFTER_SECOND_PORT:
                    return 0; /* wrong format */
            }
        } else if( isdigit(line[i]) ) {
            switch( state ) {
                case BEFORE_ADDRESS:
                    state = ADDRESS;
                    startAddr = i;
                    break;
                case ADDRESS:
                case FIRST_PORT:
                case SECOND_PORT:
                    continue;
                case AFTER_ADDRESS:
                case AFTER_FIRST_PORT:
                case AFTER_SECOND_PORT:
                    return 0; /* wrong format */
                case BEFORE_FIRST_PORT:
                    state = FIRST_PORT;
                    startMccPort = i;
                    break;
                case BEFORE_SECOND_PORT:
                    state = SECOND_PORT;
                    startInPort = i;
                    break;
            }
        }
        else if ( line[i] == '.' ) {
            switch( state ) {
                case BEFORE_ADDRESS:
                case AFTER_ADDRESS:
                case BEFORE_FIRST_PORT:
                case FIRST_PORT:
                case AFTER_FIRST_PORT:
                case BEFORE_SECOND_PORT:
                case SECOND_PORT:
                case AFTER_SECOND_PORT:
                        return 0; /* wrong format */
                case ADDRESS:
                    continue;
            }
        } else if ( line[i] == ':' ) {
            switch( state ) {
                case BEFORE_ADDRESS:
                case BEFORE_FIRST_PORT:
                case FIRST_PORT:
                case AFTER_FIRST_PORT:
                case BEFORE_SECOND_PORT:
                case SECOND_PORT:
                case AFTER_SECOND_PORT:
                    return 0;  /* wrong format */
                case ADDRESS:
                    state = BEFORE_FIRST_PORT;
                    endAddr = i - 1;
                    break;
                case AFTER_ADDRESS:
                    state = BEFORE_FIRST_PORT;
                    break;
            }
        } else if ( line[i] == ',') {
            switch( state ) {
                case BEFORE_ADDRESS:
                case ADDRESS:
                case AFTER_ADDRESS:
                case BEFORE_FIRST_PORT:
                case BEFORE_SECOND_PORT:
                case SECOND_PORT:
                case AFTER_SECOND_PORT:
                    return 0;
                case FIRST_PORT:
                    state = BEFORE_SECOND_PORT;
                    endMccPort = i - 1;
                    break;
                case AFTER_FIRST_PORT:
                    state = BEFORE_SECOND_PORT;
                    break;
            }
        } else {
            return 0;
        }

    } /* for */
    if( state == SECOND_PORT ) {
        endInPort = size;
        state = AFTER_SECOND_PORT;
    }
    if( state == BEFORE_ADDRESS ) return -1; /* empty line */
    if( state != AFTER_SECOND_PORT ) return 0;

    memcpy( address, line + startAddr, endAddr - startAddr + 1 );
    address[endAddr - startAddr + 1] = 0;
    memcpy( mccPort, line + startMccPort, endMccPort - startMccPort + 1 );
    mccPort[endMccPort - startMccPort + 1] = 0;
    memcpy( inPort, line + startInPort, endInPort - startInPort + 1 );
    inPort[endInPort - startInPort + 1] = 0;

    return 1;
}

/******************************************************************************/
/**                                 links.cfg                                **/
/******************************************************************************/
/* line format:
 *  node : node1, node2, node3....
 */

typedef enum TopologylineParserStates {
                                BEFORE_NODE_NUMBER, NODE_NUMBER, AFTER_NODE_NUMBER
                              , BEFORE_NODE_IN_LIST, NODE_IN_LIST, AFTER_NODE_IN_LIST
} TopologylineParserStates;

/******************************************************************************/
/* returns
 *   0 - wrong format
 *   1 - ok
 *  -1 - empty line
 */
int parse_topology_line( const char *line, char *node_number, List **out_nodes_list )
{
    int i, size;
    int startToken = 0;
    char out_node[LINE_MAX];
    TopologylineParserStates state = BEFORE_NODE_NUMBER;

    assertion( line != NULL, "parse_topology_line: line is NULL" );
    assertion( node_number != NULL, "parse_topology_line: node_number is NULL" );
    assertion( out_nodes_list != NULL, "parse_topology_line: out_nodes_list is NULL" );

#ifdef DEBUG_MCC
    printf("MCC: parse_topology_line, line:<%s>\n", line);
#endif
    size = strlen( line );
    for( i=0; i < size; i++ ) {

        if( isspace(line[i]) ) {
            switch( state ) {
                case BEFORE_NODE_NUMBER:
                case AFTER_NODE_NUMBER:
                case BEFORE_NODE_IN_LIST:
                case AFTER_NODE_IN_LIST:
                    continue;
                case NODE_NUMBER:
                    state = AFTER_NODE_NUMBER;
                    memcpy( node_number, line + startToken, i - startToken );
                    node_number[i - startToken] = 0;
                    break;
                case NODE_IN_LIST:
                    state = AFTER_NODE_IN_LIST;
                    memcpy( out_node, line + startToken, i - startToken );
                    out_node[i - startToken] = 0;
                    append_List( r(*out_nodes_list), r(create_String(out_node)) );
                    break;
            }
        } else if( isdigit(line[i]) ) {
            switch( state ) {
                case BEFORE_NODE_NUMBER:
                    state = NODE_NUMBER;
                    startToken = i;
                    break;
                case NODE_NUMBER:
                case NODE_IN_LIST:
                    continue;
                case AFTER_NODE_NUMBER:
                case AFTER_NODE_IN_LIST:
                    return 0;
                case BEFORE_NODE_IN_LIST:
                    state = NODE_IN_LIST;
                    startToken = i;
                    break;
            }
        } else if ( line[i] == ':') {
            switch( state ) {
                case BEFORE_NODE_NUMBER:
                case BEFORE_NODE_IN_LIST:
                case NODE_IN_LIST:
                case AFTER_NODE_IN_LIST:
                    return 0;
                case NODE_NUMBER:
                    state = BEFORE_NODE_IN_LIST;
                    memcpy( node_number, line + startToken, i - startToken );
                    node_number[i - startToken] = 0;
                    break;
                case AFTER_NODE_NUMBER:
                    state = BEFORE_NODE_IN_LIST;
                    break;
            }
        } else if ( line[i] == ',') {
            switch( state ) {
                case BEFORE_NODE_NUMBER:
                case NODE_NUMBER:
                case AFTER_NODE_NUMBER:
                case BEFORE_NODE_IN_LIST:
                    return 0;
                case NODE_IN_LIST:
                    state = BEFORE_NODE_IN_LIST;
                    memcpy( out_node, line + startToken, i - startToken );
                    out_node[i - startToken] = 0;
                    append_List( r(*out_nodes_list), r(create_String(out_node)) );
                    break;
                case AFTER_NODE_IN_LIST:
                    state = BEFORE_NODE_IN_LIST;
                    break;
            }
        } else {
            return 0;
        }

    } /* for */
    if( state == NODE_IN_LIST) {
        memcpy( out_node, line + startToken, size - startToken + 1);
        out_node[size - startToken + 1] = 0;
        append_List( r(*out_nodes_list), r(create_String(out_node)) );
        state = AFTER_NODE_IN_LIST;
    }
    if( state == BEFORE_NODE_NUMBER ) return -1; /* empty line */
    if( state != AFTER_NODE_IN_LIST ) return 0;

    return 1;
}

