/*Copyright 2010-2012 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Protsenko (protsenko@ispras.ru)
// SYNOPSIS: Fuzzy Associative Elimination 
//           (Rule # 9 in C++TESK.ErrorDiagnostics.ru.doc):
//           {[y1:x2]+[x1:y2] => [y1:y2]+[x1:x2]}
///////////////////////////////////////////////////////////////////////////////

#include <hw/diagnostics/rule/fuzzy_associative.hpp>

namespace cpptesk {
namespace hw {
namespace diagnostics {

    ReactionSet FuzzyAssociativeEliminationRule::transform(::cpptesk::hw::diagnostics::ReactionUnit &reaction_unit_1, 
                                                           ::cpptesk::hw::diagnostics::ReactionUnit &reaction_unit_2)    
    // Collapse the reactions
    {
        ReactionSet transform_out;
        
        const ::cpptesk::hw::Message *message_1_spec, *message_1_impl;
        const ::cpptesk::hw::Message *message_2_impl, *message_2_spec;
        
        // Message from reaction 1
        message_1_spec = reaction_unit_1.getReaction().getSpecMessage();
        message_1_impl = reaction_unit_1.getReaction().getImplMessage();
        // Message from reaction 2
        message_2_spec = reaction_unit_2.getReaction().getSpecMessage();
        message_2_impl = reaction_unit_2.getReaction().getImplMessage();
        
        // Get iface
        const ::cpptesk::hw::Interface *iface_1 = reaction_unit_1.getReaction().where();
        const ::cpptesk::hw::Interface *iface_2 = reaction_unit_2.getReaction().where();
        
        // Get "start" and "end" time of reactions
        cycle_t start_time   =  reaction_unit_1.getReaction().whenStarted();
        cycle_t start_time_2 =  reaction_unit_2.getReaction().whenStarted();
        cycle_t end_time     =  reaction_unit_1.getReaction().whenStopped();
        cycle_t end_time_2   =  reaction_unit_2.getReaction().whenStopped();
        
        if(start_time > start_time_2) start_time = start_time_2; // Min start time
        if(end_time < end_time_2)     end_time = end_time_2;     // Max end time
        
        ::cpptesk::hw::Reaction reaction_incorrect_1 = Reaction(INCORRECT_REACTION, *iface_1, *message_1_spec, *message_2_impl, start_time, end_time, "Incorrect reaction");
        ::cpptesk::hw::Reaction reaction_incorrect_2 = Reaction(INCORRECT_REACTION, *iface_2, *message_2_spec, *message_1_impl, start_time, end_time, "Incorrect reaction");
        
        transform_out.addReaction(reaction_incorrect_1, reaction_unit_1.getIdentifier(), reaction_unit_2.getIdentifier());
        transform_out.addReaction(reaction_incorrect_2, reaction_unit_2.getIdentifier(), reaction_unit_1.getIdentifier());
        
        // Set flag
        reaction_unit_1.setCollapsed();
        reaction_unit_2.setCollapsed();
       
        return transform_out;
    }

    int FuzzyAssociativeEliminationRule::match(::cpptesk::hw::Reaction reaction_1, 
                                               ::cpptesk::hw::Reaction reaction_2)
    // Compare the reactions
    {
        int match   = FALSE_MATCH;
        int match_11 = FALSE_MATCH, 
            match_22 = FALSE_MATCH,  
            match_12 = FALSE_MATCH, 
            match_21 = FALSE_MATCH;

        if((reaction_1.getError() == INCORRECT_REACTION)&&(reaction_2.getError() == INCORRECT_REACTION))
        {
            // Message from reaction 1
            const ::cpptesk::hw::Message *message_1_spec = reaction_1.getSpecMessage();
            const ::cpptesk::hw::Message *message_1_impl = reaction_1.getImplMessage();
            // Message from reaction 2
            const ::cpptesk::hw::Message *message_2_spec = reaction_2.getSpecMessage();
            const ::cpptesk::hw::Message *message_2_impl = reaction_2.getImplMessage();
    
            if(message_1_spec->size() == message_2_spec->size())
            {
                match = FUZZY_MATCH;
                
                // Check base match reaction # 1
                match_11 = FUZZY_MATCH;
                for(::cpptesk::hw::Message::const_iterator i = message_1_spec->begin(), j = message_1_impl->begin();
                    i != message_1_spec->end(); i++, j++)
                {
                    const Field<> &field_spec = (*message_1_spec).**i;
                    const Field<> &field_impl = (*message_1_impl).**j;

                    if(field_spec.equals(field_impl))
                    {
                        match_11--;
                    }
                }
                
                // Check base match reaction # 2
                match_22 = FUZZY_MATCH;
                for(::cpptesk::hw::Message::const_iterator i = message_2_spec->begin(), j = message_2_impl->begin();
                    i != message_2_spec->end(); i++, j++)
                {
                    const Field<> &field_spec = (*message_2_spec).**i;
                    const Field<> &field_impl = (*message_2_impl).**j;

                    if(field_spec.equals(field_impl))
                    {
                        match_22--;
                    }
                }
                
                // Check match reaction # 1 Impl & reaction # 2 Spec
                match_12 = FUZZY_MATCH;
                for(::cpptesk::hw::Message::const_iterator i = message_2_spec->begin(), j = message_1_impl->begin();
                    i != message_2_spec->end(); i++, j++)
                {
                    const Field<> &field_spec = (*message_2_spec).**i;
                    const Field<> &field_impl = (*message_1_impl).**j;

                    if(field_spec.equals(field_impl))
                    {
                        match_12--;
                    }
                }
                
                // Check match reaction # 2 Impl & reaction # 1 Spec
                match_21 = FUZZY_MATCH;
                for(::cpptesk::hw::Message::const_iterator i = message_1_spec->begin(), j = message_2_impl->begin();
                    i != message_1_spec->end(); i++, j++)
                {
                    const Field<> &field_spec = (*message_1_spec).**i;
                    const Field<> &field_impl = (*message_2_impl).**j;

                    if(field_spec.equals(field_impl))
                    {
                        match_21--;
                    }
                }
                
                if(((match_21 < match_11)&&(match_21 < match_22))||((match_12 < match_11)&&(match_12 < match_22)))
                {
                    if(match_12 < match_21)
                    {
                        match = match_12;
                    }
                    else
                    {
                        match = match_21;
                    }
                }
            }
        }
        return match;
    }
}}}
