/*Copyright 2010-2012 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Protsenko (protsenko@ispras.ru)
// SYNOPSIS: Fuzzy Spec Intersection Elimination 
//           (Rule # 12 in C++TESK.ErrorDiagnostics.ru.doc):
//           {[x:Z2]+[Z1:Y2] => [x:Y1]+[Z1:Z2]}
///////////////////////////////////////////////////////////////////////////////

#include <hw/diagnostics/rule/fuzzy_spec_intersection.hpp>

namespace cpptesk {
namespace hw {
namespace diagnostics {

    ReactionSet FuzzySpecIntersectionEliminationRule::transform(::cpptesk::hw::diagnostics::ReactionUnit &reaction_unit_1, 
                                                                ::cpptesk::hw::diagnostics::ReactionUnit &reaction_unit_2)    
    // Collapse the reactions
    {
        ReactionSet transform_out;

        // Get "start" and "end" time of reactions
        cycle_t start_time   =  reaction_unit_1.getReaction().whenStarted();
        cycle_t start_time_2 =  reaction_unit_2.getReaction().whenStarted();
        cycle_t end_time     =  reaction_unit_1.getReaction().whenStopped();
        cycle_t end_time_2   =  reaction_unit_2.getReaction().whenStopped();
        
        if(start_time > start_time_2) start_time = start_time_2; // Min start time
        if(end_time < end_time_2)     end_time = end_time_2;     // Max end time

        if(reaction_unit_1.getReaction().getError() == MISSING_REACTION)
        {
            const ::cpptesk::hw::Message *message_1_spec,
                                         *message_2_impl, 
                                         *message_2_spec;
            
            // Message from reaction 1 [UNEXPECTED] 
            message_1_spec = reaction_unit_1.getReaction().getSpecMessage();
            // Message from reaction 2 [INCORRECT]
            message_2_spec = reaction_unit_2.getReaction().getSpecMessage();
            message_2_impl = reaction_unit_2.getReaction().getImplMessage();
            
            const ::cpptesk::hw::Interface *iface_1 = reaction_unit_1.getReaction().where();
            const ::cpptesk::hw::Interface *iface_2 = reaction_unit_2.getReaction().where();
 
            ::cpptesk::hw::Reaction reaction_incorrect = Reaction(INCORRECT_REACTION, *iface_1, *message_1_spec, *message_2_impl, 
                                                                  start_time, end_time, "Incorrect reaction");
            ::cpptesk::hw::Reaction reaction_missing   = Reaction(MISSING_REACTION,   *iface_2, *message_2_spec, 
                                                                  start_time, end_time, "Missing reaction");

            transform_out.addReaction(reaction_incorrect, reaction_unit_1.getIdentifier(), reaction_unit_2.getIdentifier());
            transform_out.addReaction(reaction_missing,   reaction_unit_2.getIdentifier(), -1);
        }
        else
        {
            const ::cpptesk::hw::Message *message_1_impl, 
                                         *message_1_spec,
                                         *message_2_spec;
            
            // Message from reaction 1 [INCORRECT]
            message_1_spec = reaction_unit_1.getReaction().getSpecMessage();
            message_1_impl = reaction_unit_1.getReaction().getImplMessage();
            // Message from reaction 1 [UNEXPECTED] 
            message_2_spec = reaction_unit_2.getReaction().getSpecMessage();
            
            const ::cpptesk::hw::Interface *iface_1 = reaction_unit_1.getReaction().where();
            const ::cpptesk::hw::Interface *iface_2 = reaction_unit_2.getReaction().where();
            
            ::cpptesk::hw::Reaction reaction_incorrect = Reaction(INCORRECT_REACTION, *iface_2, *message_2_spec, *message_1_impl, 
                                                                  start_time, end_time, "Incorrect reaction");
            ::cpptesk::hw::Reaction reaction_missing   = Reaction(MISSING_REACTION,   *iface_1, *message_1_spec, 
                                                                  start_time, end_time, "Missing reaction");
            
            transform_out.addReaction(reaction_incorrect, reaction_unit_2.getIdentifier(), reaction_unit_1.getIdentifier());
            transform_out.addReaction(reaction_missing,   reaction_unit_1.getIdentifier(), -1);
        }
        
        // Set flag
        reaction_unit_1.setCollapsed();
        reaction_unit_2.setCollapsed();

        return transform_out;
    }

    int FuzzySpecIntersectionEliminationRule::match(::cpptesk::hw::Reaction reaction_1, 
                                                    ::cpptesk::hw::Reaction reaction_2)
    // Compare the reactions
    {
        int match   = FALSE_MATCH, 
            match_1 = FALSE_MATCH, 
            match_2 = FALSE_MATCH;

        if((reaction_1.getError() == INCORRECT_REACTION)&&(reaction_2.getError() == MISSING_REACTION))
        {
            // Message from reaction 1
            // Get Impl & Spec from incorrect reaction 1
            const ::cpptesk::hw::Message *message_1_spec = reaction_1.getSpecMessage();
            const ::cpptesk::hw::Message *message_1_impl = reaction_1.getImplMessage();
            // Message from reaction 2
            // Get Spec from unexpected reaction 2
            const ::cpptesk::hw::Message *message_2_spec = reaction_2.getSpecMessage();
            
            if(message_1_spec->size() == message_2_spec->size())
            {
                match = FUZZY_MATCH;
                
                match_1 = FUZZY_MATCH;
                // Check match [reaction#1 Impl] & [reaction#1 Spec]
                for(::cpptesk::hw::Message::const_iterator i = message_1_spec->begin(), j = message_1_impl->begin();
                    i != message_1_spec->end(); i++, j++)
                {
                    const Field<> &field_spec = (*message_1_spec).**i;
                    const Field<> &field_impl = (*message_1_impl).**j;

                    if(field_spec.equals(field_impl))
                    {
                        match_1--;
                    }
                }

                match_2 = FUZZY_MATCH;
                // Check match [reaction#1 Impl] & [reaction#2 Spec]
                for(::cpptesk::hw::Message::const_iterator i = message_1_impl->begin(), j = message_2_spec->begin();
                    i != message_1_impl->end(); i++, j++)
                {
                    const Field<> &field_spec = (*message_2_spec).**i;
                    const Field<> &field_impl = (*message_1_impl).**j;

                    if(field_spec.equals(field_impl))
                    {
                        match_2--;
                    }
                }
                
                if(match_1 > match_2)
                {
                    match = match_2;
                }
                // if _match2 >= _match1, then _match = FUZZY_MATCH
            }
        }

        if((reaction_2.getError() == INCORRECT_REACTION)&&(reaction_1.getError() == MISSING_REACTION))
        {
            // Message from reaction 1
            // Get Spec from unexpected reaction 1
            const ::cpptesk::hw::Message *message_1_spec = reaction_1.getSpecMessage();
            // Message from reaction 2
            // Get Impl & Spec from incorrect reaction 2
            const ::cpptesk::hw::Message *message_2_spec = reaction_2.getSpecMessage();
            const ::cpptesk::hw::Message *message_2_impl = reaction_2.getImplMessage();
    
            if(message_1_spec->size() == message_2_spec->size())
            {
                match = FUZZY_MATCH;
                
                match_1 = FUZZY_MATCH;
                // Check match [reaction#2 Impl] & [reaction#1 Spec]
                for(::cpptesk::hw::Message::const_iterator i = message_1_spec->begin(), j = message_2_impl->begin();
                    i != message_1_spec->end(); i++, j++)
                {
                    const Field<> &field_spec = (*message_1_spec).**i;
                    const Field<> &field_impl = (*message_2_impl).**j;

                    if(field_spec.equals(field_impl))
                    {
                        match_1--;
                    }
                }

                match_2 = FUZZY_MATCH;
                // Check match [reaction#2 Impl] & [reaction#2 Spec]
                for(::cpptesk::hw::Message::const_iterator i = message_2_spec->begin(), j = message_2_impl->begin();
                    i != message_2_spec->end(); i++, j++)
                {
                    const Field<> &field_spec = (*message_2_spec).**i;
                    const Field<> &field_impl = (*message_2_impl).**j;

                    if(field_spec.equals(field_impl))
                    {
                        match_2--;
                    }
                }
                
                if(match_1 < match_2)
                {
                    match = match_1;
                }
                // if _match1 >= _match2, then _match = FUZZY_MATCH
            }
        }
        
        return match;
    }
    
}}}
